/*
 *  (c) Copyright Hewlett-Packard Company 2001
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: _StoreBdb.java,v 1.9 2002/09/18 11:58:00 jeremy_carroll Exp $
 */

package com.hp.hpl.mesa.rdf.jena.bdb;

import com.hp.hpl.mesa.rdf.jena.common.*;
import com.hp.hpl.mesa.rdf.jena.model.*;

import com.hp.hpl.jena.util.Log;

import com.sleepycat.db.Db;
import com.sleepycat.db.Dbc;
import com.sleepycat.db.DbEnv;
import com.sleepycat.db.Dbt;
import com.sleepycat.db.DbMemoryException;
import com.sleepycat.db.DbException;
import com.sleepycat.db.DbTxn;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.File;
import java.io.DataOutputStream;
import java.io.IOException;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.NoSuchElementException;
import java.util.Vector;


/**
 *
 * @author  bwm
 * @version
 */
abstract class _StoreBdb extends StoreCom implements Store {
    
    protected static final int MAGICNUMBER = 0x0BDB0BDB;
    protected static final String cls = _StoreBdb.class.getName();
    
    Db db0;         // global data
    Db db1;         // sp->o
    Db db2;         // po->s
    Db db3;         // os->p
    
    Model model = null;
    
    protected static String envDir;
    
    protected File    lockFile;           // used for locking the database
    protected boolean open;               // has the store been closed?
    
    int       size;                       // number of statements in the store
    boolean   stmtAdded;                  // flag set when add adds a stmt
    boolean   stmtDeleted;                // flag set when remove deletes stmt
    
    String    filename;
    
    class Configuration {
        int  magicNumber;
        int  version;
        int size;
    };
    Configuration configuration = new Configuration();
    
    {
        configuration.magicNumber = MAGICNUMBER;
        configuration.version     = 1;
        configuration.size        = 0;
    }
    
    protected static DbEnv dbEnv = null;
    
    public _StoreBdb(String envdir, String filename, Model m)
    throws BDBException {
        final String method = cls;
        LLog.debug("called (" + envdir + ", " + filename + ", ...)",
        cls, method);
        model = m;
        this.filename = filename;
        createEnv(envdir);
        openDb(filename);
        if (readConfig() == 0) {
            size = configuration.size;
        } else {
            size = -1;
        }
        configuration.size = -1;             //@@temp mark size as dirty
        writeConfig(configuration);
        open = true;
        LLog.debug("exiting", cls, method);
    }
    
    public synchronized void close() {
        final String method = "close";
        LLog.debug("called", cls, method);
        try {
            if (open) {
                configuration.size = size;
                try {
                    writeConfig(configuration);    // clear the dirty size
                } catch (BDBException e) {
                    Log.warning("Failed to update size when closing database",
                    "_StoreDb", "close", e);
                }
                if (! lockFile.delete()) {
                    Log.warning("failed to delete lock file",
                    "_StoreDb", "close()");
                }
                db0.close(0);
                db1.close(0);
                db2.close(0);
                db3.close(0);
            }
        } catch (DbException e) {
            Log.warning("Exception whilst closing database",
            "_StoreDb", "close()", e);
        }
        open = false;
        LLog.debug("exiting", cls, method);
    }
    
    protected DbEnv createEnv(String envdir) throws BDBException {
        final String method = "createEnv";
        LLog.debug("called (" + envdir + ")", cls, method);
        // if envdir is null, use the default
        if (envdir == null) {
            envDir = Util.getProperty(
            Jena.PATH + ".bdbenv", Jena.DIR +"/RDF/bdb");
        } else {
            envDir = envdir;
        }
        
        // create the environment directory if it does not already exist
        
        File file = new File(envDir);
        if (file.exists()) {
            if (!file.isDirectory()) {
                Log.severe("BDB env is not a directory",
                "StoreDb", "createEnv");
                throw new BDBException(BDBException.BADENVDIR);
            }
        } else {
            if (!file.mkdirs()) {
                Log.severe("Cannot create DBD env directory",
                "StoreDb", "createEnv");
                throw new BDBException(BDBException.CREATEENVDIR);
            }
        }
        
        try {
            dbEnv = new MyDbEnv(0);  //@@ fix should be DbEnv
            dbEnv.open(file.getAbsolutePath(),
            //                       Db.DB_RECOVER    |
            //                       Db.DB_INIT_LOG   |
            //                       Db.DB_INIT_TXN   |
            Db.DB_CREATE     |
            Db.DB_INIT_MPOOL,
            0);
        } catch (Exception e) {
            Log.severe("failed to open Berkeley DB environment",
            "StoreDb", "createEnv", e);
            throw new RDFError(e);
        }
        
        LLog.debug("exiting", cls, method);
        return dbEnv;
    }
    
    protected void openDb(String filename) throws BDBException {
        final String method = "openDb";
        LLog.debug("called (" + filename + ")", cls, method);
        try {
            // create the lock file
            lockFile = new File(envDir + "/" + filename + ".lck");
            if (!lockFile.createNewFile()) {
                throw new BDBException(BDBException.DBALREADYOPEN);
            }
            lockFile.deleteOnExit();
        } catch (java.io.IOException e) {
            Log.severe("lockfile error",
            this.getClass().getName(),
            "openDb",
            e);
            throw new RDFError(e);
        }
        
        try {
            db0 = new Db(dbEnv, 0);
            db0.open(filename, "config", Db.DB_RECNO, Db.DB_CREATE, 0);
            // "SP->O"
            db1 = new Db(dbEnv, 0);
            db1.set_flags(Db.DB_DUPSORT);
            db1.open(filename, "SP->O", Db.DB_BTREE, Db.DB_CREATE, 0);
            // "PO->S"
            db2 = new Db(dbEnv,0);
            db2.set_flags(Db.DB_DUPSORT);
            db2.open(filename, "PO->S", Db.DB_BTREE, Db.DB_CREATE, 0);
            // "OS->P"
            db3 = new Db(dbEnv, 0);
            db3.set_flags(Db.DB_DUPSORT);
            db3.open(filename, "OS->P" , Db.DB_BTREE, Db.DB_CREATE, 0);
        } catch (Exception e) {
            Log.severe("failed to open Berkeley DB database",
            "_StoreDb", "openDb", e);
            throw new BDBException(BDBException.DBCORRUPT);
        }
        LLog.debug("exiting", cls, method);
    }
    
    protected int readConfig() throws BDBException {
        final String method = "readConfig";
        LLog.debug("calling()", cls, method);
        byte[] keydata = new byte[4];
        Dbt key = new Dbt(keydata);
        key.set_ulen(4);
        key.set_recno_key_data(1);
        key.set_flags(Db.DB_DBT_USERMEM);
        byte[] valdata = configToBytes(configuration);
        Dbt val = new Dbt(valdata);
        val.set_ulen(valdata.length);
        val.set_flags(Db.DB_DBT_USERMEM);
        
        int status;
        try {
            status = db0.get(null, key, val, 0);
        } catch (DbException e) {
            Log.severe("failed to read DB configuration record",
            this.getClass().getName(),
            "readConfig",
            e);
            throw new RDFError(e);
        }
        
        configuration = configFromBytes(val.get_data());
        
        LLog.debug("exiting", cls, method);
        return status;
    }
    
    protected void writeConfig(Configuration config) throws BDBException {
        final String method = "writeConfig";
        Log.debug("called(config)", cls, method);
        byte[] keydata = new byte[4];
        Dbt key = new Dbt(keydata);
        key.set_ulen(4);
        key.set_recno_key_data(1);
        key.set_flags(Db.DB_DBT_USERMEM);
        byte[] valdata = configToBytes(config);
        Dbt val = new Dbt(valdata);
        val.set_ulen(valdata.length);
        val.set_flags(Db.DB_DBT_USERMEM);
        try {
            db0.put(null, key, val, 0);
        } catch (DbException e) {
            throw new BDBException(e);
        }
        LLog.debug("exiting", cls, method);
    }
    
    protected byte[] configToBytes(Configuration config) {
        try {
            ByteArrayOutputStream bos = new ByteArrayOutputStream();
            DataOutputStream os = new DataOutputStream(bos);
            os.writeInt(config.magicNumber);
            os.writeInt(config.version);
            os.writeInt(config.size);
            return bos.toByteArray();
        } catch (IOException e) {
            Log.severe("Failed to convert configuration to bytes",
            "_StoreDb", "configToBytes", e);
            throw new RDFError(e);
        }
    }
    
    protected Configuration configFromBytes(byte [] bytes) throws BDBException {
        Configuration config = new Configuration();
        ByteArrayInputStream bis = new ByteArrayInputStream(bytes);
        DataInputStream is = new DataInputStream(bis);
        try {
            config.magicNumber = is.readInt();
            if (config.magicNumber != MAGICNUMBER) {
                throw new BDBException(BDBException.BADMAGIC);
            }
            config.version = is.readInt();
            config.size = is.readInt();
            return config;
        } catch (IOException e) {
            throw new BDBException(BDBException.HEADERCORRUPT);
        }
    }
    
    
    protected synchronized void add(Statement s, DbTxn txn) {
        final String method = "add";
        LLog.debug("add(" + s + ")", cls, method);
        
        byte[] keybytes = null;
        
        byte[] bSubject   = ByteSerializer.toBytes(s.getSubject());
        byte[] bPredicate = ByteSerializer.toBytes(s.getPredicate());
        byte[] bObject    = ByteSerializer.toBytes(s.getObject());
        
        try {
            
            // the following code is rather defensive as a search is made
            // for intermittent errors possible due to memory allocation
            // problems or buffer size problems.
            // tidy up later once its working
            // @@fixme
            keybytes = concat(bSubject, bPredicate);
            Dbt key = new Dbt(keybytes);
            key.set_ulen(keybytes.length);
            key.set_flags(Db.DB_DBT_USERMEM);
            
            Dbt val  = new Dbt(bObject);
            val.set_ulen(bObject.length);
            key.set_flags(Db.DB_DBT_USERMEM);
            
            if (db1.put(txn, key, val, Db.DB_NODUPDATA)
            != Db.DB_KEYEXIST) {
                
                keybytes = concat(bPredicate, bObject);
                key = new Dbt(keybytes);
                key.set_ulen(keybytes.length);
                key.set_flags(Db.DB_DBT_USERMEM);
                
                val  = new Dbt(bSubject);
                val.set_ulen(bSubject.length);
                val.set_flags(Db.DB_DBT_USERMEM);
                
                db2.put(txn, key, val, Db.DB_NODUPDATA);
                
                keybytes = concat(bObject, bSubject);
                key = new Dbt(keybytes);
                key.set_ulen(keybytes.length);
                key.set_flags(Db.DB_DBT_USERMEM);
                
                val  = new Dbt(bPredicate);
                val.set_ulen(bPredicate.length);
                val.set_flags(Db.DB_DBT_USERMEM);
                
                db3.put(txn, key, val, Db.DB_NODUPDATA);
                stmtAdded = true;
            }
        } catch (DbException e) {
            Log.severe("Failed to add statement",
            "StoreDb", "add", e);
            throw new RDFError(e);
        }
        LLog.debug("exiting", cls, method);
    }
    
    protected synchronized void remove(Statement s, DbTxn txn) {
        final String method = "remove";
        LLog.debug("called(" + s + ")", cls, method);
        byte[] bSubject   = ByteSerializer.toBytes(s.getSubject());
        byte[] bPredicate = ByteSerializer.toBytes(s.getPredicate());
        byte[] bObject    = ByteSerializer.toBytes(s.getObject());
        remove(db1, concat(bSubject, bPredicate), bObject, txn);
        remove(db2, concat(bPredicate, bObject),  bSubject, txn);
        remove(db3, concat(bObject, bSubject),    bPredicate, txn);
        Log.debug("exiting", cls, method);
    }
    
    protected void remove(Db db, byte[] k, byte[] v, DbTxn txn) {
        final String method = "remove";
        LLog.debug("called(Db, byte[], byte[], DbTxn)", cls, method);
        Dbc cur = null;
        try {
            cur = db.cursor(txn, 0);  // @@ txn
            
            Dbt key = new Dbt(k);
            key.set_flags(Db.DB_DBT_USERMEM);
            key.set_ulen(k.length);
            
            Dbt val = new Dbt(v);
            val.set_flags(Db.DB_DBT_USERMEM);
            val.set_ulen(v.length);
            
            if (cur.get(key, val, Db.DB_GET_BOTH) != Db.DB_NOTFOUND) {
                stmtDeleted = true;
                cur.del(0);
            }
        } catch (DbException e) {
            Log.severe("remove statement failed",
            "StoreDb", "remove(Db, byte[], byte[])", e);
            throw new RDFError(e);
        } finally {
            try {
                if (cur != null) cur.close();
            } catch (DbException e2) {
                Log.severe("failed to close cursor",
                "StoreDb", "remove(Db, byte[], byte[])", e2);
                throw new RDFError(e2);
            }
        }
        LLog.debug("exiting", cls, method);
    }
    
    public synchronized int size() throws RDFException {
        final String method = "size";
        LLog.debug("called", cls, method);
        //@@temp        if (size < 0) {
        // the size field in the header was dirty
        
        Iterator iter  = list();
        
        size = 0;
        
        while (iter.hasNext()) {
            size++;
            iter.next();
        }
        //@@temp        }
        
        LLog.debug("result = " + size, cls, method);
        return size;
    }
    
    public synchronized boolean contains(Statement s) throws RDFException {
        // doesn't really throw RDFException - @@Fix
        final String method = "contains";
        LLog.debug("called(" + s + ")", cls, method);
        
        byte[] bSubject   = ByteSerializer.toBytes(s.getSubject());
        byte[] bPredicate = ByteSerializer.toBytes(s.getPredicate());
/* We can' just do a straight db lookup, unfortunately.  The represenation
 * of literal includes its well formedness flag, but two literals can be
 * equal even when there well formedness flags are different.  Hence this
 * old code is commented out and we must retrieve the statements and test
 * for equality.
 
        byte[] bObject    = ByteSerializer.toBytes(s.getObject());
 
        try {
            Dbt key = new Dbt(concat(bSubject, bPredicate));
            Dbt val = new Dbt(bObject);
            key.set_flags(Db.DB_DBT_USERMEM);
            key.set_ulen(bSubject.length + bPredicate.length);
            val.set_flags(Db.DB_DBT_USERMEM);
            val.set_ulen(bObject.length);
            return !(db1.get(null,key, val, Db.DB_GET_BOTH) == Db.DB_NOTFOUND);
        } catch (Exception e) {
            Log.severe("check for presence failed",
                       "StoreDb", "contains(Statement)", e);
            throw new RDFUnexpectedException(e);
        } */
        RDFNode object = s.getObject();
        Iterator iter = listFullKey(db1,
        concat(bSubject,
        bPredicate), 4);
        boolean result = false;
        while (iter.hasNext()) {
            if (((Statement)iter.next()).getObject().equals(object)) {
                result = true;
                break;
            }
        }
        ((ClosableIterator)iter).close();
        LLog.debug("result = " + result, cls, method);
        return result;
    }
    
    public synchronized boolean contains(Resource s, Property p)
    throws RDFException {
        // doesn't really throw RDFException - @@Fix
        final String method = "contains";
        LLog.debug("called(" + s + ", " + p + ")", cls, method);
        
        byte[] bSubject   = ByteSerializer.toBytes(s);
        byte[] bPredicate = ByteSerializer.toBytes(p);
        
        try {
            byte[] keydata = concat(bSubject, bPredicate);
            Dbt key = new Dbt(keydata);
            key.set_flags(Db.DB_DBT_USERMEM);
            key.set_ulen(keydata.length);
            
            byte[] valdata = new byte[0];
            Dbt val = new Dbt(valdata);
            val.set_flags(Db.DB_DBT_USERMEM | Db.DB_DBT_PARTIAL);
            val.set_ulen(valdata.length);
            
            boolean result =  !(db1.get(null,key, val, 0) == Db.DB_NOTFOUND);
            LLog.debug("result = " + result, cls, method);
            return result;            
        } catch (DbException e) {
            Log.severe("check for presence failed",
            "StoreDb", "contains(Resource, Property)", e);
            throw new RDFError(e);
        }
    }
    
    public synchronized Iterator list() throws RDFException {
        final String method = "list";
        LLog.debug("called()", cls, method);
        
        return new Lookup(db1, this, new byte[0], 4, Db.DB_FIRST, Db.DB_NEXT);
    }
    
    public synchronized Iterator list(Resource subject,
    Property predicate, RDFNode object)
    throws RDFException {
        final String method = "list";
        LLog.debug("called(" + subject + ", " + predicate + ", " + object + ")",
        cls, method);
        int comb = 0;
        if (subject   != null) comb = comb | 1;
        if (predicate != null) comb = comb | 2;
        if (object    != null) comb = comb | 4;
        
        switch (comb) {
            case 0:
                return list();
                
            case 1:
                return listBySubject(subject);
                
            case 2:
                return listByPredicate(predicate);
                
            case 4:
                return listByObject(object);
                
            case 3:
                return listFullKey(db1,
                concat(ByteSerializer.toBytes(subject),
                ByteSerializer.toBytes(predicate)), 4);
            case 5:
                return listFullKey(db3,
                concat(ByteSerializer.toBytes(object),
                ByteSerializer.toBytes(subject)), 2);
                
            case 6:
                return listFullKey(db2,
                concat(ByteSerializer.toBytes(predicate),
                ByteSerializer.toBytes(object)), 1);
                
            case 7:
                HashSet stmts = new HashSet();
                StatementImpl stmt =
                new StatementImpl(subject, predicate, object, model);
                if (contains(stmt)) {
                    stmts.add(stmt);
                }
                return stmts.iterator();
        }
        LLog.debug("result = null", cls, method);
        return null;
    }
    
    protected Iterator listFullKey(Db db, byte[] bKey, int vx) {
        final String method = "listFullKey";
        LLog.debug("called(DB, byte[]," + vx, cls, method);
        return new Lookup(db, this, bKey, vx, Db.DB_SET, Db.DB_NEXT_DUP);
    }
    
    protected Iterator listPartialKey(Db db, RDFNode nKey, int vx) {
        final String method = "listPartialKey";
        LLog.debug("called(DB, " + nKey + ", " + vx, cls, method);
        return new LookupPartialKey(db, this, nKey, vx, Db.DB_SET_RANGE,
        Db.DB_NEXT);
    }
    
    public synchronized Iterator listBySubject(Resource subject) {
        final String method = "listBySubject";
        LLog.debug("called(" + subject + ")", cls, method );
        return listPartialKey(db1, subject, 4);
    }
    
    public synchronized Iterator listByPredicate(Property property) {
        final String method = "listByPredicate";
        LLog.debug("called(" + property + ")", cls, method);
        return listPartialKey(db2, property, 1);
    }
    
    public synchronized Iterator listByObject(RDFNode object) {
        Log.debug(Thread.currentThread() + ": "
        + "_StoreBdb.listByObject(" + object + ")" );
        return listPartialKey(db3, object, 2);
    }
    
    public synchronized ResIterator listSubjects() {
        final String method = "listSubjects";
        LLog.debug("called()", cls, method);
        return new ResIteratorImpl(new LookupUniqSubj(this), null);
    }
    
    public synchronized NsIterator listNameSpaces() {
        final String method = "listNameSpaces";
        LLog.debug("called()", cls, method);
        return new NsIteratorImpl(new LookupUniqNs(this), null);
    }
    
    protected byte[] concat(byte[] a, byte[] b) {
        byte[] result = new byte[a.length + b.length];
        int x = 0;
        System.arraycopy(a, 0, result, 0, a.length);
        System.arraycopy(b, 0, result, a.length, b.length);
        return result;
    }
    
    protected synchronized boolean getOpen() {
        return open;
    }
    
    protected class Lookup implements ClosableIterator {
        final String cls = Lookup.class.getName();
        Db  db;
        _StoreBdb store;
        int vx = 0;
        RDFNode nKey = null;
        Dbt     key;
        Dbt     val;
        Dbc     cur;
        boolean open = false;  // do we have an open cursor?
        int     flagN;
        RDFNode n1;
        RDFNode n2;
        RDFNode n3;
        boolean more;
        RDFNode filterVar = null;
        Statement stmt;
        
        // Need to keep pointers to data buffers in case garbage collector
        // runs and re-allocates data that is in use.
        
        byte[] keydata;
        byte[] valdata;
        
        Lookup(Db db, _StoreBdb store, RDFNode nKey, int vx,
        int flag1, int flagN) {
            final String method = cls;
            LLog.debug("called(db, store, " + nKey + ", " + vx + ", "
            + flag1 + ", " + flagN + ")", cls, method);
            this.nKey = nKey;
            init(db, store, ByteSerializer.toBytes(nKey), vx, flag1, flagN);
            LLog.debug("exiting", cls, method);
        }
        
        Lookup(Db db, _StoreBdb store, byte[] bKey, int vx,
        int flag1, int flagN) {
            final String method = cls;
            LLog.debug("called(db, store, byte[], "  + vx + ", "
            + flag1 + ", " + flagN + ")", cls, method);
            init(db, store, bKey, vx, flag1, flagN);
            LLog.debug("exiting", cls, method);
        }
        
        protected void init(Db db, _StoreBdb store, byte[] bKey, int vx,
        int flag1, int flagN) {
            this.db = db;
            this.store = store;
            this.vx = vx;
            this.flagN = flagN;
            
            try {
                cur  = db.cursor(null, 0);
                open = true;
                
                key = new Dbt(bKey);
                key.set_flags(Db.DB_DBT_USERMEM);
                key.set_ulen(bKey.length);
                keydata = bKey;
                
                valdata = new byte[0];
                val = new Dbt(valdata);
                val.set_flags(Db.DB_DBT_USERMEM);
                val.set_ulen(0);
                
                // get first record in the database
                getNext(flag1);
            } catch (NoSuchElementException e1) {
                more = false;
            } catch (DbException e2) {
                throw new RDFError(e2);
            }
        }
        
        public void close() {
            // have to check whether DB is still open.  If its closed and
            // we try to close the cursor, there will/may be a memory
            // access exception in the BDB code.
            final String method = "close";
            LLog.debug("called()", cls, method);
            synchronized (store) {
                LLog.debug("entering synced section", cls, method);
                if (open && store.getOpen()) {
                    try {
                        cur.close();
                        open = false;
                    } catch (DbException e) {
                        Log.warning("Finalizer failed", "Lookup", "finalize()",
                        e);
                    }
                }
                LLog.debug("exiting synced section", cls, method);
            }
            LLog.debug("exiting", cls, method);
        }
        
        public void finalize() {
            final String method = "finalize";
            LLog.debug("called()", cls, method);
            close();
            LLog.debug("exiting)", cls, method);
        }
        
        public boolean hasNext() {
            return more;
        }
        
        public Object next() throws NoSuchElementException {
            final String method = "next";
            LLog.debug("called()", cls, method);
            
            
            if (!more) {
                throw new NoSuchElementException();
            }
            
            try {
                
                // construct the statement
                switch (vx) { //@@todo replace with table lookup
                    case 1:  // subject is value
                        stmt = new StatementImpl((Resource) n3,
                        (Property) n1,
                        n2,
                        model);
                        break;
                        
                    case 2: // predicate is value
                        stmt = new StatementImpl((Resource) n2,
                        (Property) n3,
                        n1,
                        model);
                        break;
                        
                    case 4: //object is value
                        stmt = new StatementImpl((Resource) n1,
                        (Property) n2,
                        n3,
                        model);
                        break;
                }
                
                // and get the next one
                getNext(flagN);
                LLog.debug("result = " + stmt, cls, method);
                return stmt;
            } catch (Exception e) {
                Log.severe("unexpected exception", "Lookup", "next", e);
                throw new RDFError(e);
            }
        }
        
        public void remove() throws UnsupportedOperationException,
        IllegalStateException {
        /*  We don't actually have to do anything here.
         *  Remove will be called in the StmtIterator which wraps this
         *  iterator.
            try {
                store.remove(stmt);
            } catch (RDFException e) {
                // remove can throw an exception and there is nothing we
                // can do except throw an unchecked exception
                throw new RDFError(e);
            } */
        }
        
        protected void getNext(int flags) {
            final String method = "getNext";
            LLog.debug("called(" + flags + ")", cls, method);
            int result = 0;
            do {                         // need to try 3 times
                // first time to get key buffer size
                // second time to get data buffer size
                // third time to actually succeed
                for (int i=0; i<3; i++) {
                    try {
                        result = cur.get(key, val, flags);
                        break;
                    } catch (DbMemoryException e) {
                        if ( i==2) {  // should have succeeded by third try
                            Log.severe("failed to read record",
                            this.getClass().getName(),
                            "getNext",
                            e);
                            throw new RDFError(e);
                        }
                        // not enough memory in the buffers
                        int size = key.get_size();
                        if (size > key.get_ulen()) {
                            keydata = new byte[size];
                            System.arraycopy(key.get_data(), 0, keydata, 0,
                            key.get_ulen());
                            key.set_data(keydata);
                            key.set_ulen(size);
                        }
                        size = val.get_size();
                        if (size > val.get_ulen()) {
                            valdata = new byte[size];
                            val.set_data(valdata);
                            val.set_ulen(size);
                        }
                    } catch (DbException e2) {
                        Log.severe("failed to read next record",
                        this.getClass().getName(),
                        "getNext",
                        e2);
                        throw new RDFError(e2);
                    }
                }
                
                
                if (result != Db.DB_NOTFOUND) {
                    
                    // extract statement from database record
                    ByteArrayInputStream bis =
                    new ByteArrayInputStream(key.get_data());
                    DataInputStream is  = new DataInputStream(bis);
                    
                    try {
                        n1    = ByteSerializer.fromData(is, model);
                        n2    = ByteSerializer.fromData(is, model);
                        
                        bis = new ByteArrayInputStream(val.get_data());
                        is  = new DataInputStream(bis);
                        n3    = ByteSerializer.fromData(is, model);
                    } catch (java.io.IOException e1) {
                        Log.severe("failed to deserialize record",
                        this.getClass().getName(),
                        "getNext",
                        e1);
                        throw new RDFError(e1);
                    }
                    
                    // have we reached the end?
                    more = ! endCondition();
                } else {
                    more = false;
                    try {
                        cur.close();
                    } catch (DbException e) {
                        Log.severe("failed to close cursor",
                        this.getClass().getName(), "getNext", e);
                        throw new RDFError(e);
                    }
                    open = false;
                }
            } while (more && filter());
            LLog.debug("exiting", cls, method);
        }
        
        protected boolean endCondition() {
            return false;
        }
        
        protected boolean filter() {
            return false;
        }
    }
    
    protected class LookupPartialKey extends Lookup {
        
        LookupPartialKey(Db db, _StoreBdb store, RDFNode nKey, int vx,
        int flag1, int flagN) {
            super(db, store, nKey, vx, flag1, flagN);
        }
        
        protected boolean endCondition() {
            return !n1.equals(nKey);
        }
    }
    
    protected class LookupUniqSubj extends Lookup  {
        
        LookupUniqSubj(_StoreBdb store) {
            super(store.db1, store, new byte[0], 4, Db.DB_FIRST,
            Db.DB_NEXT_NODUP);
        }
        
        public Object next() throws NoSuchElementException {
            Statement stmt = (Statement) super.next();
            return stmt.getSubject();
        }
        
        protected boolean filter() {
            if (filterVar != null && n1.equals(filterVar)) {
                return true;
            }
            filterVar = n1;
            return false;
        }
    }
    
    protected class LookupUniqNs extends Lookup  {
        
        LookupUniqNs(_StoreBdb store) {
            super(store.db2, store, new byte[0], 1, Db.DB_FIRST,
            Db.DB_NEXT_NODUP);
        }
        
        public Object next() throws NoSuchElementException {
            Statement stmt = (Statement) super.next();
            return stmt.getPredicate().getNameSpace();
        }
        
        protected boolean filter() {
            Property p = (Property) n1;
            if (filterVar != null &&
            p.getNameSpace().equals(((Property)filterVar).getNameSpace())) {
                return true;
            }
            filterVar = n1;
            return false;
        }
    }
    
    // This is a kludge to make sure DbEnv is closed when finalised
    // otherwise we get irritating error messages.
    // Needs design change to fix properly.  No time at present  @@fix
    
    class MyDbEnv extends DbEnv {
        static final String cls = "MyDbEnv";
        MyDbEnv(int flags) {
            super(flags);
            String method = cls;
            LLog.debug("exiting", cls, method);
        }
        
        public synchronized void finalize() {
            final String method = "finalize";
            LLog.debug("called()", cls, method);
            try {
                super.close(0);
            } catch (DbException e) {
                Log.warning("Error on closing DbEnv", "MyDbEnv", "finalize", e);
            }
            LLog.debug("exiting", cls, method);
        }
    }
}

// temporary logging class to add thread id to log messages

class LLog {
    public static void debug( String msg, String cls, String method ) {
        Log.debug(Thread.currentThread() + " : " + msg, cls, method);
    }
}