/*
 *  (c) Copyright Hewlett-Packard Company 2001
 *  All rights reserved.
*
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Model.java
 *
 * Created on 11 March 2001, 16:07
 */

package com.hp.hpl.mesa.rdf.jena.common;

import com.hp.hpl.mesa.rdf.jena.common.*;
import com.hp.hpl.mesa.rdf.jena.model.*;
import com.hp.hpl.mesa.rdf.jena.vocabulary.RDF;

import java.io.FileReader;
import java.io.Writer;
import java.io.Reader;
import java.util.HashSet;
import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.Vector;


/** Common methods for model implementations.
 *
 * <P>This class implements common methods, mainly convenience methods, for
 *    model implementations.  It is intended use is as a base class from which
 *    model implemenations can be derived.</P>
 *
 * @author bwm
 * @version Release='$Name:  $' Revision='$Revision: 1.10 $' Date='$Date: 2001/12/06 11:50:21 $'
 */
public abstract class ModelCom extends Object implements Model, ModelI {

    protected Store store = null;

    protected RDFReaderF readerFactory = new RDFReaderFImpl();
    protected RDFWriterF writerFactory = new RDFWriterFImpl();

    // next free error code = 3

    public ModelCom() {
    }

    public Resource getResource(String uri, ResourceF f) throws RDFException {
        try {
            return f.createResource(getResource(uri));
        } catch (Exception e) {
            throw new RDFException(e);
        }
    }

    public Model add(Resource s,Property p,RDFNode o) throws RDFException {
        add(createStatement(s, p, o));
        return this;
    }

    public Model add(Resource s, Property p, boolean o) throws RDFException {
        add(s, p, new LiteralImpl(o));
        return this;
    }

    public Model add(Resource s, Property p, long o) throws RDFException {
        add(s, p, new LiteralImpl(o));
        return this;
    }

    public Model add(Resource s, Property p, char o) throws RDFException {
        add(s, p, new LiteralImpl(o));
        return this;
    }

    public Model add(Resource s, Property p, float o) throws RDFException {
        add(s, p, new LiteralImpl(o));
        return this;
    }

    public Model add(Resource s, Property p, double o) throws RDFException {
        add(s, p, new LiteralImpl(o));
        return this;
    }

    public Model add(Resource s, Property p, String o) throws RDFException {
        add(s, p, new LiteralImpl(o));
        return this;
    }

    public Model add(Resource s, Property p, String o, boolean wellFormed)
      throws RDFException {
        add(s, p, new LiteralImpl(o, wellFormed));
        return this;
    }

    public Model add(Resource s, Property p, String o, String lang,
      boolean wellFormed) throws RDFException {
        add(s, p, new LiteralImpl(o, lang, wellFormed));
        return this;
    }

    public Model add(Resource s, Property p, String o, String l)
       throws RDFException {
        add(s, p, new LiteralImpl(o,l));
        return this;
    }

    public Model add(Resource s, Property p, Object o) throws RDFException {
        if (o instanceof RDFNode) {
            add(s, p, (RDFNode) o);
        } else {
            add(s, p, new LiteralImpl(o));
        }
        return this;
    }

    public Model add(StmtIterator iter) throws RDFException {
        try {
            while (iter.hasNext()) {
                add(iter.next());
            }
            return this;
        } finally {
            iter.close();
        }
    }

    public Model add(Model m) throws RDFException {
        return add(m.listStatements());
    }

    public RDFReader getReader() throws RDFException {
        return readerFactory.getReader();
    }

    public RDFReader getReader(String lang) throws RDFException {
        return readerFactory.getReader(lang);
    }

    public String setReaderClassName(String lang, String className) {
        return readerFactory.setReaderClassName(lang, className);
    }

    public Model read(String url) throws RDFException {
        readerFactory.getReader()
                     .read(this, url);
        return this;
    }

    public Model read(Reader reader, String base) throws RDFException {
        readerFactory.getReader()
                     .read(this, reader, base);
        return this;
    }

    public Model read(String url, String lang) throws RDFException {
        readerFactory.getReader(lang)
                     .read(this, url);
        return this;
    }

    public Model read(Reader reader, String base, String lang)
      throws RDFException {
        readerFactory.getReader(lang)
                     .read(this, reader, base);
        return this;
    }

    public RDFWriter getWriter() throws RDFException {
        return writerFactory.getWriter();
    }

    public RDFWriter getWriter(String lang) throws RDFException {
        return writerFactory.getWriter(lang);
    }

    public String setWriterClassName(String lang, String className) {
        return writerFactory.setWriterClassName(lang, className);
    }

    public Model write(Writer writer) throws RDFException {
        try {
            writerFactory.getWriter()
                         .write(this, writer, "");
            return this;
        } catch (Exception e) {
            throw new RDFException(e);
        }
    }

    public Model write(Writer writer, String lang) throws RDFException {
        try {
            writerFactory.getWriter(lang)
                         .write(this, writer, "");
            return this;
        } catch (Exception e) {
            throw new RDFException(e);
        }
    }


    public Model write(Writer writer, String lang, String base)
      throws RDFException {
        try {
            writerFactory.getWriter(lang)
                         .write(this, writer, base);
            return this;
        } catch (Exception e) {
            throw new RDFException(e);
        }
    }

    public Model remove(StmtIterator iter) throws RDFException {
        while (iter.hasNext()) {
            remove((Statement)iter.next());
        }
        return this;
    }

    public Model remove(Model m) throws RDFException {
        StmtIterator iter = m.listStatements();
        try {
            remove(iter);
        } finally {
            iter.close();
        }
        return this;
    }

    public boolean contains(Resource s, Property p, boolean o)
      throws RDFException {
        return contains(s, p, new LiteralImpl(o));
    }

    public boolean contains(Resource s, Property p, long o)
      throws RDFException {
        return contains(s, p, new LiteralImpl(o));
    }

    public boolean contains(Resource s, Property p, char o)
      throws RDFException {
        return contains(s, p, new LiteralImpl(o));
    }

    public boolean contains(Resource s, Property p, float o)
      throws RDFException {
        return contains(s, p, new LiteralImpl(o));
    }

    public boolean contains(Resource s, Property p, double o)
      throws RDFException {
        return contains(s, p, new LiteralImpl(o));
    }

    public boolean contains(Resource s, Property p, String o)
      throws RDFException {
        return contains(s, p, new LiteralImpl(o));
    }

    public boolean contains(Resource s, Property p, String o, String l)
      throws RDFException {
        return contains(s, p, new LiteralImpl(o,l));
    }

    public boolean contains(Resource s, Property p, Object o)
      throws RDFException {
        if (o instanceof RDFNode) {
            return contains(s, p, (RDFNode) o);
        } else {
            return contains(s, p, new LiteralImpl(o));
        }
    }

    public boolean containsAny(StmtIterator iter) throws RDFException {
        while (iter.hasNext()) {
            if (contains(iter.next())) return true;
        }
        return false;
    }

    public boolean containsAll(StmtIterator iter) throws RDFException {
        while (iter.hasNext()) {
            if (!contains(iter.next())) return false;
        }
        return true;
    }

    public boolean containsAny(Model model) throws RDFException {
        StmtIterator iter = model.listStatements();
        try {
            return containsAny(iter);
        } finally {
            iter.close();
        }
    }

    public boolean containsAll(Model model) throws RDFException {
        StmtIterator iter = model.listStatements();
        try {
            return containsAll(iter);
        } finally {
            iter.close();
        }
    }
    public ResIterator listSubjectsWithProperty(Property p, boolean o)
    throws RDFException {
        return listSubjectsWithProperty(p, new LiteralImpl(o));
    }

    public ResIterator listSubjectsWithProperty(Property p, long o)
    throws RDFException {
        return listSubjectsWithProperty(p, new LiteralImpl(o));
    }

    public ResIterator listSubjectsWithProperty(Property p, char o)
    throws RDFException {
        return listSubjectsWithProperty(p, new LiteralImpl(o));
    }

    public ResIterator listSubjectsWithProperty(Property p, float o)
    throws RDFException {
        return listSubjectsWithProperty(p, new LiteralImpl(o));
    }

    public ResIterator listSubjectsWithProperty(Property p, double o)
    throws RDFException {
        return listSubjectsWithProperty(p, new LiteralImpl(o));
    }

    public ResIterator listSubjectsWithProperty(Property p, String o)
    throws RDFException {
        return listSubjectsWithProperty(p, new LiteralImpl(o));
    }

    public ResIterator listSubjectsWithProperty(Property p, String o, String l)
    throws RDFException {
        return listSubjectsWithProperty(p, new LiteralImpl(o,l));
    }

    public ResIterator listSubjectsWithProperty(Property p, Object o)
    throws RDFException {
        if (o instanceof RDFNode) {
            return listSubjectsWithProperty(p, (RDFNode) o);
        } else {
            return listSubjectsWithProperty(p, new LiteralImpl(o));
        }
    }

    public Resource createResource(Resource type) throws RDFException {
        return createResource().addProperty(RDF.type, type);
    }

    public Resource createResource(String uri,Resource type)
    throws RDFException {
        return getResource(uri)
                   .addProperty(RDF.type, type);
    }

    public Resource createResource(ResourceF f) throws RDFException {
        return createResource(null, f);
    }

    public Resource createResource(String uri, ResourceF f) throws RDFException {
        try {
            return f.createResource(createResource(uri));
        } catch (Exception e) {
            throw new RDFException(RDFException.NESTEDEXCEPTION, e);
        }
    }

    public Literal createLiteral(boolean v) throws RDFException {
        return createLiteral(String.valueOf(v), "");
    }
    public Literal createLiteral(long v) throws RDFException {
        return createLiteral(String.valueOf(v), "");
    }
    public Literal createLiteral(char v) throws RDFException {
        return createLiteral(String.valueOf(v), "");
    }
    public Literal createLiteral(float v) throws RDFException {
        return createLiteral(String.valueOf(v), "");
    }
    public Literal createLiteral(double v) throws RDFException {
        return createLiteral(String.valueOf(v), "");
    }
    public Literal createLiteral(String v) throws RDFException {
        return createLiteral(v, "");
    }

    public Literal createLiteral(String v, String l) throws RDFException {
        return new LiteralImpl(v, l);
    }

    public Literal createLiteral(String v, String l, boolean wellFormed) {
        return new LiteralImpl(v, l, wellFormed);
    }

    public Literal createLiteral(Object v) throws RDFException {
        return createLiteral(v.toString(), "");
    }

    public Statement createStatement(Resource r, Property p, boolean o)
    throws RDFException {
        return createStatement(r, p, createLiteral(o));
    }
    public Statement createStatement(Resource r, Property p, long o)
    throws RDFException {
        return createStatement(r, p, createLiteral(o));
    }
    public Statement createStatement(Resource r, Property p, char o)
    throws RDFException {
        return createStatement(r, p, createLiteral(o));
    }
    public Statement createStatement(Resource r, Property p, float o)
    throws RDFException {
        return createStatement(r, p, createLiteral(o));
    }
    public Statement createStatement(Resource r, Property p, double o)
    throws RDFException {
        return createStatement(r, p, createLiteral(o));
    }
    public Statement createStatement(Resource r, Property p, String o)
    throws RDFException {
        return createStatement(r, p, createLiteral(o));
    }

    public Statement createStatement(Resource r, Property p, Object o)
    throws RDFException {
        if (o instanceof RDFNode) {
            return createStatement(r, p, (RDFNode) o);
        } else {
            return createStatement(r, p, createLiteral(o));
        }
    }
    public Statement createStatement(Resource r, Property p, String o,
                                     boolean wellFormed) throws RDFException {
        return createStatement(r, p, new LiteralImpl(o, wellFormed));
    }

    public Statement createStatement(Resource r, Property p, String o, String l)
      throws RDFException {
        return createStatement(r, p, new LiteralImpl(o,l));
    }
    public Statement createStatement(Resource r, Property p, String o, String l,
                                     boolean wellFormed) throws RDFException {
        return createStatement(r, p, new LiteralImpl(o,l,wellFormed));
    }
    public Bag createBag() throws RDFException {
        return createBag(null);
    }
    public Alt createAlt() throws RDFException {
        return createAlt(null);
    }
    public Seq createSeq() throws RDFException {
        return createSeq(null);
    }

    public Resource getResource(String uri) throws RDFException {
        return store.getResource(uri, this);
    }

    public Property getProperty(String uri) throws RDFException {
        return store.getProperty(uri, this);
    }

    public Property getProperty(String nameSpace,String localName)
      throws RDFException {
        return store.getProperty(nameSpace, localName, this);
    }

    public Bag getBag(String uri) throws RDFException {
      return store.getBag(uri, this);
    }
    public Bag getBag(Resource r) throws RDFException {
        return store.getBag(r, this);
    }

    public Alt getAlt(String uri) throws RDFException {
        return store.getAlt(uri, this);
    }

    public Alt getAlt(Resource r) throws RDFException {
        return store.getAlt(r, this);
    }

    public Seq getSeq(String uri) throws RDFException {
        return store.getSeq(uri, this);
    }
    public Seq getSeq(Resource r) throws RDFException {
        return store.getSeq(r, this);
    }

    public long size() throws RDFException {
        return store.size();
    }

    public ResIterator listSubjects() throws RDFException {
        return store.listSubjects();
    }

    public NsIterator listNameSpaces() throws RDFException {
        return store.listNameSpaces();
    }

    public StmtIterator listStatements() throws RDFException {
        return new StmtIteratorImpl(store.list(), null);
    }

    public StmtIterator listReifiedStatements() throws RDFException {
        Vector reifiedStatements = new Vector();
        Iterator iter = store.list();
        Object subject;
        while (iter.hasNext()) {
            subject = ((Statement) iter.next()).getSubject();
            if (subject instanceof Statement) {
                reifiedStatements.add(subject);
            }
        }
        if (iter instanceof ClosableIterator) {
            ((ClosableIterator) iter).close();
        }
        return new StmtIteratorImpl(reifiedStatements.iterator(),
                                    reifiedStatements);
    }

    public Model add(Statement s) throws RDFException {
        if (! (s instanceof StatementImpl)
          || (((StatementImpl)s).getModel() != this) ) {
            s = new StatementImpl(s.getSubject(),
                                  s.getPredicate(),
                                  s.getObject(),
                                  this);
        }
        store.add(s);
        return this;
    }

    public Model remove(Statement s) throws RDFException {
        store.remove(s);
        return this;
    }

    public boolean contains(Statement s) throws RDFException {
        return store.contains(s);
    }

    public boolean contains(Resource s, Property p) throws RDFException {
        return store.contains(s, p);
    }

    public boolean contains(Resource s, Property p, RDFNode o)
      throws RDFException {
       return contains(new StatementImpl(s, p, o));
    }

    public boolean isReified(Statement s) throws RDFException {
        StmtIterator iter = null;
        try {
            iter = listStatements(new SelectorImpl(s, null, (RDFNode) null));
            if (iter.hasNext()) {
                return true;
            }
            iter = listStatements(new SelectorImpl(null, null, s));
            return iter.hasNext();
        } finally {
            iter.close();
        }
    }

    public Statement getProperty(Resource s,Property p) throws RDFException {
        StmtIterator iter = null;
        try {
            iter = listStatements(new SelectorImpl(s, p, (RDFNode) null));
            if (iter.hasNext()) {
                return iter.next();
            } else {
                throw new RDFException(RDFException.PROPERTYNOTFOUND);
            }
        } finally {
            iter.close();
        }
    }

    public ResIterator listSubjectsWithProperty(Property p)
    throws RDFException {
        HashSet subjects = new HashSet();
        Iterator iter = store.listByPredicate(p);
        Statement stmt;
        try {
            while (iter.hasNext()) {
                stmt = (Statement)(iter.next());
                if (stmt.getPredicate().equals(p)) {
                    subjects.add(stmt.getSubject());
                }
            }
        } finally {
            if (iter instanceof ClosableIterator) {
                ((ClosableIterator) iter).close();
            }
        }
        return new ResIteratorImpl(subjects.iterator(), subjects);
    }

    public ResIterator listSubjectsWithProperty(Property p, RDFNode o)
    throws RDFException {
        HashSet subjects = new HashSet();
        Iterator iter = store.list(null, p, o);
        Statement stmt;
        try {
            while (iter.hasNext()) {
                stmt = (Statement)(iter.next());
// we need to check here because ModelMem does not return an exact match
// maybe this should be fixed @@FIX
                if (stmt.getPredicate().equals(p)
                  && stmt.getObject().equals(o)) {
                    subjects.add(stmt.getSubject());
                }
            }
        } finally {
            if (iter instanceof ClosableIterator) {
                ((ClosableIterator) iter).close();
            }
        }
        return new ResIteratorImpl(subjects.iterator(), subjects);
    }

    public NodeIterator listObjects() throws RDFException {
        HashSet objects = new HashSet();
        Iterator iter = store.list();
        Statement stmt = null;
        try {
            while (iter.hasNext()) {
                objects.add(((Statement) iter.next()).getObject());
            }
        } finally {
            if (iter instanceof ClosableIterator) {
                ((ClosableIterator) iter).close();
            }
        }
        return new NodeIteratorImpl(objects.iterator(), objects);
    }

    public NodeIterator listObjectsOfProperty(Property p) throws RDFException {
        HashSet objects = new HashSet();
        Iterator iter = store.listByPredicate(p);
        try {
            while (iter.hasNext()) {
                Statement stmt = (Statement) iter.next();
                objects.add(stmt.getObject());
            }
        } finally {
            if (iter instanceof ClosableIterator) {
                ((ClosableIterator) iter).close();
            }
        }
        return new NodeIteratorImpl(objects.iterator(), objects);
    }

    public NodeIterator listObjectsOfProperty(Resource s, Property p)
      throws RDFException {
        HashSet objects = new HashSet();
        Iterator iter = store.list(s, p, null);
        try {
            Statement stmt = null;
            while (iter.hasNext()) {
                stmt = (Statement) iter.next();
                if (stmt.getPredicate().equals(p))
                {
                    objects.add(stmt.getObject());
                }
            }
        } finally {
            if (iter instanceof ClosableIterator) {
                ((ClosableIterator) iter).close();
            }
        }
        return new NodeIteratorImpl(objects.iterator(), objects);
    }

    public StmtIterator listStatements(Selector selector)
      throws RDFException {
        Iterator iter;

        if (selector instanceof SelectorImpl) {
            SelectorImpl s = (SelectorImpl) selector;
            iter = store.list(s.getSubject(),
                              s.getPredicate(),
                              s.getObject());
        } else {
            iter = store.list();
        }

        if (iter == null) {
            return new StmtIteratorImpl((new HashSet()).iterator(), null);
        }
        return listStatements(selector, iter);
    }

    protected StmtIterator listStatements(Selector s, Iterator iter)
      throws RDFException {
        Vector matching = new Vector();
        Statement stmt;
        while (iter.hasNext()) {
            stmt = (Statement) iter.next();
            try {
                if (s.test(stmt)) {
                    matching.add(stmt);
                }
            } catch (Exception e) {
                if (iter instanceof ClosableIterator) {
                    ((ClosableIterator)iter).close();
                }
                throw new RDFException(RDFException.SELECTOREXCEPTION, e);
            }
        }
        return new StmtIteratorImpl(matching.iterator(), matching);
    }

    public Model begin() throws RDFException {
        throw new RDFException(RDFException.UNSUPPORTEDOPERATION);
    }

    public Model abort() throws RDFException {
        throw new RDFException(RDFException.UNSUPPORTEDOPERATION);
    }

    public Model commit() throws RDFException {
        throw new RDFException(RDFException.UNSUPPORTEDOPERATION);
    }

    public boolean independent() {
        return true;
    }
    public Resource createResource() throws RDFException {
        return store.createResource(this);
    }

    public Resource createResource(String uri) throws RDFException {
        return store.createResource(uri, this);
    }

    public Property createProperty(String uri) throws RDFException {
        return store.createProperty(uri, this);
    }
    public Property createProperty(String nameSpace, String localName)
    throws RDFException {
        return store.createProperty(nameSpace, localName, this);
    }

    public Statement createStatement(Resource r, Property p, RDFNode o)
    throws RDFException {
        return new StatementImpl(r, p, o, this);
    }

    public Bag createBag(String uri) throws RDFException {
        Bag bag = store.createBag(uri, this);
        bag.addProperty(RDF.type, RDF.Bag);
        return bag;
    }
    public Alt createAlt(String uri) throws RDFException {
        Alt alt = store.createAlt(uri, this);
        alt.addProperty(RDF.type, RDF.Alt);
        return alt;
    }
    public Seq createSeq(String uri) throws RDFException {
        Seq seq = store.createSeq(uri, this);
        seq.addProperty(RDF.type, RDF.Seq);
        return seq;
    }

    public NodeIterator listContainerMembers(Container cont,
                                             NodeIteratorFactory f)
                                                  throws RDFException {
        Iterator iter = store.listBySubject(cont);
        Statement    stmt;
        String       rdfURI = RDF.getURI();
        Vector       result = new Vector();
        int          maxOrdinal = 0;
        int          ordinal;
        while (iter.hasNext()) {
            stmt = (Statement) iter.next();
            ordinal = stmt.getPredicate().getOrdinal();
            if (stmt.getSubject().equals(cont) && ordinal != 0) {
                if (ordinal > maxOrdinal) {
                    maxOrdinal = ordinal;
                    result.setSize(ordinal);
                }
                result.setElementAt(stmt, ordinal-1);
            }
        }
        if (iter instanceof ClosableIterator) {
            ((ClosableIterator) iter).close();
        }
        try {
             return f.createIterator(result.iterator(), result, cont);
        } catch (Exception e) {
            throw new RDFException(e);
        }
    }

    public int containerSize(Container cont) throws RDFException {
        int result = 0;
        Iterator iter = store.listBySubject(cont);
        Property     predicate;
        Statement    stmt;
        String       rdfURI = RDF.getURI();
        while (iter.hasNext()) {
            stmt = (Statement) iter.next();
            predicate = stmt.getPredicate();
            if (stmt.getSubject().equals(cont)
             && predicate.getOrdinal() != 0
               ) {
                result++;
            }
        }
        if (iter instanceof ClosableIterator) {
            ((ClosableIterator) iter).close();
        }
        return result;
    }

    public int containerIndexOf(Container cont, RDFNode n) throws RDFException {
        int result = 0;
        Iterator iter = store.listBySubject(cont);
        Property     predicate;
        Statement    stmt;
        String       rdfURI = RDF.getURI();
        while (iter.hasNext()) {
            stmt = (Statement) iter.next();
            predicate = stmt.getPredicate();
            if (stmt.getSubject().equals(cont)
             && predicate.getOrdinal() != 0
             && n.equals(stmt.getObject())
              ) {
                result = predicate.getOrdinal();
                break;
            }
        }
        if (iter instanceof ClosableIterator) {
            ((ClosableIterator) iter).close();
        }
        return result;
    }

   public boolean containerContains(Container cont, RDFNode n)
    throws RDFException {
        return containerIndexOf(cont, n) != 0;
    }

    public Resource convert(Resource r) throws RDFException {
        if (r.getModel() == this) {
            return r;
        } else {
            return ((ResourceI)r).port(this);
        }
    }

    public Property convert(Property p) throws RDFException {
        if (p.getModel() == this) {
            return p;
        } else {
            return (Property) ((ResourceI)p).port(this);
        }
    }

    public RDFNode convert(RDFNode n) throws RDFException {
        if (n instanceof Property) {
            return convert((Property) n);
        } else if (n instanceof Resource) {
            return convert((Resource) n);
        } else {
            return n;
        }
    }

    public boolean equals(Object model ) {
        try {
            if (model==null || ! (model instanceof Model)) {
                return false;
            }
            return //match( (Model) model);
            ModelMatcher.equals(this, (Model) model);
        } catch (RDFException e) {
            throw new RDFError(e);
        }
    }

    public void close() {
        store.close();
    }
}