/*
 *  (c) Copyright Hewlett-Packard Company 2001, 2002
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: NTripleReader.java,v 1.8 2002/06/24 07:53:34 jeremy_carroll Exp $
 */

package com.hp.hpl.mesa.rdf.jena.common;



import com.hp.hpl.mesa.rdf.jena.model.*;
import com.hp.hpl.mesa.rdf.jena.vocabulary.RDF;

import java.net.URL;
import java.io.InputStreamReader;
import java.io.BufferedReader;
import java.io.Reader;
import java.io.IOException;
import java.util.Hashtable;


/** N-Triple Reader
 *
 * @author  Brian McBride, Jeremy Carroll, Dave Banks
 * @version  Release=$Name:  $ Date=$Date: 2002/06/24 07:53:34 $
 */
public class NTripleReader extends Object implements RDFReader {  
        
    Model model = null;
    Hashtable anons = new Hashtable();

    IStream        in = null;
    boolean        inErr = false;
    int            errCount = 0;
    private static final int sbLength = 200;
    
    RDFErrorHandler errorHandler = new RDFDefaultErrorHandler();

    NTripleReader() {} 

    public void read(Model model, Reader reader, String base)
      throws RDFException {
          
        if (! (reader instanceof BufferedReader)) {
            reader = new BufferedReader(reader);
        }
        
        this.model = model;
        in = new IStream(reader);
        readRDF();
        if (errCount != 0) {
            throw new RDFException(RDFException.SYNTAXERROR);
        }
    }
    
    public void read(Model model, String url) throws RDFException {
        try {
            read(model, 
                 new InputStreamReader(((new URL(url))).openStream()), url);
        } catch (Exception e) {
            throw new RDFException(e);
        } finally {
            if (errCount != 0) {                
                throw new RDFException(RDFException.SYNTAXERROR);
            }
        }
    }
    
    public Object setProperty(String propName,Object propValue)
      throws RDFException {
        errorHandler.error(new RDFException(RDFException.UNKNOWNPROPERTY));
        return null;
    }

    public RDFErrorHandler setErrorHandler(RDFErrorHandler errHandler) {
        RDFErrorHandler old = this.errorHandler;
        this.errorHandler = errHandler;
        return old;
    }    
    
    protected void readRDF() throws RDFException {
        Resource  subject;
        Property  predicate = null;
        RDFNode   object;
        
        while (!in.eof()) {
            while (!in.eof()) {
                inErr = false;
                
                skipWhiteSpace();
                if (in.eof()) {
                    return;
                }
                
                subject = readResource();
                if (inErr) break;
                
                skipWhiteSpace();
                try {
                    predicate = model.createProperty(readResource().getURI());
                } catch (Exception e1) {
                    errorHandler.fatalError(e1);
                }
                if (inErr) break;
                
                skipWhiteSpace();
                object = readNode();
                if (inErr) break;
                
                skipWhiteSpace();
                if (badEOF())
                      break;
                
                if (!expect("."))
                    break;
                    
                try {
                    model.add(subject, predicate, object);
                } catch (Exception e2) {
                    errorHandler.fatalError(e2);
                }              
            }
            if (inErr) {
                errCount++;
                while (! in.eof() && in.readChar() != '\n') {}
            }
        }                
    }
    
    public Resource readResource() throws RDFException {
        char inChar = in.readChar();
        if (badEOF()) 
            return null;
            
        if (inChar == '_') { // anon resource
            if (!expect(":"))
              return null;
            String name = readName();
            if (name == null) {
                errorHandler.error(
                  new RDFException(RDFException.SYNTAXERROR,
                                       syntaxErrorMessage(
                                             "expected bNode label",
                                             in.getLinepos(),
                                             in.getCharpos())));
                inErr = true;
                return null;
            }
            return lookupResource(name);
        } else if (inChar == '<') { // uri
            String uri = readURI();
            if (uri == null) {
                inErr = true;
                return null;
            }
            inChar = in.readChar();
            if (inChar != '>') {
                errorHandler.error(
                      new RDFException(RDFException.SYNTAXERROR,
                                           syntaxErrorMessage(
                                                 "expected '>'",
                                                 in.getLinepos(),
                                                 in.getCharpos())));
                inErr = true;
                return null;
            }
            return model.createResource(uri);
        } else {
                errorHandler.error(
                      new RDFException(RDFException.SYNTAXERROR,
                                           syntaxErrorMessage(
                                                 "unexpected input",
                                                 in.getLinepos(),
                                                 in.getCharpos())));
            inErr = true;
            return null;
        }
    }
    
    public RDFNode readNode() throws RDFException {
        skipWhiteSpace();
        switch ( in.nextChar() ) {
        	case '"':
            return readLiteral(false);
        	case 'x':
            return readLiteral(true);
            case '<':
            case '_':
            return readResource();
            default:
             errorHandler.error(
                      new RDFException(RDFException.SYNTAXERROR,
                                           syntaxErrorMessage(
                                                 "unexpected input",
                                                 in.getLinepos(),
                                                 in.getCharpos())));
             inErr = true;
             return null;
        }
    }
    
   protected Literal readLiteral(boolean wellFormed) throws RDFException {
        
        StringBuffer lit = new StringBuffer(sbLength);

        
        if ( wellFormed ) {
        	if(!expect("xml"))
        	  return null;
        } 
        
        if(!expect("\""))
          return null;
        
        while (true) {
            char inChar = in.readChar();
            if (badEOF()) 
                return null;
            if (inChar == '\\') {
                char c = in.readChar();
                if (in.eof()) {
                    inErr = true;
                    return null;
                }
                if (c == 'n') { 
                    inChar = '\n';
                } else if (c == 'r') {
                    inChar = '\r';
                } else if (c == 't') {
                    inChar = '\t';
                } else if (c == '\\' || c == '"') {
                    inChar = c;
                } else if ( c == 'u' ) {
                	inChar = readUnicode4Escape();
                	if ( inErr )
                	  return null;
                } else {
                    errorHandler.error(
                      new RDFException(RDFException.SYNTAXERROR,
                                           syntaxErrorMessage(
                                         "illegal escape sequence '" + c + "'",
                                                 in.getLinepos(),
                                                 in.getCharpos())));
                     inErr = true;
                     return null;                           
                }                    
            } else if (inChar == '"') {
            	String lang;
            	if ( '-' == in.nextChar() ) {
            		lang = readLang();
            	} else {
            		lang = "";
            	}
                return model.createLiteral(lit.toString(),lang,wellFormed);
            }
            lit = lit.append(inChar);
        }
    }
    
    private char readUnicode4Escape() {
    	char buf[] = new char[]{
    		in.readChar(),
    		in.readChar(),
    		in.readChar(),
    		in.readChar()
    	};
    	if ( badEOF()  ) {
             return 0;
    	}
    	try {
    		return (char)Integer.parseInt(new String(buf),16);
    	}
    	catch (NumberFormatException e) {
    		inErr = true;
            errorHandler.error(
                      new RDFException(RDFException.SYNTAXERROR,
                                           syntaxErrorMessage(
                                                 "bad unicode escape sequence",
                                                 in.getLinepos(),
                                                 in.getCharpos())));
            return 0;   
    	}
    }
    private String readLang()  {
        StringBuffer lang = new StringBuffer(15);
    
        if(!expect("-"))
          return null;
        
        while (true) {
            char inChar = in.nextChar();
            if ( Character.isWhitespace(inChar)
               || inChar == '.' ) 
               	return lang.toString();
            lang = lang.append(in.readChar());
        }
    }
    private boolean badEOF() {
    	if ( in.eof() ) {
    		inErr = true;
            errorHandler.error(
                      new RDFException(RDFException.SYNTAXERROR,
                                           syntaxErrorMessage(
                                                 "premature end of file",
                                                 in.getLinepos(),
                                                 in.getCharpos())));
    	}
    	return inErr;
    }
    protected String readURI() {
        StringBuffer uri = new StringBuffer(sbLength);
        
        while (in.nextChar() != '>') {
            char inChar = in.readChar();
            
            if (inChar == '\\') {
            	expect("u");
            	inChar = readUnicode4Escape();
            }
            if (badEOF()) {
                return null;
            }
            uri = uri.append(inChar);
        }
        return uri.toString();
    }
    
    protected String readName() {
        StringBuffer name = new StringBuffer(sbLength);
        
        while (! Character.isWhitespace(in.nextChar())) {
            name = name.append(in.readChar());
            if (badEOF()) 
                return null;
        }
        return name.toString();
    }
    private boolean expect(String str) {
        for (int i=0;i<str.length();i++) {
        	char want = str.charAt(i);
        	
            if (badEOF()) 
                return false;
                
        	char inChar = in.readChar();
                
            if (inChar != want) { 
            	System.err.println("N-triple reader error");
                errorHandler.error(
                  new RDFException(RDFException.SYNTAXERROR,
                                       syntaxErrorMessage(
                                             "expected \""+str+"\"",
                                             in.getLinepos(),
                                             in.getCharpos())));
                inErr = true;
                return false;
            }
        }
        return true;
    }
    protected void skipWhiteSpace() {
        while (Character.isWhitespace(in.nextChar()) || in.nextChar() == '#') {
            char inChar = in.readChar();
            if (in.eof()) {
                return;
            }
            if (inChar == '#') {
                while (inChar != '\n') {
                    inChar = in.readChar();
                    if (in.eof()) {
                        return;
                    }
                }
            }            
        }
    }
    
    protected Resource lookupResource(String name) throws RDFException{
        Resource r;
        r = (Resource) anons.get(name);
        if (r == null) {
            r = model.createResource();
            anons.put(name, r);
        }
        return r;
    } 
    
    protected String syntaxErrorMessage(String msg, int linepos, int charpos) {
        return "syntax error at line " + linepos +
                              " position " + charpos +
                              ": " + msg;
    }
}

class IStream {
    
    // simple input stream handler

    Reader in;
    char[] thisChar = new char[1];
    boolean eof;
    int charpos = 1;
    int linepos = 1;
    
    protected IStream(Reader in) {
        try {
            this.in = in;
            eof = (in.read(thisChar, 0, 1) == -1);
        } catch (IOException e) {
            throw new RDFError(e);
        }
    }
    
    protected char readChar() {
        try {
            if (eof) return '\000';
            char rv = thisChar[0];
            eof = (in.read(thisChar, 0, 1) == -1);
            if (rv == '\n') {
                linepos++;
                charpos = 0;
            } else {
                charpos++;
            }
            return rv;
        } catch (java.io.IOException e) {
            throw new RDFError(e);
        }
    }
    
    protected char nextChar() {
        if (eof) {
            return '\000';
        } else {
            return thisChar[0];
        }
    }
    
    protected boolean eof() {
        return eof;
    }
    
    protected int getLinepos() {
        return linepos;
    }
    
    protected int getCharpos() {
        return charpos;
    }
}