/*
 *  (c) Copyright Hewlett-Packard Company 2000 
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Store.java
 *
 * Created on 11 January 2001, 10:09
 */

package com.hp.hpl.mesa.rdf.jena.common;

import com.hp.hpl.mesa.rdf.jena.model.*;
import java.util.Iterator;

/** Interface to a triple store.
 *
 * @author  bwm
 * @version Release='$Name:  $' Revision='$Revision: 1.3 $' Date='$Date: 2001/08/25 12:27:01 $'
 */
public interface Store {
    
    public Resource createResource(Model m) throws RDFException;
    
    public Resource createResource(String uri, Model m) throws RDFException;
    
    public Property createProperty(String uri, Model m) throws RDFException;
    
    public Property createProperty(String namespace, String localPart, Model m)
       throws RDFException;
    
    public Statement createStatement(Resource s, Property p, RDFNode o, Model m)
       throws RDFException;
    
    public Bag createBag(Model m) throws RDFException;
    
    public Bag createBag(String uri, Model m) throws RDFException;
    
    public Alt createAlt(Model m) throws RDFException;
    
    public Alt createAlt(String uri, Model m) throws RDFException;
    
    public Seq createSeq(Model m) throws RDFException;
    
    public Seq createSeq(String uri, Model m) throws RDFException;
    
    public Resource getResource(String uri, Model m) throws RDFException;
    
    public Property getProperty(String uri, Model m) throws RDFException;
    
    public Property getProperty(String namespace, String localPart, Model m)
       throws RDFException;
    
    public Property getProperty(Property p) throws RDFException;
    
    public Statement getStatement(Resource s, Property p, RDFNode o, Model m)
       throws RDFException;
    
    public Bag getBag(String uri, Model m) throws RDFException;
    
    public Bag getBag(Resource r, Model m) throws RDFException;
    
    public Alt getAlt(String uri, Model m) throws RDFException;
    
    public Alt getAlt(Resource r, Model m) throws RDFException;
    
    public Seq getSeq(String uri, Model m) throws RDFException;
    
    public Seq getSeq(Resource r, Model m) throws RDFException;

/** Add a statement to the store
 * @param s the statement to add
 */    
   public void add(Statement s) throws RDFException;
    
/** remove a statement from the store
 * @param s the statement to remove
 */   
    public void remove(Statement s) throws RDFException;
    
/** the number of statements in the store
 * @return the number of statements in the store
 */    
    public int size() throws RDFException;
    
/** test whether the store contains a statement
 * @return true if the statement is in the store, false otherwise
 * @param s the statement to test for
 */    
    public boolean contains(Statement s) throws RDFException;
    
/** test whether the store contains a statement with a given subject
 *  and predicate
 * @return true if the statement is in the store, false otherwise
 * @param s the subject of the statement to test for
 * @param p the predicate of the statement to test for
 */    
    public boolean contains(Resource s, Property p) throws RDFException;
    
/** return an iterator on all the statements in the store
 * @return in iterator over all the statements in the store
 */    
    public Iterator list() throws RDFException;

/** return an iterator over all the statements which match the given
 *  subject predicate and object.
 *
 *<p>The store is searched for all statements which match the given
 *   subject predicate and object.  A null arguement matches anything.
 *   </p>
 *
 * @param subject the subject of the statements to be returned or null
 * @param predicate the predicate of the statments to be returned or null
 * @param object the object of the statements to be returned or null
 * @return an iterator over all statements in the store which match the
 * given subject predicate and object
 */   
    public Iterator list(Resource subject, Property predicate, RDFNode object)
      throws RDFException;
    
/** return an iterator over all the statements with a given subject
 * @param subject the subject of the statements to be returned
 * @return an iterator over all statements in the store with the
 * given subject
 */    
    public Iterator listBySubject(Resource subject) throws RDFException;
    
/** list all the statements in the store with a given predicate
 * @param property the predicate to search for
 * @return an iterator over all statements in the store with
 * the given predicate
 */    
    public Iterator listByPredicate(Property property) throws RDFException;
    
/** list all the statements in the store with a given object
 * @param object the object to search for
 * @return an iterator over all statements in the store with
 * the given object
 */    
    public Iterator listByObject(RDFNode object) throws RDFException;
    
/** list all the subjects of statements in the store
 *
 *  <P>Each subject is listed only once.</P>
 * @return an iterator over a set of the subjects of all statements in the store
 */ 
    public ResIterator listSubjects() throws RDFException; 
    
/** list namespaces of the predicates of all the statements in ths store.
 *
 *  <P>Each namespace is listed only once.</P>
 * @return an iterator over a set of the subjects of all statements in the store
 */        
    public NsIterator listNameSpaces() throws RDFException; 
    
/** Close the store and free up resources held.
 *
 *  <p>Not all implementations of Store require this method to be called.  But
 *     some do, so in general its best to call it when done with the object,
 *     rather than leave it to the finalizer.</p>
 */        
    public void close();
}
