/*
 *  (c) Copyright Hewlett-Packard Company 2001 
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: PrettyWriter.java,v 1.6 2002/08/03 05:43:03 jeremy_carroll Exp $
 *
 * PrettyWriter.java
 *
 * Created on June 28, 2001, 4:58 AM
 */

package com.hp.hpl.mesa.rdf.jena.common.prettywriter;

import com.hp.hpl.mesa.rdf.jena.model.*;
import com.hp.hpl.mesa.rdf.jena.common.*;

import com.hp.hpl.mesa.rdf.jena.vocabulary.RDFS;

import org.apache.xerces.readers.MIME2Java;

import java.util.Map;
import java.util.HashMap;

import java.io.*;
//Writer;
//import java.io.PrintWriter;

/** Writes out RDF in the abbreviated syntax,  for human consumption 
   not only machine readable.
 * It is not normal to call the constructor directly, but to use
 * the method RDFWriterF.getWriter("RDF/XML-ABBREV").
 * Does not support the <code>NSPREFIXPROPBASE</code> system properties.
 * Use <code>setNsPrefix</code>.
 * For best results it is necessary to set the property 
   <code>"prettyTypes"</code>. See setProperty for information.
   @see com.hp.hpl.mesa.rdf.jena.model.RDFWriterF#getWriter
 * @author jjc
 * @version  Release='$Name:  $' Revision='$Revision: 1.6 $' Date='$Date: 2002/08/03 05:43:03 $'
 */
public class PrettyWriter implements RDFWriter, RDFErrorHandler {
    
    
    /** Set an error handler.
     * @param errHandler The new error handler to be used.
     * @return the old error handler
     */
    public RDFErrorHandler setErrorHandler(RDFErrorHandler errHandler) {
        // null means no user defined error handler.
        // We implement this using defaultErrorHandler,
        // but hide this fact from the user.
        RDFErrorHandler rslt = errorHandler;
        if ( rslt == defaultErrorHandler )
            rslt = null;
        errorHandler = errHandler == null
        ? defaultErrorHandler
        : errHandler;
        return rslt;
    }
    
    /** Set the the namespace prefix to be used for a particular namespace.
     * @param prefix the prefix to be used
     * @param ns the namespace
     */
    public void setNsPrefix(String prefix,String ns) {
        nameSpacePrefices.put(prefix,ns);
    }
    
    /** Sets properties on this writer.
     *  Current properties are:
     * <dl>
     * <dt>xmlbase
     * <dd>Allows the specification of the value for xml:base in the
     * file, as a string.
     * <dt>showXmlDeclaration
     * <dd>can be true, false or "default" (null)
     *  If true, an XML Declaration is included in the output, if false
     *  no XML declaration is included.
     *  The default behaviour only gives an XML Declaration when
     *  asked to write to an OutputStreamWriter that uses some
     *  encoding other than UTF-8. In this case the encoding is shown
     *  in the XML declaration.
     * <dt>prettyTypes</dt>
     * <dd>
     * the types of the principal objects in the model.  PrettyWriter
     *  will tend to create RDF/XML with resources of these types at the
     *  top level.
     * <br>
     * Example usage showing the default value:
     <pre>
    * prettyWriter.setProperty("prettyTypes",new Resource[]{
    *               DAML.Ontology,
    *               DAML.Class,
    *               DAML.Datatype,
    *               DAML.Property,
    *               DAML.ObjectProperty,
    *               DAML.DatatypeProperty,
    *               DAML.TransitiveProperty,
    *               DAML.UnambigousProperty,
    *               DAML.UniqueProperty,
    *               });
     </pre>
     </dd>
     * @param propName Must be one of  "xmlbase", "showXmlDeclaration", "prettyTypes"
     * @param propValue Appropriate value for the property. i.e. For
     *                   <dd>
     *                   <dt>xmlbase</dt>
     *                   <dd>A string, representing a URI.</dd>
     *                   <dt>showXmlDeclaration</dt>
     *                   <dd>A Boolean, null, or the strings "true", "false", or "default"
     *                   <dt>prettyTypes
     *                   <dd>
     *  An array of Resource's being types of objects to show
     *                  at the top level.
     *                   </dl>
     * @return the old value for this property, or <code>null</code>
     * if no value was set.
     */
    public Object setProperty(String propName,Object propValue) throws RDFException {
        if ( propName.equalsIgnoreCase("showXmlDeclaration") ) {
        	String oldValue;
        	if (showXmlDeclaration == null) 
        	  oldValue = null;
        	else 
        	   oldValue = showXmlDeclaration.toString();
        	if ( propValue == null )
        	   showXmlDeclaration = null;
        	else if ( propValue instanceof Boolean )
        	   showXmlDeclaration = (Boolean)propValue;
        	else if ( propValue instanceof String ) {
        		String propValueStr = (String)propValue;
        		if ( propValueStr.equalsIgnoreCase("default") ) {
        			showXmlDeclaration = null;
        		} 
        		if ( propValueStr.equalsIgnoreCase("true") )
        		   showXmlDeclaration = Boolean.TRUE;
        		else if ( propValueStr.equalsIgnoreCase("false") )
        		   showXmlDeclaration = Boolean.FALSE;
        		else
        		  // Also overloading the error condition.
        		  throw new RDFException(RDFException.INVALIDBOOLEANFORMAT);
        	}
        	 return oldValue;
        } else
        if ( propName.equalsIgnoreCase("xmlbase") ) {
            String result = xmlBase;
            xmlBase = (String) propValue;
            return result;
        } else
        if ( propName.equalsIgnoreCase("prettyTypes") ) {
            Object rslt = types;
            types = (Resource[])propValue;
            return rslt;
        } else {
            // Copied from Brian! This is overloading the error condition!
            
            throw new RDFException(RDFException.UNKNOWNPROPERTY);
        }
    }
    
    /** Serialize Model <code>model</code> to Writer <code>out</out>.
     * @param out The Writer to which the serialization should
     * be sent.
     * @param model The model to be written.
     * @param base the base URI for relative URI calculations.  <code>
     * null</code> means use only absolute URI's.
     * @throws RDFException Generic RDF exception.
     */
    public void write(Model model,Writer out,String base) throws RDFException {
        PrintWriter pw = new PrintWriter(out);
        Unparser unp = new Unparser(this,base,model,pw);
        
        unp.setTopLevelTypes(types);
        unp.useNameSpaceDecl(nameSpacePrefices);
        if (xmlBase!=null)
           unp.setXMLBase(xmlBase);
        if ( !Boolean.FALSE.equals(showXmlDeclaration) ) {
            String decl = null;
        	if ( out instanceof OutputStreamWriter ) {
        		String javaEnc = ((OutputStreamWriter)out).getEncoding();
        		if (!javaEnc.equals("UTF8")) {
        			decl = "<?xml version='1.0' encoding='"+
        			          MIME2Java.reverse(javaEnc) +"'?>";
        		}
        	}
        	if ( decl == null && showXmlDeclaration != null )
        			decl = "<?xml version='1.0'?>";
            if (decl != null )
              unp.setDeclaration( decl );
        		
        }
        unp.write();
        pw.flush();
    }
    
    public void write(Model model,OutputStream out,String base) throws RDFException {
    	try {
    	  write(model,new OutputStreamWriter(out,"UTF-8"),base);
    	}
    	catch (UnsupportedEncodingException ee) {
    		throw new Error("Java installation broken: does not support UTF-8 encoding.");
    	}
    }
    
    
    // Implemenatation of RDFErrorHandler
    public void error(Exception e) {
        errorHandler.error(e);
    }
    
    public void warning(Exception e) {
        errorHandler.warning(e);
    }
    
    public void fatalError(Exception e) {
        errorHandler.fatalError(e);
    }
   
    static public void main(String args[]) throws Exception {
        System.out.println("Test code for bug 77");
        Model m = new com.hp.hpl.mesa.rdf.jena.mem.ModelMem();
        m.read(new FileReader("modules/rdf/regression/arp/bug51_0.rdf"),"http://example.org/file");
        RDFWriter pw = m.getWriter("RDF/XML-ABBREV");
        pw.setNsPrefix("eg","http://example.org/");
        pw.setNsPrefix("eg2","http://example.org/foo#");
        pw.write(m,new OutputStreamWriter(System.out),"http://example.org/file");
    }
    
    static private RDFDefaultErrorHandler defaultErrorHandler = new RDFDefaultErrorHandler();
    private RDFErrorHandler errorHandler = defaultErrorHandler;
    
    private Map nameSpacePrefices = new HashMap();
    private String xmlBase;
    private Boolean showXmlDeclaration = null;
    
    private Resource types[] = new Resource[]{
        DAML.Ontology,
        DAML.Datatype,
        DAML.Class,
        RDFS.Class,
        DAML.Property,
        DAML.ObjectProperty,
        DAML.DatatypeProperty,
        DAML.TransitiveProperty,
        DAML.UnambigousProperty,
        DAML.UniqueProperty,
    };
}
