/*
 *  (c) Copyright Hewlett-Packard Company 2000, 2001 
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * ModelMem.java
 *
 * Created on 03 August 2000, 15:02
 */

package com.hp.hpl.mesa.rdf.jena.mem;

import com.hp.hpl.mesa.rdf.jena.common.*;
import com.hp.hpl.mesa.rdf.jena.model.*;
import com.hp.hpl.mesa.rdf.jena.vocabulary.RDF;

import java.io.FileReader;
import java.io.Reader;
import java.io.Writer;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Map;
import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.Vector;


/** A main memory implemention of the RDF API.
 *
 * <p>This implementation of the RDF API stores its all its data in main memory.
 *   It supports all methods except the transaction calls.</p>
 * <p>This is a prototype implementation, and should be considered to be alpha
 *   quality code.  It has not yet been performance optimised and is known to be
 *   wasteful of memory.  It has been tested by loading the wordnet database
 *   (see Sample3) which results in a model containing over 600,000 statements.
 *   </p>
 * @author bwm
 * @version Release='$Name:  $' Revision='$Revision: 1.7 $' Date='$Date: 2001/09/23 13:51:18 $'
 */
public class ModelMem extends ModelCom implements Model, ModelI {    
    
    // next free error code = 3

    /** Create an empty model.
     */
    public ModelMem() {
        store = new StoreMem();
    }
    public boolean supportsTransactions() {return false;}
    
    public boolean supportsSetOperations() {return true;}
    
    
    
    public Model query(Selector selector) throws RDFException {
        ModelMem model = new ModelMem();
        StmtIterator iter = null;
        try {
            iter = listStatements(selector);
            while (iter.hasNext()) {
                model.add(iter.next());
            }
            return model;
        } finally {
            if (iter != null) {
                iter.close();
            }
        }
    }
    
    public Model union(Model model) throws RDFException {
        return (new ModelMem()).add(this)
                               .add(model);
    }
    
    public Model intersection(Model model) throws RDFException {
        Model largerModel = this;
        Model smallerModel  = model;
        ModelMem resultModel = new ModelMem();
        StmtIterator iter = null;
        Statement stmt;
        
        if (model.size() > this.size()) {
            largerModel = model;
            smallerModel = this;
        }
        try {
            iter = smallerModel.listStatements();
            while (iter.hasNext()) {
                stmt = iter.next();
                if (largerModel.contains(stmt)) {
                    resultModel.add(stmt);
                }
            }
            return resultModel;
        } finally {
            iter.close();
        }
    }
    
    public Model difference(Model model) throws RDFException {
        ModelMem resultModel = new ModelMem();
        StmtIterator iter = null;
        Statement stmt;
        try {
            iter = listStatements();
            while (iter.hasNext()) {
                stmt = iter.next();
                if (! model.contains(stmt)) {
                    resultModel.add(stmt);
                }
            }
            return resultModel;
        } finally {
            iter.close();
        }
    }    
    
/* The following is some simple graph matching code
 * This is the original code which has been removed because it was
 * too slow.  However, it is retained here, unmaintained lest it
 * prove useful for future debugging.
 *    
    protected boolean match(Model model) throws RDFException {
        
        if( size() != model.size()) {
            return false;
        }
        BindingTable bindings = new BindingTable();

	// Prepare
	StmtIterator iter = listStatements();
	HashSet stmtsWithAnon = new HashSet();
        while( iter.hasNext()) {
            Statement stmt = iter.next();
            boolean hasAnon = false;
	    
            Resource r = stmt.getSubject();
            if (r.isAnon()) {
                bindings.add(r);
                hasAnon = true;
            }
            
            r = stmt.getPredicate();
            if (r.isAnon()) {
                bindings.add(r);
                hasAnon = true;
            }
            
            RDFNode n = stmt.getObject();
            if (n instanceof Resource && ((Resource)n).isAnon()) {
                bindings.add((Resource) n);
                hasAnon = true;
            }
            
            if (hasAnon) {
                stmtsWithAnon.add(stmt);
            } else if (! model.contains(stmt)) {
                return false;
            }
        }
        
        // by this point we have tested that all statements which
        // do not contain an anonymous resource are in the other model
        // and we have built up a binding table
        
        // now call the recursive function to find a match
        
        return matchAnons(stmtsWithAnon, bindings, model);
    }
    
    protected boolean matchAnons(HashSet statements, BindingTable bindings, 
                                   Model model)
      throws RDFException {
        if( statements.isEmpty()) {
	    return true;               // none left - all done
        }
    

	// Recursive case there exist a Selector that has not
	// been bound. Bind it then Bind the rest recursively
        
	Statement stmt1 =  (Statement) statements.iterator().next();
        Resource  subject   = stmt1.getSubject();
        Property  predicate = stmt1.getPredicate();
        RDFNode   object    = stmt1.getObject();
        
        // now compute the query
        // for each component of triple
        //  if it is an anon resource if it has a binding query for that
        //                               otherwise query for null
        //           otherwise query for that node
        
        Resource qs = subject.isAnon() 
                        ? (Resource) bindings.getValue(subject)
                        : subject;
        Property qp = predicate.isAnon() 
                        ? (Property) bindings.getValue(predicate)
                        : predicate;
        RDFNode qo = (  object instanceof Resource) 
                      && ((Resource)object).isAnon()
                        ? (Resource) bindings.getValue(object)
                        : object;
        
	StmtIterator possibles 
              = model.listStatements( new AnonSelector(qs, qp, qo));
        statements.remove(stmt1);
	
	while(possibles.hasNext()) {
	    Statement stmt2 = possibles.next();
            if (subject.isAnon()) {
                bindings.bind(subject, stmt2.getSubject());
            }
            if (predicate.isAnon()) {
                bindings.bind(predicate, stmt2.getPredicate());
            }
            if (object instanceof Resource && ((Resource)object).isAnon()) {
                bindings.bind(object, stmt2.getObject());
            }
            
	    if(matchAnons(statements, bindings, model))
		return true;
            
            if (subject.isAnon()) {
                bindings.unbind(subject);
            }
            if (predicate.isAnon()) {
                bindings.unbind(predicate);
            }
            if (object instanceof Resource && ((Resource)object).isAnon()) {
                bindings.unbind(object);
            }            
	}

//	bindings.unBind(s);      // in the wrong place in the loop
	statements.add(stmt1);
	
	return false;
    }
    
   
}

class BindingTable
{
    HashMap bindings = new HashMap();

    protected void add(RDFNode r )
    {
        if( !bindings.containsKey(r)) {
	    bindings.put(r, null);
	}
    }

    protected RDFNode getValue(RDFNode r)
    {
        if(bindings.containsKey( r )) {
	    return (RDFNode) bindings.get(r);
        } else {
	    return null; // was r
        }
    }
    
    protected void bind(RDFNode a, RDFNode b )
    {
        bindings.put( a, b );
    }

    protected void unbind(RDFNode r )
    {
        bindings.put( r, null );
    }

    public String toString()
    {
	Iterator i = bindings.entrySet().iterator();
	String r = "";

	while( i.hasNext()) {
	    Map.Entry e = (Map.Entry)i.next();
	    r += e.getKey() + " -> " + e.getValue() + "\n";
	}
	return r;
    }
}

class AnonSelector extends SelectorImpl {
    
    protected AnonSelector( Resource s, Property p, RDFNode o )
    {
	super(s, p, o);
    }

    public boolean selects(Statement stmt )
    {
	if( subject == null ) {
            if( !stmt.getSubject().isAnon()) {
                return false;
            }
        }

	if( predicate == null ) {
            if( !stmt.getPredicate().isAnon()) {
                return false;
            }
        }
	if( object == null ) {
            RDFNode obj = stmt.getObject();
            if (obj instanceof Literal || !((Resource)obj).isAnon()) {
                return false;
            }
        }
	
	return true;
    } */
}