/*
 *  (c) Copyright Hewlett-Packard Company 2000, 2001
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Store.java
 *
 * Created on 15 August 2000, 14:25
 */

package com.hp.hpl.mesa.rdf.jena.mem;

import com.hp.hpl.mesa.rdf.jena.common.*;
import com.hp.hpl.mesa.rdf.jena.model.*;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;


/**
 *
 * @author  bwm
 * @version 
 */
class StoreMem extends StoreCom implements Store {
    
    HashSet statements = new HashSet();
    
    NodeMap subjects = new NodeMap();
    NodeMap predicates = new NodeMap();
    NodeMap objects = new NodeMap();

    /** Creates new Store */
    public StoreMem() {
    }
    
    public Resource getResource(String uri,Model m) throws RDFException {
        Resource newR = new ResourceImpl(uri, m);

        // check and see if we already have an object for this resource
        Statement s = subjects.getFirst(newR);
        if (s != null) return s.getSubject();
        s = objects.getFirst(newR);
        if (s != null) return (Resource) s.getObject();
        s = predicates.getFirst(newR);
        if (s != null) return s.getPredicate();
        
        // we don't so create a new one
        return newR;
    }
    
    public Property getProperty(Property p) {

        // check and see if we already have an object for this resource
        Statement s = predicates.getFirst(p);
        if (s != null) return s.getPredicate();
        
        // we don't so return the one we were passed
        return p;
    }

    public synchronized void add(Statement s) throws RDFException {
        if (statements.contains(s)) return;
        statements.add(s);
        subjects.add(s.getSubject(), s);
        predicates.add(s.getPredicate(), s);
        objects.add(s.getObject(),s);
    }
    
    public synchronized void remove(Statement s) throws RDFException {
        statements.remove(s);
        subjects.remove(s.getSubject(), s);
        predicates.remove(s.getPredicate(), s);
        objects.remove(s.getObject(), s);
    }
    
    public synchronized int size() throws RDFException {
        return statements.size();
    }
    
    public synchronized boolean contains(Statement s) throws RDFException {
        return statements.contains(s);
    }
    
    public synchronized boolean contains(Resource s, Property p) 
      throws RDFException {
        Iterator iter = null;
        iter = listBySubject(s);
        while (iter.hasNext()) {
            if ( ((Statement)iter.next()).getPredicate().equals(p)) {
                return true;
            }
        }
        return false;
    }
    
    public synchronized Iterator list() throws RDFException {
        return statements.iterator();
    }
    
    public synchronized Iterator list(Resource subject, Property predicate, RDFNode object)
      throws RDFException {
          
        // note that extra filtering may be required at by the caller this
        // implementation does not return the exact match
          
        // really should be bit smarter here - possibly test for lengths
        // of various searches instead of hard coding the order.
          
        if (subject != null) {
            return listBySubject(subject);
        } else if (object != null) {
            return listByObject(object);
        } else if (predicate != null) {
            return listByPredicate(predicate);
        } else {
            return list();
        }
   }
    
    public synchronized Iterator listBySubject(Resource subject) {
        //TODO there is a problem here
        //     this iterator should not be returned beyond the
        //     synchronized boundary
        return subjects.iterator(subject);
    }
    
    public synchronized Iterator listByPredicate(Property property) {
        return predicates.iterator(property);
    }
    
    public synchronized Iterator listByObject(RDFNode object) {
        return objects.iterator(object);
    }   
     
    public synchronized ResIterator listSubjects() throws RDFException {
        try {
            HashSet rSet = new HashSet();
            Iterator iter = subjects.listFirstStatements();
            while (iter.hasNext()) {
                rSet.add(((Statement) iter.next()).getSubject());
            }
            return new ResIteratorImpl(rSet.iterator(), rSet);
        } catch (Exception e) {
            ErrorHelper.logInternalError(this.getClass().getName(), 1, e);
            throw new RDFException(e);
        }        
    }    
    
    public synchronized NsIterator listNameSpaces() throws RDFException {
        try {
            HashSet nameSpaces = new HashSet();
            Iterator iter = predicates.listFirstStatements();
            while (iter.hasNext()) {
                nameSpaces.add(((Statement) iter.next()).getPredicate()
                                                        .getNameSpace());
            }
            return new NsIteratorImpl(nameSpaces.iterator(), nameSpaces);
        } catch (Exception e) {
            ErrorHelper.logInternalError(this.getClass().getName(), 2, e);
            throw new RDFException(e);
        }
    }
    
    protected class NodeMap {
        HashMap map = new HashMap();
        
        protected void add(RDFNode o, Statement s) {
            LinkedList l = (LinkedList) map.get(o);
            if (l==null) {
                l = new LinkedList();
                map.put(o,l);
            }
            l.add(s);
        }
        
        protected void remove(RDFNode o, Statement s) {
            LinkedList l = (LinkedList) map.get(o);
            if (l != null) {
                l.remove(s);
                if (l.size() == 0) {
                    map.put(o, null);
                }
            }
        }
        
        protected Iterator iterator(RDFNode o) {
            LinkedList l = (LinkedList) map.get(o);
            if (l==null) {
                return (new LinkedList()).iterator();
            } else {
                return l.iterator();
            }
        }
        
        protected Statement getFirst(RDFNode o) {
            LinkedList l = (LinkedList) map.get(o);
            if (l==null) {
                return null;
            } else {
                return (Statement) l.getFirst();
            }
        }
        
        protected Iterator listFirstStatements() {
            HashSet stmts = new HashSet();
            Iterator iter = ((Collection) map.values()).iterator();
            while (iter.hasNext()) {
                LinkedList l = ((LinkedList) iter.next());
                
                //TODO there is a possible problem here.
                //     Should not have to test of the hashmap value being null
                //     but the entry doesn't seem to get removed.
                //     will it ever be removed?
                //     And thats why we can't use keySet here
                if (l != null) {
                    stmts.add((Statement) l.getFirst());
                }
            }
            return stmts.iterator();
        }
    }        
}