/* -*-Java-*-
*******************************************************************
*
* File:         ConstraintsGeneric.java
* RCS:          $Header: /cvsroot/jena/jena/src/com/hp/hpl/mesa/rdf/jena/rdb/ConstraintsGeneric.java,v 1.2 2001/11/26 18:20:08 der Exp $
* Author:       Dave Reynolds, HPLabs, Bristol
* Created:      3/11/01
* Modified:     $Date: 2001/11/26 18:20:08 $
* Status:       Experimental (Do Not Distribute)
*
* (c) Copyright 2001, Hewlett-Packard Company, all rights reserved.
* [See end of file]
*********************************************************************/

//=======================================================================
// Package
package com.hp.hpl.mesa.rdf.jena.rdb;

//=======================================================================
// Imports
import com.hp.hpl.mesa.rdf.jena.model.*;
import java.sql.*;
import java.util.Iterator;

//=======================================================================
/**
* Implemention of the IConstraints interface used to specify search
* constraints on Jena statements. This implementation is suitable for
* any of the generic table layouts (whether hash or generator based) and
* should work for any reasonable database.
*
* @author <a href="mailto:der@hplb.hpl.hp.com">Dave Reynolds</a>
* @version $Revision: 1.2 $ on $Date: 2001/11/26 18:20:08 $
*/

public class ConstraintsGeneric implements IConstraints {

    /** Cached version of the sql as a prepared statement. Set to null to force rebuild. */
    protected PreparedStatement m_preparedSQL;

    /** Subject constraint or null if unconstrained */
    protected Resource m_subject;

    /** Property constraint or null if unconstrained */
    protected Resource m_property;

    /** Object constraint or null if unconstrained */
    protected RDFNode m_object;

    /** Flag if the object should be a literal */
    protected boolean m_objectIsLiteral;

    /** Object string constraint */
    protected String m_stringConstraint;

    /** The model being searched */
    protected ModelRDB m_model;

    /**
     * Constructor.
     */
    public ConstraintsGeneric(ModelRDB model) {
        m_model = model;
    }

    /**
     * Execute the query implied by this set of constraints against the statements
     * in the given model.
     */
    public synchronized ResultSetIterator runQuery(IRDBDriver driver, ModelRDB m) throws RDFException {
        try {
            if (m_preparedSQL == null) {
                // Format the constrained query
                StringBuffer query = new StringBuffer(800);
                boolean customConstraintSQL = false;
                if (m_stringConstraint != null) {
                    String constraintSQL = driver.getSQLCache().getSQLStatement("listConstraint");
                                // Some driver files require a special version of list select suitable for constraint extensions
                    customConstraintSQL = constraintSQL != null;
                    if (!customConstraintSQL) {
                        constraintSQL =  driver.getSQLCache().getSQLStatement("list");
                    }
                    query.append(constraintSQL);
                } else {
                    query.append(driver.getSQLCache().getSQLStatement("list"));
                }
                if (m_subject != null) query.append(" AND S.SUBJECT=?");
                if (m_property != null) query.append(" AND S.PREDICATE=?");
                if (m_object != null) {
                    query.append(" AND S.OBJECT=? AND S.OBJECT_ISLITERAL=");
                    query.append(m_objectIsLiteral ? "1" : "0");
                } else if (m_stringConstraint != null) {
                    if (customConstraintSQL) {
                        // In the custom version assume that L is aliased to RDF_LITERALS and this is a non-nested select
                        query.append(" AND S.OBJECT_ISLITERAL=1 AND S.OBJECT = RDF_LITERALS.ID AND " + m_stringConstraint);
                    } else {
                        // Default, assume we can use nested selects
                        query.append(" AND S.OBJECT_ISLITERAL=1 AND S.OBJECT IN (SELECT ID FROM RDF_LITERALS WHERE " + m_stringConstraint + ")");
                    }
                }
                //System.out.println("Constrained query: " + query);
                m_preparedSQL = driver.getConnection().prepareStatement(query.toString());
            }
            int args = 1;
            if (driver.supportsMultipleModels()) m_preparedSQL.setObject(args++, m.getDBID().getID());
            if (m_subject != null) {
                IDBID id = driver.getResourceID(m_subject);
                if (id == null) return new ResultSetIterator();     // Can't match
                m_preparedSQL.setObject(args++, id.getID());
            }
            if (m_property != null) {
                IDBID id = driver.getResourceID(m_property);
                if (id == null) return new ResultSetIterator();     // Can't match
                m_preparedSQL.setObject(args++, id.getID());
            }
            if (m_object != null) {
                IDBID id = null;
                if (m_objectIsLiteral) {
                    id = driver.getLiteralID((Literal)m_object);
                } else {
                    id = driver.getResourceID((Resource)m_object);
                }
                if (id == null) return new ResultSetIterator();     // Can't match
                m_preparedSQL.setObject(args++, id.getID());
            }
            return driver.getSQLCache().executeSQL(m_preparedSQL, null, new ResultSetStatementIterator(m, driver));
        } catch (SQLException e) {
            m_preparedSQL = null;
            throw new RDFRDBException("Problem formatting constrained query", e);
        }
    }

    /**
     * Add a constraint that the subject field should match the given resource
     */
    public IConstraints addSubjectConstraint(Resource r) throws RDFException {
        m_preparedSQL = null;
        m_subject = m_model.convert(r);
        return this;
    }

    /**
     * Add a constraint that the property field should match the given resource
     */
    public IConstraints addPropertyConstraint(Resource r) throws RDFException {
        m_preparedSQL = null;
        m_property = m_model.convert(r);
        return this;
    }

    /**
     * Add a constraint that the object field should match the given resource
     */
    public IConstraints addObjectConstraint(Resource r) throws RDFException {
        m_preparedSQL = null;
        m_object = m_model.convert(r);
        m_objectIsLiteral = false;
        return this;
    }

    /**
     * Add a constraint that the object field should match the given literal exactly
     */
    public IConstraints addObjectConstraint(Literal l) throws RDFException {
        m_preparedSQL = null;
        m_object = m_model.convert(l);
        m_objectIsLiteral = true;
        return this;
    }

    /**
     * Clear all constraints
     */
    public IConstraints reset() {
        m_preparedSQL = null;
        m_subject = null;
        m_property = null;
        m_object = null;
        m_objectIsLiteral = false;
        m_stringConstraint = null;
        return this;
    }

    /**
     * Add a constraint that the object field should match the given string literal
     * using the sql comparison operator op - such as "LIKE" or "NOT LIKE"
     */
    public IConstraints addStringConstraint(String op, String constraint) {
        return doAddSqlConstraint("RDF_LITERALS.LITERAL_IDX " + op + " '" + constraint + "'");
    }

    /**
     * Helper function to add an extra sql constraint
     */
    protected IConstraints doAddSqlConstraint(String sql) {
        m_preparedSQL = null;
        m_object = null;
        if (m_stringConstraint == null) {
            m_stringConstraint = sql;
        } else {
            m_stringConstraint += " AND " + sql;
        }
        m_objectIsLiteral = true;
        return this;
    }

    /**
     * Add a constraint that the object field should match the given integer literal
     * using the sql comparison operator op - examples include "&ge;" and "&le;=".
     */
    public IConstraints addIntConstraint(String op, int constraint) {
        return doAddSqlConstraint("RDF_LITERALS.INT_LITERAL " + op + " " + constraint + " AND RDF_LITERALS.INT_OK = 1");
    }

}

/*
 *  (c) Copyright Hewlett-Packard Company 2000, 2001
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

