/* -*-Java-*-
*******************************************************************
*
* File:         DBConnection.java
* RCS:          $Header: /cvsroot/jena/jena/src/com/hp/hpl/mesa/rdf/jena/rdb/DBConnection.java,v 1.6 2002/07/08 13:57:09 der Exp $
* Author:       Dave Reynolds, HPLabs, Bristol
* Created:      17/9/01
* Modified:     $Date: 2002/07/08 13:57:09 $
* Status:       Experimental (Do Not Distribute)
*
* (c) Copyright 2001, Hewlett-Packard Company, all rights reserved.
* [See end of file]
*********************************************************************/

//=======================================================================
// Package
package com.hp.hpl.mesa.rdf.jena.rdb;

//=======================================================================
// Imports
import java.sql.*;
import java.util.*;
import java.io.*;
import com.hp.hpl.jena.util.Log;

//=======================================================================
/**
* Encapsulate the specification of a jdbc connection, mostly used to
* simplify the calling pattern for ModelRDB factory methods.
*
* @author <a href="mailto:der@hplb.hpl.hp.com">Dave Reynolds</a>
* @version $Revision: 1.6 $ on $Date: 2002/07/08 13:57:09 $
*/

public class DBConnection implements IDBConnection {

    /** The jdbc connection being wrapped up */
    protected Connection m_connection;

    /** The url for the connection, may be null if the connection was passed in pre-opened */
    protected String m_url;

    /** The user name for the connection, may be null if the connection was passed in pre-opened */
    protected String m_user;

    /** The password for the connection, may be null if the connection was passed in pre-opened */
    protected String m_password;

    /** The RDF driver to use for access data on this connection */
    protected IRDBDriver m_driver;

    /** Flag to indicate that we have tried to load the layout info table */
    protected boolean m_loadedLayout = false;

    /** Flag to indicate that the database seems to be formated - only valid when m_loadedLayout is true */
    protected boolean m_formated = false;

    /** Set of database format properties loaded from an standard RDF_LAYOUT_INFO table */
    protected Properties m_properties = new Properties();

    /**
     * Create a connection specification based on jdbc address and
     * appropriate authentication information.
     * @param url the jdbc url for the database, note that the format of this
     * is database dependent and that the appropriate jdbc driver will need to
     * be specified either in the jdbc.drivers property or via the standard pattern
     * <pre>
     *     Class.forName("my.sql.driver");
     * </pre>
     * @param user the user name to log on with
     * @param password the password corresponding to this user
     */
    public DBConnection(String url, String user, String password) {
        m_url = url;
        m_user = user;
        m_password = password;
    }

    /**
     * Create a connection specification that just wraps up an existing database
     * connection.
     * @param connection the open jdbc connection to use
     */
    public DBConnection(Connection connection) {
        m_connection = connection;
    }

    /**
     * Return the jdbc connection.
     * If the connection has been closed but we know the db uri we re-open the connection,
     * otherwise we throw an SQL exception.
     */
    public Connection getConnection() throws SQLException {
        if (m_connection == null) {
            if (m_url != null) {
                m_connection = DriverManager.getConnection(m_url, m_user, m_password);
                m_connection.setAutoCommit(true);
            }
            if (m_connection == null) {
                // Can no longer open a connection, throw an exception
                throw new SQLException("Attempted to access closed SQLConnection");
            }
        }
        return m_connection;
    }

    /**
     * Close the jdbc connection.
     * It might be reopend with a getConnection() if the full database uri, user and password
     * were provided.
     */
    public void close() throws SQLException {
        if (m_connection != null) {
            m_connection.close();
            m_connection = null;
        }
    }

    /**
     * Return true if the database seems to be formated for RDF storage.
     * This is <em>not</em> an integrity check this is simply a flag
     * recording that a base level table exists.
     * Any access errors are treated as the database not being formated.
     */
    public boolean isFormatOK() throws RDFRDBException {
        try {
            loadProperties();
            return m_formated;
        } catch (SQLException e) {
            return false;
        }
    }

    /**
     * Returns a property value for the database at the end of the
     * connection. This is used to retrieve appropriate formating and
     * driver information from a standard RDF_LAYOUT_INFO table.
     * For a database that has not been formatted all calls will
     * return null.
     * Any access errors are treated as the database not being formated.
     */
    public String getProperty(String propname) throws SQLException {
        loadProperties();
        return m_properties.getProperty(propname);
    }

    /**
     * Returns a set of property values for the database at the end of the
     * connection. This is used to retrieve appropriate formating and
     * driver information from a standard RDF_LAYOUT_INFO table.
     * For a database that has not been formatted all calls will
     * return null.
     * Any access errors are treated as the database not being formated.
     */
    public Properties getProperties() throws SQLException {
        loadProperties();
        return m_properties;
    }

    /**
     * Return an IRDBDriver suitable for this database connection.
     * This is only usable for a preformatted database, returns null if the
     * database has not been formatted or is corrupt or the driver is not loadable.
     * Throws an Exception if can't even connect to the database.
     */
    public IRDBDriver getDriver() throws RDFRDBException {
        if (m_driver != null) return m_driver;
        try {
            String driverClassName = getProperty(DRIVER_CLASSNAME);
            if (driverClassName == null) {
                String databaseType = getProperty(DATABASE_TYPE);
                String layoutType = getProperty(LAYOUT_TYPE);
                if (databaseType == null || layoutType == null) return null;
                m_driver = getDriver(layoutType, databaseType);
            } else {
                m_driver = getDriver(driverClassName);
            }
            m_driver.loadProperties(getProperties());
            m_driver.setConnection(this);
            return m_driver;
        } catch (SQLException e) {
            throw new RDFRDBException("Failure to instansiate DB driver", e);
        }
    }

    /**
     * Set the IRDBDriver to use for this connection.
     * Useful to enable external drivers to be registered outside of the
     * standard driver package. If you set this to null then the next
     * attempt to fetch a driver will re-run the driver selection logic.
     */
    public void setDriver(IRDBDriver driver) {
        m_driver = driver;
    }

    /**
     * Helper function to locate and instantiate the driver class corresponding
     * to a given layout and database name
     * Throws an RDFRDBexception if the driver can't be instantiated
     */
    public IRDBDriver getDriver(String layout, String database) throws RDFRDBException {
        if (m_driver != null) return m_driver;
        // First try for a specific driver class
        String qualifiedClassName = "com.hp.hpl.mesa.rdf.jena.rdb.Driver" + database + layout;
        try {
            m_driver = getDriver(qualifiedClassName);
        } catch (RDFRDBException e) {
            // Nope, so try to find a database/layout specifc config file
            String defaultConfig = "etc/" + database + "_" + layout + ".config";
            InputStream is = DBConnection.class.getClassLoader().getResourceAsStream(defaultConfig);
            if (is == null) {
                // Nope, try a generic layout and database config file
                defaultConfig = "etc/" + database + ".config";
                m_driver = getDriver(layout, "Generic");
                is = DBConnection.class.getClassLoader().getResourceAsStream(defaultConfig);
                if (is == null) {
                    // Nope, give in
                    throw new RDFRDBException("Failure to instantiate DB driver for " + layout + " on " + database);
                }
            }
            try {
                m_properties = new Properties();
                m_properties.load(is);
                is.close();
                if (m_driver == null) {
                    String baseClass = m_properties.getProperty("baseClass");
                    m_driver = (IRDBDriver) Class.forName(baseClass).newInstance();
                }
                m_driver.loadProperties(m_properties);
            } catch (Exception e2) {
                throw new RDFRDBException("Failure to instantiate DB driver for " + layout + " on " + database);
            }
        }
        if (m_driver != null) {
            m_driver.setConnection(this);
        }
        return m_driver;
    }

    /**
     * Force the connection and driver properties to be reloaded from the database
     */
    public void reloadProperties() throws RDFRDBException {
        m_loadedLayout = false;
        try {
            loadProperties();
            if (m_driver != null) {
                m_driver.loadProperties(m_properties);
                m_driver.setConnection(this);
            }
        } catch (SQLException e) {
            throw new RDFRDBException("Problem reloading driver properties from database", e);
        }
    }

    /**
     * Helper function to locate and instantiate the driver class corresponding
     * to a given classname.
     * Throws an RDFRDBexception if the driver can't be instantiated
     */
    private IRDBDriver getDriver(String driverClassName) throws RDFRDBException {
        if (m_driver != null) return m_driver;
        try {
            m_driver = (IRDBDriver) Class.forName(driverClassName).newInstance();
            //m_driver.setConnection(this);
            return m_driver;
        } catch (Exception e) {
            // Try fall back of config file on generic driver
            throw new RDFRDBException("Failure to instansiate DB driver: " + driverClassName, e);
        }
    }

    /**
     * Add a new property value to both RDF_LAYOUT_INFO table.
     * Only works if the database has already been formated.
     */
    public void addProperty(String propname, String value) throws SQLException {
        PreparedStatement st = getConnection().prepareStatement("INSERT INTO RDF_LAYOUT_INFO (NAME, VAL) VALUES (?,?)");
        st.setString(1, propname);
        st.setString(2, value);
        st.executeUpdate();
        st.close();
        m_properties.put(propname, value);
    }

    /**
     * Attempt to load the database layout information from the
     * standard RDF_LAYOUT_INFO table
     */
    protected void loadProperties() throws SQLException {
        if (!m_loadedLayout) {
            Connection con =  getConnection();
            if (con == null) throw new SQLException("Failed to open database " + m_url);
            try {
                Statement st = con.createStatement();
                ResultSet rs = st.executeQuery("SELECT NAME, VAL FROM RDF_LAYOUT_INFO");
                m_properties = new Properties();
                while (rs.next()) {
                    m_properties.put(rs.getString(1), rs.getString(2));
                }
                m_formated = true;
            } catch (SQLException e) {
                // Any other error is treated as a badly formated database
                m_formated = false;
            }
            m_loadedLayout = true;
        }
    }

}

/*
 *  (c) Copyright Hewlett-Packard Company 2000, 2001
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */