/* -*-Java-*-
*******************************************************************
*
* File:         DriverGenericAttribute.java
* RCS:          $Header: /cvsroot/jena/jena/src/com/hp/hpl/mesa/rdf/jena/rdb/DriverGenericAttribute.java,v 1.2 2001/11/26 18:20:08 der Exp $
* Author:       Dave Reynolds, HPLabs, Bristol
* Created:      15/11/01
* Modified:     $Date: 2001/11/26 18:20:08 $
* Status:       Experimental (Do Not Distribute)
*
* (c) Copyright 2001, Hewlett-Packard Company, all rights reserved.
* [See end of file]
*********************************************************************/

//=======================================================================
// Package
package com.hp.hpl.mesa.rdf.jena.rdb;

//=======================================================================
// Imports
import com.hp.hpl.mesa.rdf.jena.common.*;
import com.hp.hpl.mesa.rdf.jena.model.*;
import java.io.*;
import java.sql.*;
import java.util.*;
import com.hp.hpl.jena.util.Log;

//=======================================================================
/**
* Adaption of the base generic layout driver to use separate tables
* for specific predicates (called attributes).
* This version just duplicates the generic store do a replacement
* version if this looks useful.
*
* @todo store and retrieve attributeMap in database. This is just a hack for
* timing, not for real use.
*
* @author <a href="mailto:der@hplb.hpl.hp.com">Dave Reynolds</a>
* @version $Revision: 1.2 $ on $Date: 2001/11/26 18:20:08 $
*/

public class DriverGenericAttribute extends DriverGenericGeneric {

//=======================================================================
// Internal variables

    /** Table of properties (given by dbid value) which are optmized as separate
     *  attribute tables, map value is the is the ID of the attribute entry.
     */
    protected Map attributeMap = new HashMap();

    /** Index to use for next entry in attributeMap */
    protected int attributeIndexCount = 0;

//=======================================================================
// Constructor

    /**
     * Create a bare instance of the driver. It is not functional until a
     * database connection has been supplied via setConnection.
     */
    public DriverGenericAttribute() {
        super();
        // Set configuration parameters
        LAYOUT_TYPE = "Attribute";
        DATABASE_TYPE = "Generic";
    }


    /**
     * Hint to the store that the given property could be stored as an
     * attribute table.
     */
    public void hintPropertyTable(Property p) throws RDFException {
        IDBID dbid =  addResource(p);
        Integer idx = new Integer(attributeIndexCount++);
        attributeMap.put(dbid.getID(), idx);
        // Create an attribute table.
        try {
            m_sql.flushPreparedStatementCache();
            java.sql.Statement st = getConnection().createStatement();
            Iterator it = m_sql.getSQLStatementGroup("createAttributeTable").iterator();
            while (it.hasNext()) {
                String op = m_sql.substitute((String)it.next(), "${a}", idx.toString());
                st.executeUpdate(op);
            }
            st.close();
        } catch (SQLException e) {
            throw new RDFRDBException("Failed to create attributed table", e);
        }
    }

    /**
     * Register a statement in the database.
     * @param subjId database index of the subject entry
     * @param predId database index of the predicate entry
     * @param objId database index of the object entry
     * @param isLiteral if the object is a literal rather than a resource
     */
    public void addStatement(com.hp.hpl.mesa.rdf.jena.model.Statement s, ModelRDB m) throws RDFException {
        // Check if is already asserted
        if (!SKIP_DUPLICATE_CHECK) {
            if (isStatementPresent(s.getSubject(), s.getPredicate(), s.getObject(), m)) return;
        }
        // First ensure the components are all added
        IDBID predId = addResource(s.getPredicate());
        IDBID subjId = addResource(s.getSubject());
        RDFNode obj = s.getObject();
        IDBID objId = null;
        boolean isLiteral = obj instanceof Literal;
        if (isLiteral) {
            objId = addLiteral((Literal)obj);
        } else {
            objId = addResource((Resource)obj);
        }
        // Now add the statement itself
        try {
            Object[] args = new Object[] {subjId.getID(),
                                          predId.getID(),
                                          objId.getID(),
                                          new Short((short)(isLiteral ? 1 : 0)),
                                          m.getDBID().getID()};
            m_sql.runSQLQuery(INSERT_BY_PROCEDURE ? "insertStatementProc" : "insertStatement", args);
            // Also add to attribute table
            Integer aid = (Integer) attributeMap.get(predId.getID());
            if (aid != null) {
                String aTable = "RDF_A_" + aid;
                /** @todo - replace by sql file statement with substitution */
                /** @todo - package up substitutional caller better */
                PreparedStatement ps = getConnection().prepareStatement("INSERT INTO " + aTable + "(SUBJECT, PREDICATE, OBJECT, OBJECT_ISLITERAL, MODEL) VALUES(?,?,?,?,?)");
                ps.setObject(1, subjId.getID());
                ps.setObject(2, predId.getID());
                ps.setObject(3, objId.getID());
                ps.setInt(4, isLiteral ? 1 : 0);
                ps.setObject(5, m.getDBID().getID());
                ps.executeUpdate();
                ps.close();
            }
        } catch (SQLException e1) {
            throw new RDFRDBException("Failed to register statement", e1);
        }

    }

    /**
     * List all the statements in the model which match the given subject,
     * predicate object - each of which can be null (meaning wildcard).
     */
    public ResultSetIterator listStatements(Resource subject, Property predicate, RDFNode object, ModelRDB m) throws RDFException {
        // Determine the SQL operation name
        String opname = "list";
        if (subject != null) opname += "S";
        if (predicate != null) opname += "P";
        if (object != null) opname += "O";
        try {
            if (predicate != null) {
                IDBID predID = getResourceID(predicate);
                if (predID != null) {
                    Integer aid = (Integer) attributeMap.get(predID.getID());
                    if (aid != null) {
                        // Special attribute table lookup, just testing one case so far
                        String aopname = "a" + opname;
                        int argi = 1;
                        String attr = aid.toString();
                        PreparedStatement ps = m_sql.getPreparedSQLStatement(aopname, attr);
                        if (subject != null) {
                            IDBID sid = getResourceID(subject);
                            if (sid == null) return new ResultSetIterator();
                            ps.setObject(argi++, sid.getID());
                        }
                        ps.setObject(argi++, predID.getID());
                        if (object != null) {
                            if (object instanceof Literal) {
                                IDBID lid = getLiteralID((Literal)object);
                                if (lid == null) {
                                    return new ResultSetIterator();
                                }
                                ps.setObject(argi++, lid.getID());
                                ps.setInt(argi++, 1);
                            } else {
                                IDBID lid = getResourceID((Resource)object);
                                if (lid == null) return new ResultSetIterator();
                                ps.setObject(argi++, lid.getID());
                                ps.setInt(argi++, 0);
                            }
                        }
                        return m_sql.executeSQL(ps, aopname + attr, new ResultSetStatementIterator(m, this));
                    }
                }
            }
            Object[] args = statementArgs(subject, predicate, object, m);
            if (args == null) return new ResultSetIterator();
            return  m_sql.runSQLQuery(opname, args, new ResultSetStatementIterator(m, this));
        } catch (SQLException e) {
            throw new RDFRDBException("Internal sql failure in " + opname, e);
        }
    }


    /**
     * Remove all RDF information from a database.
     * This will only work if there are no other clients us the DB so is mostly to
     * simplify testing that a practical tool. Partial DB config or cleaning can leave
     * the DB in a strange state hence it will continue with each cleanup phase even
     * if earlier phases have generated errors and erros are only logged, not propagated
     * as exceptions.
     */
    public void cleanDB() {
        super.cleanDB();
        Iterator it = attributeMap.values().iterator();
        try {
            java.sql.Statement st = getConnection().createStatement();
            while (it.hasNext()) {
                Integer id = (Integer)it.next();
                try {
                    st.executeUpdate("DROP TABLE RDF_A_"+id);
                } catch (SQLException e) {
                    Log.warning("Problem cleaning DB: ", e);
                }
            }
           st.close();
        } catch (Exception e) {
            Log.warning("Problem in DB access during clean up", e);
        }
    }
}

/*
 *  (c) Copyright Hewlett-Packard Company 2000, 2001
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
