/* -*-Java-*-
*******************************************************************
*
* File:         DriverGenericHash.java
* RCS:          $Header: /cvsroot/jena/jena/src/com/hp/hpl/mesa/rdf/jena/rdb/DriverGenericHash.java,v 1.4 2002/07/01 15:20:21 der Exp $
* Author:       Dave Reynolds, HPLabs, Bristol
* Created:      30/10/01
* Modified:     $Date: 2002/07/01 15:20:21 $
* Status:       Experimental (Do Not Distribute)
*
* (c) Copyright 2001, Hewlett-Packard Company, all rights reserved.
* [See end of file]
*********************************************************************/

//=======================================================================
// Package
package com.hp.hpl.mesa.rdf.jena.rdb;

//=======================================================================
// Imports
import com.hp.hpl.mesa.rdf.jena.common.*;
import com.hp.hpl.mesa.rdf.jena.model.*;
import java.io.*;
import java.sql.*;
import java.util.*;
import java.util.zip.CRC32;
import java.security.MessageDigest;
import com.hp.hpl.jena.util.Log;

//=======================================================================
/**
* Adaption of the base generic layout driver to use unique hashes of
* as the index terms instead of relying on database sequences. This is
* useful for databases without useful sequences/generators it may also
* yield marginally higher performance.
*
* @author <a href="mailto:der@hplb.hpl.hp.com">Dave Reynolds</a>
* @version $Revision: 1.4 $ on $Date: 2002/07/01 15:20:21 $
*/

public class DriverGenericHash extends DriverGenericGeneric {

//=======================================================================
// Internal variables

    /** MD5 hash generator */
    protected static MessageDigest m_hashGenerator;
    //protected static MD5 m_hashGenerator = new MD5();

    /** Base64 encoder  */
    protected static sun.misc.BASE64Encoder m_base64Encoder = new sun.misc.BASE64Encoder();

    /** table of default sql statements */
    protected static Properties m_defaultHashSQL = null;

    /** The name of the default sql definition file */
    protected final static String HASHSQL_DEFAULT_FILE = "etc/generic_hash.sql";

    /**
     * Create a bare instance of the driver. It is not functional until a
     * database connection has been supplied via setConnection.
     */
    public DriverGenericHash() {
        super();
        // Set configuration parameters
        LAYOUT_TYPE = "Hash";
        DATABASE_TYPE = "Generic";
        SQL_FILE = HASHSQL_DEFAULT_FILE;
        try {
            m_hashGenerator = MessageDigest.getInstance("MD5");
        } catch (java.security.NoSuchAlgorithmException e) {
            Log.severe("Failed to find an MD5 provider", e);
        }
        m_anonNSId = uniqueHash(ANON_NAMESPACE);
        cacheNamespaceId(ANON_NAMESPACE, m_anonNSId);
        SKIP_ALLOCATE_ID = false;
        INSERT_BY_PROCEDURE = false;
        ID_SQL_TYPE = "CHAR(24)";
    }

    /**
     * Supporting function to load the table of default SQL operations.
     * Override base one to use a different base set of SQL that supports
     * hash indexing
     */
    protected Properties getDefaultSQL() throws IOException {
        if (m_defaultHashSQL == null) {
            m_defaultHashSQL = SQLCache.loadSQLFile(HASHSQL_DEFAULT_FILE, null, ID_SQL_TYPE);
        }
        return m_defaultHashSQL;
    }

    /**
     * Convert the raw SQL object used to store a database identifier into a java object
     * which meets the IDBID interface.
     */
    public IDBID wrapDBID(Object id) throws RDFRDBException {
        if (id instanceof String) {
            return new DBIDHash((String)id);
        } else if (id == null) {
            return null;
        } else {
            /** @todo modify signature so this can raise an exception, easy once we switch to non-checked exceptions */
            throw new RDFRDBException("Unexpected DB identifier type: " + id);
            //return null;
        }
    }

//=======================================================================
// Support for registering/looking up resources, literals, namespaces
// For most categories we need operations to:
//    getXID - determine the DBID, if available
//    getX   - reconstruct the object from its DBID
//    addX   - add to the database, finds DBID as side effect
//=======================================================================

    /**
     * Generate a unique hash for a string.
     * The result is an 16-byte array, which, unfortunately is too big for most bigint
     * database types. Some databases (e.g. Interbase) can directly pack these
     * into char(16) types but others (e.g. Postgres) can't even do that so we
     * end up base64 encoding and using a char(24) key for everything.
     */
    protected synchronized IDBID uniqueHash(String arg) {
        m_hashGenerator.update(arg.getBytes());
        byte[] res = m_hashGenerator.digest();
        m_hashGenerator.reset();
        return new DBIDHash(m_base64Encoder.encode(res));
    }

    /**
     * Return the namespace ID, if it doesn't have one yet then allocate one.
     */
    protected IDBID allocateNamespaceID(String namespace) throws RDFRDBException{
        if (m_namespaceIDMap.containsKey(namespace)) {
            return (IDBID) m_namespaceIDMap.get(namespace);
        } else {
            return uniqueHash(namespace);
        }
    }

    /**
     * Return the database ID for a resource
     */
    protected IDBID allocateResourceID(Resource resource) throws RDFException {
        String namespace = resource.isAnon() ? ANON_NAMESPACE : resource.getNameSpace();
        IDBID nsid = getNamespaceID(namespace);
        if (nsid == null) nsid = allocateNamespaceID(namespace);
        String localname = resource.isAnon() ? resource.getId().toString() : resource.getLocalName();
        return uniqueHash(localname + nsid.getID().toString());
    }

    /**
     * Return the database ID for the literal
     */
    protected IDBID allocateLiteralID(Literal l) throws RDFRDBException {
        return uniqueHash(l.toString() + l.getLanguage());
    }

    /**
     * Allocate a new ID for the statement
     */
    protected IDBID allocateStatementID(com.hp.hpl.mesa.rdf.jena.model.Statement statement, IDBID modelID) throws RDFException {
        StringBuffer fingerPrint = new StringBuffer();
        fingerPrint.append(allocateResourceID(statement.getSubject()).getID().toString());
        fingerPrint.append("|");
        fingerPrint.append(allocateResourceID(statement.getPredicate()).getID().toString());
        fingerPrint.append("|");
        RDFNode obj = statement.getObject();
        if (obj instanceof Literal) {
            fingerPrint.append(allocateLiteralID((Literal)obj).getID().toString());
        } else {
            fingerPrint.append(allocateResourceID((Resource)obj).getID().toString());
        }
        fingerPrint.append("|");
        fingerPrint.append(modelID.toString());
        return uniqueHash(fingerPrint.toString());
    }

}

/*
 *  (c) Copyright Hewlett-Packard Company 2000, 2001
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
