/* -*-Java-*-
*******************************************************************
*
* File:         DriverGenericMMGeneric.java
* RCS:          $Header: /cvsroot/jena/jena/src/com/hp/hpl/mesa/rdf/jena/rdb/DriverGenericMMGeneric.java,v 1.5 2002/07/15 17:53:21 der Exp $
* Author:       Dave Reynolds, HPLabs, Bristol
* Created:      5/10/01
* Modified:     $Date: 2002/07/15 17:53:21 $
* Status:       Experimental (Do Not Distribute)
*
* (c) Copyright 2001, Hewlett-Packard Company, all rights reserved.
* [See end of file]
*********************************************************************/

//=======================================================================
// Package
package com.hp.hpl.mesa.rdf.jena.rdb;

//=======================================================================
// Imports
import com.hp.hpl.mesa.rdf.jena.common.*;
import com.hp.hpl.mesa.rdf.jena.model.*;
import java.io.*;
import java.sql.*;
import java.util.*;
import com.hp.hpl.jena.util.Log;

//=======================================================================
/**
* Adaption of the base generic layout driver to support multiple
* models in one database.
*
* @author <a href="mailto:der@hplb.hpl.hp.com">Dave Reynolds</a>
* @version $Revision: 1.5 $ on $Date: 2002/07/15 17:53:21 $
*/

public class DriverGenericMMGeneric extends DriverGenericGeneric implements IRDBDriver {

    /** table of default sql statements */
    protected static Properties m_defaultMMSQL = null;

    /** The name of the default sql definition file */
    protected final static String MMSQL_DEFAULT_FILE = "etc/generic_mmgeneric.sql";

    /**
     * Create a bare instance of the driver. It is not functional until a
     * database connection has been supplied via setConnection.
     */
    public DriverGenericMMGeneric() {
        super();
        // Set configuration parameters
        LAYOUT_TYPE = "MMGeneric";
        DATABASE_TYPE = "Generic";
        SQL_FILE = MMSQL_DEFAULT_FILE;
    }

    /**
     * Supporting function to load the table of default SQL operations.
     * Override base one to use a different base set of SQL that supports model id
     * checking.
     */
    protected Properties getDefaultSQL() throws IOException {
        if (m_defaultMMSQL == null) {
            m_defaultMMSQL = SQLCache.loadSQLFile(MMSQL_DEFAULT_FILE, null, ID_SQL_TYPE);
        }
        return m_defaultMMSQL;
    }

//=======================================================================
// Support multiple models

    /**
     * Returns true of the database layout supports multiple RDF models
     * in the same database.
     */
    public boolean supportsMultipleModels() {
        return true;
    }

    /**
     * Allocate a new databaseID for a new model in the same database.
     * Unlike other allcoate calls this also record the allocation in the database directly.
     * Throws RDFRDBException if the driver does not support multiple models per db.
     */
    public IDBID allocateModelID(String modelName) throws RDFRDBException {
        try {
            IDBID id = null;
            ResultSetIterator rs = m_sql.runSQLQuery("getModelID", new Object[] {modelName});
            if (rs != null && rs.hasNext()) {
                id = wrapDBID(rs.getSingleton());
            } else {
                if (!SKIP_ALLOCATE_ID) {
                    rs = m_sql.runSQLQuery("allocateModelID", null);
                    id = wrapDBID(rs.getSingleton());
                    m_sql.runSQLQuery("insertModelID", new Object[] {modelName, id.getID()});
                } else {
                    m_sql.runSQLQuery("insertModelID", new Object[] {modelName});
                    id = getModelID(modelName);
                }
            }
            return id;
        } catch (SQLException e) {
            throw new RDFRDBException("Failed to allocate new model id", e);
        }
    }

    /**
     * Return the databaseID for an existing named model within the database.
     * Throws RDFRDBException if the driver does not support multiple models per db.
     */
    public IDBID getModelID(String modelName) throws RDFRDBException {
        try {
            ResultSetIterator rs = m_sql.runSQLQuery("getModelID", new Object[] {modelName});
            return wrapDBID(rs.getSingleton());
        } catch (SQLException e) {
            throw new RDFRDBException("No such model called: " + modelName + " found in database", e);
        }
    }

    /**
     * List the names of all models in this database
     */
    public Iterator listModels() throws RDFRDBException {
        try {
            return m_sql.runSQLQuery("listModels", null, new ResultSetStringIterator());
        } catch (SQLException e) {
            throw new RDFRDBException("Problem listing models", e);
        }
    }


    /**
     * Delete the named from from within the database.
     * This will remove any statements associated with the model and
     * remove the model from the table of known models.
     * Throws RDFRDBException if the driver does not support multiple models per db.
     */
    public void deleteModel(String modelName) throws RDFRDBException {
        try {
            IDBID mid = getModelID(modelName);
            m_sql.runSQLQuery("deleteAllStatements", new Object[] {mid.getID()});
            m_sql.runSQLQuery("deleteModelID", new Object[] {modelName});
        } catch (SQLException e) {
            throw new RDFRDBException("Problem listing models", e);
        }
    }


}

/*
 *  (c) Copyright Hewlett-Packard Company 2000, 2001
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
