/* -*-Java-*-

*******************************************************************

*

* File:         DriverOracleMMGeneric.java

* RCS:          $Header: /cvsroot/jena/jena/src/com/hp/hpl/mesa/rdf/jena/rdb/DriverOracleMMGeneric.java,v 1.4 2002/07/18 12:55:54 der Exp $

* Author:       Dave Reynolds, HPLabs, Bristol

* Created:      20/2/02

* Modified:     $Date: 2002/07/18 12:55:54 $

* Status:       Experimental (Do Not Distribute)

*

* (c) Copyright 2001, Hewlett-Packard Company, all rights reserved.

* [See end of file]

*********************************************************************/



//=======================================================================

// Package

package com.hp.hpl.mesa.rdf.jena.rdb;



//=======================================================================

// Imports

import com.hp.hpl.mesa.rdf.jena.common.*;

import com.hp.hpl.mesa.rdf.jena.model.*;

import java.io.*;

import java.sql.*;

import java.util.*;

import com.hp.hpl.jena.util.Log;



//=======================================================================

/**

* Customize the MMGeneric driver for use with Oracle.

* The key issues with Oracle are:

* (1) Possible return of bignums for IDs - now handled in generic driver

* (2) Empty strings are stored as nulls and so do not match other empty strings,

*     this breaks all the "languge" field processing in literals in several ways.

*     It also causes problems with resource localnames that are empty

*     strings (because they re-appear from the database as nulls) and empty literals

*     (because the they go in as nulls which then violate the non-null index

*     condition).

*

* @author <a href="mailto:der@hplb.hpl.hp.com">Dave Reynolds</a>

* @version $Revision: 1.4 $ on $Date: 2002/07/18 12:55:54 $

*/



public class DriverOracleMMGeneric extends DriverGenericMMGeneric implements IRDBDriver {



    /** table of default sql statements */

    protected static Properties m_defaultMMSQL = null;



    /** String used inplace of empty literals to avoid the NULL-to-empty string problem */

    protected static String EMPTY_LITERAL_MARKER = "empty-literal";



    /**

     * Create a bare instance of the driver. It is not functional until a

     * database connection has been supplied via setConnection.

     */

    public DriverOracleMMGeneric() {

        super();

        // Set configuration parameters

        DRIVER_NAME = "oracle.jdbc.driver.OracleDriver";

        LAYOUT_TYPE = "MMGeneric";

        DATABASE_TYPE = "Oracle";

        SQL_FILE = "etc/oracle_mmgeneric.sql";

        MAX_LITERAL = 1000;

        ID_SQL_TYPE = "INTEGER";

    }



//=======================================================================

// Patched functions to adapt to oracle jdbc driver expectations



    /**

     * Return the database ID for the literal, if it exists

     */

    public IDBID getLiteralID(Literal l) throws RDFRDBException {

        try {

            String ls = l.toString();

            int len = ls.length();

            if (len >= MAX_LITERAL) {

                ls = getLiteralIdx(ls);

            }

            String lang = l.getLanguage();

            boolean nullLang = (lang == null) || (lang.equals(""));

            // Oracle (thin driver at least) fails to match empty strings so have to watch out for when lang is empty

            String opName = nullLang ? "getLiteralIDNoLang" : "getLiteralID";

            if (len == 0) {

                opName += "NullLiteral";

                ls = EMPTY_LITERAL_MARKER;

            }

            PreparedStatement ps = m_sql.getPreparedSQLStatement(opName);

            ps.setString(1, ls);

            if (!nullLang) ps.setString(2, lang);

            ResultSet rs = ps.executeQuery();

            IDBID result = null;

            if (rs.next()) {

                result = wrapDBID(rs.getObject(1));

            }

            rs.close();

            m_sql.returnPreparedSQLStatement(ps, opName);

            return result;

        } catch (SQLException e1) {

            // /* DEBUG */ System.out.println("Literal truncation (" + l.toString().length() + ") " + l.toString().substring(0, 150));

            throw new RDFRDBException("Failed to find literal", e1);

        }

    }



    /**

     * Wrap up a boolean flag as a object which the jdbc driver can assert into a boolean/short column

     */

    public Object wrapFlag(boolean flag) {

        return  flag ? new Integer(1) : new Integer(0);

    }



    /**

     * Fetch a literal just knowing its literal rdb-id.

     * Can be null if the ID can from a hash function and literal isn't registered yet.

     * Patches up any mangled lang fields from Oracle.

     */

    public Literal getLiteral(IDBID id) throws RDFException {

        Literal l = super.getLiteral(id);

        if (l != null && l.getLanguage() == null) {

            // At this point Oracle has retrieved something which was an empty string as a null

            // Can't just fixup inplace (no setLanguage operator on LiteralImpl) so create new one.

            // The overhead is probably trivial compared to the DB call anyway.

            return new LiteralImpl(l.getString(), "", l.getWellFormed());

        } else {

            return l;

        }

    }



    /**

     * Lookup the unique ID for a resource.

     * Return null if it does not have one yet.

     */

    public IDBID getResourceID(Resource resource) throws RDFException {

        try {

            HasRDBID resourceRDB = (HasRDBID) resource;

            IDBID id = resourceRDB.getRDBId();

            if (id != null) return id;

            String namespace = resource.isAnon() ? ANON_NAMESPACE : resource.getNameSpace();

            IDBID nsid = getNamespaceID(namespace);

            if (nsid == null) return null;

            resourceRDB.setRDBNSId(nsid);

            String localname = resource.isAnon() ? resource.getId().toString() : resource.getLocalName();

            boolean nullName = localname == null || localname.equals("");

            String opName = nullName ? "getResourceIDNoName" : "getResourceID";

            Object[] args = nullName ? new Object[] {nsid.getID()} : new Object[] {localname, nsid.getID()};

            ResultSetIterator rs = m_sql.runSQLQuery(opName, args);

            if (rs.hasNext()) {

                id = wrapDBID(rs.getSingleton());

                resourceRDB.setRDBId(id);

                return id;

            } else {

                return null;

            }

        } catch (SQLException e1) {

            // /* DEBUG */ System.out.println("Literal truncation (" + l.toString().length() + ") " + l.toString().substring(0, 150));

            throw new RDFRDBException("Failed to find resource", e1);

        }

    }



    /**

     * Register a literal

     * @return the db index of the added literal

     */

    protected IDBID addLiteral(Literal l) throws RDFRDBException {

        IDBID id = null;

        if (!SKIP_DUPLICATE_CHECK) {

            id = getLiteralID(l);

            if (id != null) return id;

        }

        try {

            int val = 0;               // Integer translation of l, if possible

            boolean isInt = false; // true if literal can be interpreted as an integer

            // Check if the literal can be translated to an int

            try {

                val = Integer.parseInt(l.toString());

                isInt = true;

            } catch (NumberFormatException e) {

                isInt = false;

            }

            String opname = isInt ? "insertLiteralInt" : "insertLiteral";

            if (INSERT_BY_PROCEDURE) opname = opname + "Proc";

            PreparedStatement ps = m_sql.getPreparedSQLStatement(opname);

            int argi = 1;

            if (!SKIP_ALLOCATE_ID) {

                id = allocateLiteralID(l);

                ps.setObject(argi++, id.getID());

            }

            String lit = l.toString();

            int len = lit.length();

            if (len >= MAX_LITERAL || len == 0) {

                // Literal too long to fit, insert a subset for indexing and put whole into a blob

                ps.setString(argi++, len == 0 ? EMPTY_LITERAL_MARKER : getLiteralIdx(lit));

                // First convert the literal to a UTF-16 encoded byte array

                // (this wouldn't be needed for jdbc 2.0 drivers but not all db's have them)

                byte[] temp = lit.getBytes("UTF-8");

                int lenb = temp.length;

                //System.out.println("utf-16 len = " + lenb);

                byte[] litData = new byte[lenb + 4];

                litData[0] = (byte)(lenb & 0xff);

                litData[1] = (byte)((lenb >> 8) & 0xff);

                litData[2] = (byte)((lenb >> 16) & 0xff);

                litData[3] = (byte)((lenb >> 24) & 0xff);

                System.arraycopy(temp, 0, litData, 4, lenb);

                ps.setBinaryStream(argi++, new ByteArrayInputStream(litData), litData.length);

            } else {

                ps.setString(argi++, lit);

                ps.setNull(argi++, Types.BLOB);

            }

            ps.setString(argi++, l.getLanguage());

            ps.setInt(argi++, l.getWellFormed() ? 1 : 0);

            if (isInt) {

                ps.setInt(argi++, val);

            }

            if (INSERT_BY_PROCEDURE) {

                ResultSet rs = ps.executeQuery();

                ResultSetIterator it = new ResultSetIterator(rs, ps, m_sql, opname);

                if (it != null && it.hasNext())

                    id = wrapDBID(it.getSingleton());

            } else {

                ps.executeUpdate();

                m_sql.returnPreparedSQLStatement(ps, opname);

            }

            if (id == null)

                id = getLiteralID(l);

            return id;

        } catch (Exception e1) {

            // /* DEBUG */ System.out.println("Problem on literal (l=" + l.toString().length() + "): " + l);

            /* DEBUG */ System.out.println("Problem on literal (l=" + l.toString().length() + ") ");

            /* DEBUG */ System.out.println("ID is: " + id);

            throw new RDFRDBException("Failed to register literal", e1);

        }

    }





}



/*

 *  (c) Copyright Hewlett-Packard Company 2000, 2001

 *  All rights reserved.

 *

 * Redistribution and use in source and binary forms, with or without

 * modification, are permitted provided that the following conditions

 * are met:

 * 1. Redistributions of source code must retain the above copyright

 *    notice, this list of conditions and the following disclaimer.

 * 2. Redistributions in binary form must reproduce the above copyright

 *    notice, this list of conditions and the following disclaimer in the

 *    documentation and/or other materials provided with the distribution.

 * 3. The name of the author may not be used to endorse or promote products

 *    derived from this software without specific prior written permission.



 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR

 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES

 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.

 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,

 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT

 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,

 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY

 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT

 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF

 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 */

