/* -*-Java-*-

*******************************************************************

*

* File:         IRDBSpec.java

* RCS:          $Header: /cvsroot/jena/jena/src/com/hp/hpl/mesa/rdf/jena/rdb/IRDBDriver.java,v 1.9 2002/09/04 08:47:38 der Exp $

* Author:       Dave Reynolds, HPLabs, Bristol

* Created:      28/8/01

* Modified:     $Date: 2002/09/04 08:47:38 $

* Status:       Experimental (Do Not Distribute)

*

* (c) Copyright 2001, Hewlett-Packard Company, all rights reserved.

* [See end of file]

*********************************************************************/



//=======================================================================

// Package

package com.hp.hpl.mesa.rdf.jena.rdb;



//=======================================================================

// Imports

import com.hp.hpl.mesa.rdf.jena.common.*;

import com.hp.hpl.mesa.rdf.jena.model.*;

import java.io.*;

import java.util.*;

import java.sql.*;

import java.util.Iterator;



//=======================================================================

/**

* Generic database interface used for implementing RDF Stores.

* Different database table layouts and different SQL dialects should all

* be supportable via this generic interface. Each driver instance can

* support multiple stores so long as "supportsMultipleModels" is true.

*

* @todo add <em>backdoor</em> functions to support queries.

* @author <a href="mailto:der@hplb.hpl.hp.com">Dave Reynolds</a>

* @version $Revision: 1.9 $ on $Date: 2002/09/04 08:47:38 $

*/



public interface IRDBDriver {



//=======================================================================

// Database operations



    /**

     * Link an existing instance of the driver to a specific jdbc connection

     */

    public void setConnection(IDBConnection dbcon) throws RDFRDBException;



    /**

     * Return the jdbc connection to the database, opening it if necessary

     */

    public Connection getConnection() throws RDFRDBException;



    /**

     * Close the databse connection

     * @throws RDFRDBException if there is an access problem

     */

    public void close() throws RDFRDBException;



    /**

     * Check that the database has a legal set of RDF tables.

     * @return true if the database seems to be legally formated.

     * @throws RDFRDBException if the is a problem opening the connection or an internal SQL error.

     */

    public boolean isDBFormatOK() throws RDFRDBException;



    /**

     * Initialise a database ready to store RDF tables.

     * Currently the table format depends on the RDBSpec type. In future it

     * may become an explicit part of operations like this.

     * @throws RDFRDBException if the is a problem opening the connection or an internal SQL error.

     */

    public void formatDB() throws RDFRDBException;



    /**

     * Remove all RDF information from a database.

     */

    public void cleanDB();



    /**

     * Allocate a new databaseID for a new model in the same database.

     * Throws RDFRDBException if the driver does not support multiple models per db.

     */

    public IDBID allocateModelID(String modelName) throws RDFRDBException;



    /**

     * Return the databaseID for an existing named model within the database.

     * Throws RDFRDBException if the driver does not support multiple models per db.

     */

    public IDBID getModelID(String modelName) throws RDFRDBException;



    /**

     * Delete the named from from within the database.

     * This will remove any statements associated with the model and

     * remove the model from the table of known models.

     * Throws RDFRDBException if the driver does not support multiple models per db.

     */

    public void deleteModel(String modelName) throws RDFRDBException;



    /**

     * List the names of all models in this database

     */

    public Iterator listModels() throws RDFRDBException;



    /**

     * Load a set of configuration parameters into the driver.

     */

    public void loadProperties(Properties props);



     /**

      * Set cache limits for caching resources and literals. In each case

      * a limit of 0 indicates no caching and a limit of -1 indicates

      * unlimited caching.

      * @param literalLimit limit on number of literals to cache

      * @param resouceLimit limit on number of resources to cache

      */

     public void setCacheLimits(int literalLimit, int resourceLimit);



     /**

      * Preload the resource and literal caches with all data in the DB.

      * Used for higher performance bulk loading in situations where the DB

      * is just used for persistance, not for volume scaling.

      */

     public void preloadCaches(ModelRDB m) throws RDFException;



    /**

     * Fetch a resource from the cache just knowing its resource rdb-id.

     * If the resource is not cached, does not attempt to load it from the database.

     */

    public Resource getResourceFromCache(IDBID id, ModelRDB m) throws RDFException;



    /**

     * Fetch a property from the cache just knowing its resource rdb-id.

     * If the resource is not cached, does not attempt to load it from the database.

     */

    public Property getPropertyFromCache(IDBID id, ModelRDB m) throws RDFException;



    /**

     * Fetch a literal from the cache just knowing its literal rdb-id.

     * If it is not in the cache, do not attempt to retrieve it from the database.

     */

    public Literal getLiteralFromCache(IDBID id);



    /**

     * Re-create a property from its local name and namespace identifier

     */

    public Property makeProperty(IDBID id, String localname, IDBID nsid, ModelRDB m) throws RDFException;



//=======================================================================

// Store operations



    /**

     * Register a store as a user of this driver.

     * One driver can support multiple store instances.

     */

    public void attachStore(StoreRDB store);



    /**

     * Notify that a store is no longer using this driver.

     */

    public void detachStore(StoreRDB store);



    /**

     * Register a statement in the database.

     */

    public void addStatement(com.hp.hpl.mesa.rdf.jena.model.Statement s, ModelRDB m) throws RDFException;



    /**

     * Delete the statements in the model which match the given subject,

     * predicate object.

     * @return true if the statement was successfully deleted

     */

    public boolean deleteStatement(Resource subject, Property predicate, RDFNode object, ModelRDB m) throws RDFException;



    /**

     * Delete the statement in the model corresponding to the given statement ID

     * @return true if the statement was successfully deleted

     */

    public boolean deleteStatement(IDBID statementID, ModelRDB m) throws RDFException;



    /**

     * List all the subject resources in the model

     * @return iterator over Resource objects

     */

    public ResultSetIterator listSubjects(ModelRDB m) throws RDFRDBException;



    /**

     * List all namespaces referenced in the model

     * @return iterator over Strings

     */

    public ResultSetIterator listNamespaces(ModelRDB m) throws RDFRDBException;



    /**

     * List all the statements in the model which match the given subject,

     * predicate object - each of which can be null (meaning wildcard).

     */

    public ResultSetIterator listStatements(Resource subject, Property predicate, RDFNode object, ModelRDB m) throws RDFException;



    /**

     * Check if a statement is in the database.

     */

    public boolean isStatementPresent(Resource subject, Property predicate, RDFNode object, ModelRDB m) throws RDFException;



    /**

     * List all the statements in the given moddel which make the given set of constraints

     * (a constraint set can be obtained using {@link #createConstraints}).

     */

    public ResultSetIterator listStatements(IConstraints constraints, ModelRDB m) throws RDFException;



    /**

     * Create a constraint object appropriate for this type of database layout

     */

    public IConstraints createConstraints(ModelRDB model);



    /**

     * Remove all the statements from the given model.

     * N.B. The implementation is not guaranteed to garbage collect resources and literal tables

     * but this fact is not detectable from the API.

     */

    public void clear(ModelRDB model) throws RDFRDBException;



//=======================================================================

// Misc



    /**

     * Returns true if the database layout supports multiple RDF models

     * in the same database.

     */

    public boolean supportsMultipleModels();



    /**

     * Returns true if the database layout supports implicit reification

     * of statements (i.e. statements can be treated as resources).

     */

    public boolean supportsJenaReification();



    /**

     * Return an iterator over all statements which are have been

     * added as reified statements.

     */

    public ResultSetIterator listReifiedStatements(ModelRDB m) throws RDFException;



    /**

     * Re-create a resource from its local name and namespace identifier

     */

    public Resource makeResource(IDBID id, String localname, IDBID nsidw, ModelRDB m) throws RDFException;



    /**

     * Re-create a resource from its local name and namespace identifier.

     * If stmtID is not null then the resource is actually a reified statement

     * and with that as its id.

     */

    public Resource makeResource(IDBID id, String localname, IDBID nsid, ModelRDB m, IDBID stmtID) throws RDFException;



    /**

     * Convert the raw SQL object used to store a database identifier into a java object

     * which meets the IDBID interface

     */

    public IDBID wrapDBID(Object id) throws RDFException;



    /**

     * Fetch a resource just knowing its resource rdb-id.

     */

    public Resource getResource(IDBID id, ModelRDB m) throws RDFException;



    /**

     * Fetch a reiefied statement resource just knowing its resource rdb-id.

     * Can be null if the ID can from a hash function and resource isn't registered yet.

     * This variant will never attempt to instantiate the contents of the Statement

     * (as does getResource) - if the statement is not cached then an empty StatementImplRDB

     * will be returned.

     */

    public StatementImplRDB getBareStatementResource(IDBID id, ModelRDB m) throws RDFException;



    /**

     * Fetch a property just knowing its resource rdb-id.

     */

    public Property getProperty(IDBID id, ModelRDB m) throws RDFException;



    /**

     * Fetch a literal just knowing its literal rdb-id

     */

    public Literal getLiteral(IDBID id) throws RDFException;



    /**

     * Return the database ID for the literal

     */

    public IDBID getLiteralID(Literal l) throws RDFRDBException;



    /**

     * Lookup the unique ID for a resource.

     */

    public IDBID getResourceID(Resource resource) throws RDFException;



    /**

     * Return the namespace string corresponding to a given namespace id

     */

    public String getNamespace(IDBID nsid) throws RDFRDBException;



    /**

     * Return the SQLCache used for low level database commands.

     */

    SQLCache getSQLCache();



    /**

     * Fetch a resource, knowing its uri. If the resource does not yet exist in the database it is added

     */

    public Resource getResource(String uri, ModelRDB m) throws RDFException;



    /**

     * Fetch a property, knowing its uri. If the property does not yet exist in the database it is added

     */

    public Property getProperty(String uri, ModelRDB m) throws RDFException;



    /**

     * Hint to the store that the given property could be stored as an

     * attribute table.

     */

    public void hintPropertyTable(Property p) throws RDFException;

}



/*

 *  (c) Copyright Hewlett-Packard Company 2000, 2001

 *  All rights reserved.

 *

 * Redistribution and use in source and binary forms, with or without

 * modification, are permitted provided that the following conditions

 * are met:

 * 1. Redistributions of source code must retain the above copyright

 *    notice, this list of conditions and the following disclaimer.

 * 2. Redistributions in binary form must reproduce the above copyright

 *    notice, this list of conditions and the following disclaimer in the

 *    documentation and/or other materials provided with the distribution.

 * 3. The name of the author may not be used to endorse or promote products

 *    derived from this software without specific prior written permission.



 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR

 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES

 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.

 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,

 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT

 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,

 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY

 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT

 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF

 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 */