/* -*-Java-*-
*******************************************************************
*
* File:         ModelRDB.java
* RCS:          $Header: /cvsroot/jena/jena/src/com/hp/hpl/mesa/rdf/jena/rdb/ModelRDB.java,v 1.12 2002/09/04 08:47:38 der Exp $
* Author:       Dave Reynolds, HPLabs, Bristol
* Created:      28/8/01
* Modified:     $Date: 2002/09/04 08:47:38 $
* Status:       Experimental (Do Not Distribute)
*
* (c) Copyright 2001, Hewlett-Packard Company, all rights reserved.
* [See end of file]
*********************************************************************/

//=======================================================================
// Package
package com.hp.hpl.mesa.rdf.jena.rdb;

//=======================================================================
// Imports
import com.hp.hpl.mesa.rdf.jena.common.*;
import com.hp.hpl.mesa.rdf.jena.model.*;
import com.hp.hpl.mesa.rdf.jena.mem.*;
import java.io.*;
import java.util.*;

//=======================================================================
/**
* This implementation of the Model interface uses a relational database
* to hold the model statements. This provides persistence, transactions and
* scaling at the expense of performance. It doesn't provide query support
* since the jena query interface is currently procedural rather than
* declarative.
* <p>
* Different SQL dialects and different table layout structures are supported
* by delegating all database access to a separate IRDBSpec implementation.
* <p>
* To reopen an existing database use the open() static factory methods.
* <br>
* To create a database for the first time use the create() static factory
* methods.
*
* @author <a href="mailto:der@hplb.hpl.hp.com">Dave Reynolds</a>
* @version $Revision: 1.12 $ on $Date: 2002/09/04 08:47:38 $
*/

public class ModelRDB extends ModelCom implements Model, ModelI {

    /** The name of the model, used for layouts that support multiple models in one db */
    protected String m_name = "default";

    /** The database index number of the model, used for layouts that support multiple models in one db */
    protected IDBID m_dbid;

     /**
     * Return the default model from the specified database. If the database
     * does not appear to be formated it formats it and creates an empty default model.
     * @param dbcon a DBConnectionI specifying the database connection
     * @throws RDFRDBException if the database cannot be opened
     */
    public ModelRDB(IDBConnection dbcon) throws RDFRDBException {
        this(new StoreRDB(dbcon.getDriver()));
    }

    /**
     * Construct a Model which uses the given store interface.
     * Used to support static factory methods, not for public consumption.
     */
    protected ModelRDB(Store store) throws RDFRDBException {
        this.store = store;
        if (store instanceof StoreRDB) {
            ((StoreRDB)store).attachModel(this);
        }
        m_dbid = ((StoreRDB)store).getDriver().allocateModelID(m_name);
    }

    /**
     * Construct a Model which uses the given store interface.
     * Used to support static factory methods, not for public consumption.
     */
    protected ModelRDB(Store store, String name, IDBID dbid) {
        this.store = store;
        m_name = name;
        m_dbid = dbid;
        if (store instanceof StoreRDB) {
            ((StoreRDB)store).attachModel(this);
        }
    }

    /**
     * Open an existing rdf database. The layout and datatype type information
     * will be dynamically loaded from the database.
     * Will throw an RDFRDBException if the database layout does not support
     * multiple models or if the database does not seem to formated.
     * @param dbcon a DBConnectionI specifying the database connection
     * @param name the name of the RDF model to open, some database layouts support
     * multiple models in a single database
     */
    public static ModelRDB open(IDBConnection dbcon, String name) throws RDFRDBException {
        IRDBDriver driver = dbcon.getDriver();
        if (driver == null) throw new RDFRDBException("Failed to open RDB driver, probably an unformatted database");
        IDBID dbid = driver.getModelID(name);
        return new ModelRDB(new StoreRDB(driver), name, dbid);
    }

    /**
     * Open an existing rdf database. The layout and datatype type information
     * will be dynamically loaded from the database. If the database supports more
     * than one model then the "default" model will be opended.
     * @param dbcon a DBConnectionI specifying the database connection
     */
    public static ModelRDB open(IDBConnection dbcon) throws RDFRDBException {
        IRDBDriver driver = dbcon.getDriver();
        if (driver == null) throw new RDFRDBException("Failed to open RDB driver, probably an unformatted database");
        return new ModelRDB(StoreRDB.openDB(driver));
    }

    /**
     * Create a new model on an existing, preformatted database.
     * Will throw an RDFRDBException if the database layout does not support
     * multiple models or if the database does not seem to formated.
     * @param dbcon a DBConnectionI specifying the database connection
     * @param name the name to give the newly created model
     */
    public static ModelRDB createModel(IDBConnection dbcon, String name) throws RDFRDBException {
        IRDBDriver driver = dbcon.getDriver();
        if (driver == null) throw new RDFRDBException("Failed to open RDB driver, probably an unformatted database");
        IDBID dbid = driver.allocateModelID(name);
        return new ModelRDB(new StoreRDB(driver), name, dbid);
    }

    /**
     * Create a named model from an existing multi-model database.
     * Will throw an RDFRDBException if the database layout does not support
     * multiple models or if the database does not seem to formated.
     * @param dbcon a DBConnectionI specifying the database connection
     * @param name the name to give the newly created model
     */
    public static void deleteModel(IDBConnection dbcon, String name) throws RDFRDBException {
        IRDBDriver driver = dbcon.getDriver();
        if (driver == null) throw new RDFRDBException("Failed to open RDB driver, probably an unformatted database");
        driver.deleteModel(name);
    }

    /**
     * List the name of all models stored in the database
     */
    public static java.util.Iterator listModels(IDBConnection dbcon) throws RDFRDBException {
        IRDBDriver driver = dbcon.getDriver();
        if (driver == null) throw new RDFRDBException("Failed to open RDB driver, probably an unformatted database");
        return driver.listModels();
    }

    /**
     * Create a new database suitable for storing RDF data. In fact the database has
     * to exist since jdbc can't create an empty database from a vacuum but it can be empty
     * and this call will format it with appropriate tables and stored procedures.
     * <p>
     * The appropriate RDF-RDB driver to use is assumed to be the class Driver<DatabaseType><LayoutType>.
     * If that can't be found it defaults to looking for a property file in /etc/Driver<DatabaseType><LayoutType>.config
     * and uses that to determine the driver class and parameters.
     *
     * @param dbcon a DBConnectionI specifying the database connection
     * @param layoutType the name of the layout style to use. Currently one of:
     * "Generic", "Hash", "MMGeneric", "MMHash", "Proc", "ThinProc".
     * @param databaseType the name of the database type. Currently one of:
     * "Interbase" "Postgresql" "Mysql" "Oracle". This may seem a little redundant
     * given that the jdbc uri implicitly contains this information but there is no
     * standard way of extracting this (esp. if the user connects via a bridge).
     */
    public static ModelRDB create(IDBConnection dbcon, String layoutType, String databaseType) throws RDFRDBException {
        IRDBDriver driver = dbcon.getDriver(layoutType, databaseType);
        if (driver.isDBFormatOK()) {
            // Database already formated so reload configuration properties from there
            // and don't rely on the getDriver() results since layouts may have changed
            dbcon.reloadProperties();
        }
        return new ModelRDB(StoreRDB.createDB(driver));
    }

    /**
     * Create a new database suitable for storing RDF data. In fact the database has
     * to exist since jdbc can't create an empty database from a vacuum but it can be empty
     * and this call will format it with appropriate tables and stored procedures.
     * <p>
     * Uses a default layout format which is able to support multiple models in a single database.
     *
     * @param databaseType the name of the database type. Currently one of:
     * "Interbase" "Postgresql" "Mysql" "Oracle". This may seem a little redundant
     * given that the jdbc uri implicitly contains this information but there is no
     * standard way of extracting this (esp. if the user connects via a bridge).
     */
    public static ModelRDB create(IDBConnection dbcon, String databaseType) throws RDFRDBException {
        IRDBDriver driver = dbcon.getDriver("MMGeneric", databaseType);
        return new ModelRDB(StoreRDB.createDB(driver));
    }

    /**
     * Close the Model and free up the resources held.
     * <p>This will discard the underlying StoreRDB object which in turn.
     * Multiple models can share a single underlying DBConnection object (when
     * the database layout is type "MMxxx") so it is only when the last ModelRDB
     * associated with a driver is closed that the driver itself will
     * be closed which will close.
     * <p>In earlier releases the final driver close resulted in closing the
     * DBConnection which in turn closed the jdbc connection. This proved unpopular
     * and has been removed. The close will free up associated prepared statements
     * but the DBConnection will remain open so you should close that explicitly if
     * you wish to shut down the jdbc connection.
     */
    public void close() {
        store.close();
        store = null;
    }

    /**
     * Finalizer to clean up the database collections
     * Nullified now that close does not close the JDBC connection.
     */
    protected void finalize() {
        //close();
    }

    /**
     * Return the store which underpins this model
     */
    public StoreRDB getStore() {
        return (StoreRDB) store;
    }

    /**
     * Return the database index id for this model.
     * This will be 0 unless the underlying database layout supports
     * multiple models per database.
     */
    public IDBID getDBID() {
        return m_dbid;
    }

    /**
     * Remove all the statements from the database which are associated with just this model.
     * This no longer reformats the database (which makes it safer and useful for multi-model
     * databases) but
     * means that it is not guaranteed to garbage collect the resource table.
     */
    public void clear() throws RDFRDBException {
        ((StoreRDB)store).clear();
    }

    /**
     * Loads all the statements for this model into an in-memory model.
     * @return a ModelMem containing the whole of the RDB model
     */
    public Model loadAll() throws RDFException {
        ModelMem m = new ModelMem();
        // Preload resources and literals
        getStore().getDriver().preloadCaches(this);
        // Load all statements but don't move them to the model until all are in
        // to avoid sequence problems with reified statements
        ArrayList statementCache = new ArrayList();
        for (StmtIterator i = this.listStatements(); i.hasNext(); ) {
            statementCache.add(i.next());
        }
        // Move the loaded statements to the memory model
        for (Iterator j = statementCache.iterator(); j.hasNext(); ) {
            m.add((Statement)j.next());
        }
        return m;
    }

//=======================================================================
// Extra operations to exploit db indexing

    /**
     * List all the statements in the given moddel which make the given set of constraints
     * (a constraint set can be obtained using {@link #createConstraints}).
     */
    public StmtIterator listStatements(IConstraints constraints) throws RDFException {
        return new StmtIteratorImpl(((StoreRDB)store).listStatements(constraints, this), null);
    }

    /**
     * Create a constraint object appropriate for this type of database layout
     */
    public IConstraints createConstraints() {
        return ((StoreRDB)store).createConstraints(this);
    }


//=======================================================================
// ModelI interface for converting Resource types

    /**
     * Convert a resource to a database implementation
     */
    public Resource convert(Resource r) throws RDFException {
        if (r == null) return null;
        if (r.getModel() == this && (r instanceof ResourceImplRDB || r instanceof StatementImplRDB)) {
            return r;
        } else if (supportsJenaReification() && r instanceof Statement) {
            return StatementImplRDB.createFrom((Statement)r, this);
        } else {
            return new ResourceImplRDB(r, this);
        }
    }

    /**
     * Convert a property to a database implementation
     */
    public Property convert(Property p) throws RDFException {
        if (p == null) return null;
        if (p.getModel() == this && p instanceof PropertyImplRDB) {
            return p;
        } else {
            return new PropertyImplRDB(p, this);
        }
    }

    /**
     * Convert a node (property, resource or literal) to a database implementation
     */
    public RDFNode convert(RDFNode n) throws RDFException {
        if (n == null) return null;
        if (n instanceof Property) {
            return convert((Property) n);
        } else if (n instanceof Resource) {
            return convert((Resource) n);
        } else {
            return n;
        }
    }

//=======================================================================
// Reification support

    /**
     * Returns true if the database layout supports implicit reification
     * of statements (i.e. statements can be treated as resources).
     */
    public boolean supportsJenaReification() {
        return ((StoreRDB)store).supportsJenaReification();
    }

    /**
     * list all statements which are present in reified form
     */
    public StmtIterator listReifiedStatements() throws RDFException {
        if (supportsJenaReification()) {
            return new StmtIteratorImpl(((StoreRDB)store).listReifiedStatements(this), null);
        } else {
            return super.listReifiedStatements();
        }
    }

    /** Add a statement to the model */
    public Model add(Statement s) throws RDFException {
        if (! (s instanceof StatementImplRDB)
          || (((StatementImplRDB)s).getModel() != this) ) {
            s = StatementImplRDB.createFrom(s, this);
        }
        store.add(s);
        return this;
    }


    /** Test if the model contains a triple */
    public boolean contains(Resource s, Property p, RDFNode o)
      throws RDFException {
       return contains(new StatementImplRDB(s, p, o));
    }

    /** create a new statement, not yet added to the modl */
    public Statement createStatement(Resource r, Property p, RDFNode o)
    throws RDFException {
        return new StatementImplRDB(r, p, o, this);
    }

//=======================================================================
// Transaction support

/** Begin a new transation.
 *
 * <p> All changes made to a model within a transaction, will either
 * be made, or none of them will be made.</p>
 * @return this model to enable cascading.
 * @throws RDFException Generic RDF Exception
 */
    public Model begin() throws RDFException {
        ((StoreRDB)store).begin();
        return this;
    }

/** Abort the current transaction and abandon any changes in progress.
 * @return this model to enable cascading.
 * @throws RDFException Generic RDF Exception
 */
    public Model abort() throws RDFException {
        ((StoreRDB)store).abort();
        return this;
    }

/** Commit the current transaction.
 * @return this model to enable cascading.
 * @throws RDFException Generic RDF Exception
 */
    public Model commit() throws RDFException {
        ((StoreRDB)store).commit();
        return this;
    }

/** Determine whether this model is independent.
 *
 *  <p>For efficiency reasons, some implementations may create models which
 *  which are dependent on others, i.e. a change in one model may cause
 *  a change in another.  If this is the case this method will return false,
 *  otherwise it will return true.</p>
 *
 * @return  true if this model is indepdent of others
 */
    public boolean independent() {
        return true;
    }

/** Determine whether this model supports transactions.
 * @return  true if this model supports transactions.
 */
    public boolean supportsTransactions() {
        return true;
    }

//=======================================================================
// Below here is boiler plate code copied from ModelMem.
// This is just temporary. A real implementation would use the database
// for implementing the set operations but this is only possible if the
// database layout supports multiple models.

    public boolean supportsSetOperations() {
        return true;
    }

    public Model query(Selector selector) throws RDFException {
        ModelMem model = new ModelMem();
        StmtIterator iter = null;
        try {
            iter = listStatements(selector);
            while (iter.hasNext()) {
                model.add(iter.next());
            }
            return model;
        } finally {
            iter.close();
        }
    }

    public Model union(Model model) throws RDFException {
        return (new ModelMem()).add(this)
                               .add(model);
    }

    public Model intersection(Model model) throws RDFException {
        Model largerModel = this;
        Model smallerModel  = model;
        ModelMem resultModel = new ModelMem();
        StmtIterator iter = null;
        Statement stmt;

        if (model.size() > this.size()) {
            largerModel = model;
            smallerModel = this;
        }
        try {
            iter = smallerModel.listStatements();
            while (iter.hasNext()) {
                stmt = iter.next();
                if (largerModel.contains(stmt)) {
                    resultModel.add(stmt);
                }
            }
            return resultModel;
        } finally {
            iter.close();
        }
    }

    public Model difference(Model model) throws RDFException {
        ModelMem resultModel = new ModelMem();
        StmtIterator iter = null;
        Statement stmt;
        try {
            iter = listStatements();
            while (iter.hasNext()) {
                stmt = iter.next();
                if (! model.contains(stmt)) {
                    resultModel.add(stmt);
                }
            }
            return resultModel;
        } finally {
            iter.close();
        }
    }

}

/*
 *  (c) Copyright Hewlett-Packard Company 2000, 2001
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */