/* -*-Java-*-
*******************************************************************
*
* File:         ResultSetStatementIterator.java
* RCS:          $Header: /cvsroot/jena/jena/src/com/hp/hpl/mesa/rdf/jena/rdb/ResultSetStatementIterator.java,v 1.5 2002/09/04 08:47:38 der Exp $
* Author:       Dave Reynolds, HPLabs, Bristol
* Created:      11/9/01
* Modified:     $Date: 2002/09/04 08:47:38 $
* Status:       Experimental (Do Not Distribute)
*
* (c) Copyright 2001, Hewlett-Packard Company, all rights reserved.
* [See end of file]
*********************************************************************/

//=======================================================================
// Package
package com.hp.hpl.mesa.rdf.jena.rdb;

//=======================================================================
// Imports
import com.hp.hpl.mesa.rdf.jena.model.*;
import com.hp.hpl.mesa.rdf.jena.common.LiteralImpl;
import com.hp.hpl.mesa.rdf.jena.common.StatementImpl;
import com.hp.hpl.jena.util.Log;
import java.sql.*;

//=======================================================================
/**
* Version of ResultSetIterator that extracts database rows as statments
* assuming that the SQL returns rows of form:
* <pre>
*    subject [sid, slname, snsid]
*    predicate [pid, plname, pnsid]
*    object [oid, olitid, oislit]
* </pre>
* It would be simpler to just return resource and literal ids
* and issue separate calls to retrieve those instances which are not
* currently cached. Early experiments show this as x3 slower when the cache
* misses (which is likely to be often).
*
* @author <a href="mailto:der@hplb.hpl.hp.com">Dave Reynolds</a>
* @version $Revision: 1.5 $ on $Date: 2002/09/04 08:47:38 $
*/
public class ResultSetStatementIterator extends ResultSetIterator {

    /** The rdf model in which to instantiate any resources */
    protected ModelRDB m_model;

    /** The database driver, used to access namespace and resource caches */
    protected IRDBDriver m_driver;

    /** The number of rows in the result set (-1 if don't yet know) */
    protected int m_nrows = -1;

    /** Over-ride row data type */
    protected StatementImplRDB m_row;

    /**
     * Create an iterator which which convert SQL result rows into RDF statements.
     * @param model The rdf model in which to instantiate any resources
     * @param driver The database driver, used to access namespace and resource caches
     */
    public ResultSetStatementIterator(ModelRDB model, IRDBDriver driver) {
        m_model = model;
        m_driver = driver;
    }

    /**
     * Extract the current row
     */
    protected void extractRow() throws SQLException, RDFException {
        // Check if this is a wide row or a narrow row
        if (m_nrows == -1) {
            m_nrows = m_resultSet.getMetaData().getColumnCount();
        }
        boolean wideRow = m_nrows >= 8;
        int rx = 1;     // result column counter
        // Get subject
        Resource subject = null;
        IDBID id = m_driver.wrapDBID(m_resultSet.getObject(rx++));
        if (wideRow) {
            String localname = m_resultSet.getString(rx++);
            IDBID nsid = m_driver.wrapDBID(m_resultSet.getObject(rx++));
            if (subject == null) {
                if (m_model.supportsJenaReification()) {
                    IDBID stmtID = m_driver.wrapDBID(m_resultSet.getObject(rx++));
                    subject = m_driver.makeResource(id, localname, nsid, m_model, stmtID);
                } else {
                    subject = m_driver.makeResource(id, localname, nsid, m_model);
                }
            }
        } else {
            subject = m_driver.getResource(id, m_model);
        }
        // Get predicate
        Property predicate = null;
        id = m_driver.wrapDBID(m_resultSet.getObject(rx++));
        if (wideRow) {
            String localname = m_resultSet.getString(rx++);
            IDBID nsid = m_driver.wrapDBID(m_resultSet.getObject(rx++));
            predicate = m_driver.makeProperty(id, localname, nsid, m_model);
        } else {
            predicate = m_driver.getProperty(id, m_model);
        }
        // Get object
        RDFNode object = null;
        IDBID literalId = m_driver.wrapDBID(m_resultSet.getObject(rx++));
        int isLiteral = m_resultSet.getInt(rx++);
        if (isLiteral == 1) {
            object = m_driver.getLiteral(literalId);
        } else {
            object = m_driver.getResource(literalId, m_model);
        }
        if (m_model.supportsJenaReification()) {
            // Pick up the extra reification info - statementID, isReified, reification-resource
            IDBID thisID = m_driver.wrapDBID(m_resultSet.getObject(rx++));
            int isOnlyReified = m_resultSet.getInt(rx++);
            Object reification = m_resultSet.getObject(rx++);
            if (reification != null) {
                // The statement points to a resource entry that reifies it
                // need to retrieve that to get the correct URI/anonid info
                m_row = m_driver.getBareStatementResource(m_driver.wrapDBID(reification), m_model);
                m_row.setSPO(subject, predicate, object);
            } else {
                m_row = new StatementImplRDB(subject, predicate, object, m_model);
            }
            m_row.setRDBStatementId(thisID);
            m_row.setIsReifiedOnly(isOnlyReified == 1);
        } else {
            m_row = new StatementImplRDB(subject, predicate, object, m_model);
        }
    }

    /**
     * Return the current row,should have already been extracted.
     * Override in subclasses.
     */
    protected Object getRow() {
        return m_row;
    }

    /**
     * Delete the current row entry.
     * This version is for databases (like InterBase) whose jdbc driver doesn support concurrently updatable cursors.
     */
    public void remove() {
        if (!m_finished && m_row != null) {
            try {
                m_model.remove(m_row);
            } catch (RDFException e) {
                throw new IllegalStateException("Error during remove(): " + e);
            }
        }
    }

} // End class

/*
 *  (c) Copyright Hewlett-Packard Company 2000, 2001
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */





