/* -*-Java-*-
*******************************************************************
*
* File:         StatementImplRDB.java
* RCS:          $Header: /cvsroot/jena/jena/src/com/hp/hpl/mesa/rdf/jena/rdb/StatementImplRDB.java,v 1.2 2002/09/04 08:47:38 der Exp $
* Author:       Dave Reynolds, HPLabs, Bristol
* Created:      23/6/02
* Modified:     $Date: 2002/09/04 08:47:38 $
* Status:       Experimental (Do Not Distribute)
*
* (c) Copyright 2001, Hewlett-Packard Company, all rights reserved.
* [See end of file]
*********************************************************************/

//=======================================================================
// Package
package com.hp.hpl.mesa.rdf.jena.rdb;

//=======================================================================
// Imports
import com.hp.hpl.mesa.rdf.jena.common.*;
import com.hp.hpl.mesa.rdf.jena.model.*;
import com.hp.hpl.jena.util.Log;
import java.io.*;

//=======================================================================
/**
* A variation on the default Statement implementation supports
* reference to a reified version via an RDB resource ID.
*
* @author <a href="mailto:der@hplb.hpl.hp.com">Dave Reynolds</a>
* @version $Revision: 1.2 $ on $Date: 2002/09/04 08:47:38 $
*/

public class StatementImplRDB extends StatementImpl implements HasRDBID {

    /** Unique ID for resource corresponding to this statement */
    protected IDBID m_rdbId = null;

    /** Unique ID for the resource's namespace */
    protected IDBID m_rdbNSId = null;

    /** Unique ID for the statement itself */
    protected IDBID m_rdbStatementId = null;

    /** flag to indicate that the statement is only reified and not asserted */
    protected boolean isOnlyReified = false;

    /** Creates new StatementImpl */
    public StatementImplRDB(Resource subject, Property predicate, RDFNode object) {
        super(subject, predicate, object);
    }

    /** Creates new StatementImpl */
    public StatementImplRDB(Resource subject,
                         Property predicate,
                         RDFNode object,
                         Model model) throws RDFException {
        super(subject, predicate, object, model);
    }

    /** Creates new StatementImpl */
    public StatementImplRDB(Resource subject,
                         Property predicate,
                         RDFNode object,
                         AnonId id,
                         Model model) throws RDFException {
        super(subject, predicate, object, id, model);
    }

    /** Creates new StatementImpl */
    public StatementImplRDB(Resource subject,
                         Property predicate,
                         RDFNode  object,
                         String   URI,
                         Model    model) throws RDFException {
        super(subject, predicate, object, URI, model);
    }

    /**
     * Factory method to copy a Statement with all its parameters including
     * uri/anonid to an RDB statement
     * @param s the statement to be copied
     * @param m the model to which the new statement should be attached
     */
    public static StatementImplRDB createFrom(Statement s, Model m) throws RDFException {
        if (s.isAnon()) {
            return new StatementImplRDB(s.getSubject(), s.getPredicate(), s.getObject(),
                                        s.getId(), m);
        } else {
            return new StatementImplRDB(s.getSubject(), s.getPredicate(), s.getObject(),
                                        s.getURI(), m);
        }
    }

    /**
     * Return the unique ID for the statement itself
     */
    public IDBID getRDBStatementId() {
        return m_rdbStatementId;
    }

    /**
     * Set the unique ID for the statement itself
     */
    public void setRDBStatementId(IDBID id) {
        m_rdbStatementId = id;
    }

    /**
     * Return the unique ID for the resource associated with this statement.
     */
    public IDBID getRDBId() {
        return m_rdbId;
    }

    /**
     * Set the unique ID for the resource associated with this statement
     */
    public void setRDBId(IDBID id) {
        m_rdbId = id;
    }

    /**
     * Return the unique ID for the resource's namespace or null for resources
     * which have not yet be assigned an ID.
     */
    public IDBID getRDBNSId() {
        return m_rdbNSId;
    }

    /**
     * Set the unique ID for the resource's namespace
     */
    public void setRDBNSId(IDBID id) {
        m_rdbNSId = id;
    }

    /**
     * get flag to indicate that the statement is only reified and not asserted
     */
    public boolean getIsReifiedOnly() {
        return isOnlyReified;
    }

    /**
     * set flag to indicate that the statement is only reified and not asserted
     */
    public void setIsReifiedOnly(boolean isOnlyReified) {
        this.isOnlyReified = isOnlyReified;
    }

    /**
     * Change the uri/anonid assigned to the statement to match that
     * of the given resource
     */
    public void setURI(Resource r) {
        if (r.isAnon()) {
            try{
                this.id = r.getId();
            } catch (RDFException e) {
                // This is one of those "can't happen" errors ...
                Log.finest("error during statement setURI", e);
            }
        } else {
            this.nameSpace = r.getNameSpace();
            this.localName = r.getLocalName();
            this.uri = r.getURI();
        }
    }

    /**
     * Change the s/p/o values bound into this statement.
     * This is only used during bulk loading to fix up reified statements
     * left dangling by the preloader.
     */
    protected void setSPO(Resource subject, Property predicate, RDFNode object) {
        this.subject = subject;
        this.predicate = predicate;
        this.object = object;
    }

} // End class

/*
 *  (c) Copyright Hewlett-Packard Company 2000, 2001
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
