/* -*-Java-*-
*******************************************************************
*
* File:         StoreRDB.java
* RCS:          $Header: /cvsroot/jena/jena/src/com/hp/hpl/mesa/rdf/jena/rdb/StoreRDB.java,v 1.6 2002/08/09 12:12:50 der Exp $
* Author:       Dave Reynolds, HPLabs, Bristol
* Created:      28/8/01
* Modified:     $Date: 2002/08/09 12:12:50 $
* Status:       Experimental (Do Not Distribute)
*
* (c) Copyright 2001, Hewlett-Packard Company, all rights reserved.
* [See end of file]
*********************************************************************/

//=======================================================================
// Package
package com.hp.hpl.mesa.rdf.jena.rdb;

//=======================================================================
// Imports
import com.hp.hpl.mesa.rdf.jena.common.*;
import com.hp.hpl.mesa.rdf.jena.model.*;
import java.io.*;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.*;
import com.hp.hpl.jena.util.Log;

//=======================================================================
/**
* Generic store implementation for RDB backed RDF storage.
* The underlying database operations are implemented by separate driver
* classes which conform to the
* {@link com.hp.hpl.mesa.rdf.jena.rdb.IRDBDriver IRDBDriver}
* interface. Different driver instances may support different database table
* layouts as well as different specific databases.
*
* <p>A single driver instance can be shared by several StoreRDB
* instances whereas each ModelRDB requires a separate StoreRDB instance.
*
* @todo add <em>backdoor</em> functions to support queries.
* @author <a href="mailto:der@hplb.hpl.hp.com">Dave Reynolds</a>
* @version $Revision: 1.6 $ on $Date: 2002/08/09 12:12:50 $
*/

public class StoreRDB extends StoreCom implements Store {

//=======================================================================
//  Variables

    /** The database driver which implements the physical layout for the given db */
    protected IRDBDriver m_dbDriver;

    /** The model to which this store is attached */
    protected ModelRDB m_model;

//=======================================================================
// Constructors and database set up functions

    /**
     * Construct an Store handler for storing Jena models in an RDB.
     * This is not normally called by application programs but directly
     * via
     * {@link #createDB createDB}
     * or
     * {@link #openDB openDB}
     *
     * @param driver the RDB driver specification class which defines both
     * the table layout scheme and the database specific commands.
     * @param connection the open jdbc connection to an initialized database.
     */
    protected StoreRDB(IRDBDriver driver) {
        m_dbDriver = driver;
        m_dbDriver.attachStore(this);
    }

    /**
     * Construct a new Store using the given database spec. If the database
     * exists but does not seem to be formated for RDF storage new tables
     * will be created.
     * @param driver defines the database driver, host machine, login details
     * and root path for the databases.
     * @return the RDB-based Store
     * @throws RDFRDBException if the database cannot be opened
     */
    public static StoreRDB createDB(IRDBDriver driver) throws RDFRDBException {
        if (!driver.isDBFormatOK()) {
            driver.formatDB();
        }
        return new StoreRDB(driver);
    }

    /**
     * Open a named database and check that it is formated for RDF storage.
     * @param driverpec defines the database driver, host machine, login details
     * and root path for the databases.
     * @return the RDB-based Store
     * @throws RDFRDBException if the database cannot be opened or if once opened it does
     * not seem to contain a valid set of RDF storage tables.
     */
    public static StoreRDB openDB(IRDBDriver driver) throws RDFRDBException {
        if (!driver.isDBFormatOK()) {
            throw new RDFRDBException("Database is not formating for RDF storage, try createDB");
        }
        return new StoreRDB(driver);
    }

    /**
     * Note the model which this store is associated with.
     */
    public void attachModel(ModelRDB m) {
        m_model = m;
    }

    /**
     * Return the database driver underpining this store
     */
    public IRDBDriver getDriver() {
        return m_dbDriver;
    }

    /**
     * Close this store
     */
    public void close() {
        if (m_dbDriver != null) {
            m_dbDriver.detachStore(this);
            m_dbDriver = null;
        }
    }

    /**
     * Finalizer to clean up the database collections.
     * Nullified now that close does not close the JDBC connection.
     */
    protected void finalize() {
        //close();
    }

    /**
     * Remove all the statements from the database.
     * This no longer reformats the database (which makes it safer) but
     * means that it is not guaranteed to garbage collect the resource table.
     */
    public void clear() throws RDFRDBException {
        m_dbDriver.clear(m_model);
        //m_dbDriver.cleanDB();
        //m_dbDriver.formatDB();
    }


//=======================================================================
// Transaction support

/** Begin a new transation.
 *
 * <p> All changes made to a model within a transaction, will either
 * be made, or none of them will be made.</p>
 * @return this model to enable cascading.
 * @throws RDFException Generic RDF Exception
 */
    public void begin() throws RDFException {
        try {
            m_dbDriver.getSQLCache().flushPreparedStatementCache();
            Connection c = m_dbDriver.getConnection();
            c.setAutoCommit(false);
            c.setTransactionIsolation(Connection.TRANSACTION_READ_COMMITTED);
        } catch (SQLException e) {
            throw new RDFRDBException("Transaction support failed: ", e);
        }
    }

/** Abort the current transaction and abandon any changes in progress.
 * @return this model to enable cascading.
 * @throws RDFException Generic RDF Exception
 */
    public void abort() throws RDFException {
        try {
            Connection c = m_dbDriver.getConnection();
            c.rollback();
            c.setAutoCommit(true);
        } catch (SQLException e) {
            throw new RDFRDBException("Transaction support failed: ", e);
        }
    }

/** Commit the current transaction.
 * @return this model to enable cascading.
 * @throws RDFException Generic RDF Exception
 */
    public void commit() throws RDFException {
        try {
            Connection c = m_dbDriver.getConnection();
            c.commit();
            c.setAutoCommit(true);
        } catch (SQLException e) {
            throw new RDFRDBException("Transaction support failed: ", e);
        }
    }

//=======================================================================
// Store interface implementation

    /** Create a new resource implementation for a model */
    public Resource createResource(Model m) throws RDFException {
        return new ResourceImplRDB(m);
    }

    /** Create a new resource implementation for a model */
    public Resource createResource(String uri,Model m) throws RDFException {
        return new ResourceImplRDB(uri, m);
    }

    /** Create a new property resource implementation for a model */
    public Property createProperty(String uri, Model m) throws RDFException {
          return new PropertyImplRDB(uri, m);
    }

    /** Create a new property resource implementation for a model */
    public Property createProperty(String namespace,String localPart,Model m)
      throws RDFException {
          return new PropertyImplRDB(namespace, localPart, m);
    }

    /** create a new statement implementation for a model */
    public Statement createStatement(Resource s, Property p, RDFNode o, Model m)
      throws RDFException {
          return new StatementImplRDB(s, p, o, m);
    }


    /** Fetch a resource, adding it to the database if necessary */
    public Resource getResource(String uri,Model m) throws RDFException {
        return m_dbDriver.getResource(uri, (ModelRDB)m);
    }

    /** Fetch a property, adding it to the database if necessary */
    public Property getProperty(Property p) throws RDFException {
        return m_dbDriver.getProperty(p.getURI(), m_model);
    }

    /**
     * Add a new statement to the model and load it to the database
     * @todo might need to refactor most of this into the driver since ID labeling might be layout dependent
     */
    public synchronized void add(Statement s) throws RDFException {
        m_dbDriver.addStatement(convertStatement(s), m_model);
    }

    public synchronized void remove(Statement s) throws RDFException {
        /*
          Problem: fails in test 13, n=390
          Cause:
          Unfortunetely there are cases where the rest of the jena API can lose
          the statementID information. If you use the set() opertions to modify
          a statement you actually delete it an add a new one. However the user code
          retains a pointer to the original statement with any statement id information.
          If (s)he then does a second set the incorrect statement id will be used
          by the remove.
          However, without using the sid it is hard to delete statements which refer
          to reified statements that no longer exist (when cleaning out a database).
          Solution:
          Try to use the statement ID but if there is no such statement to delete
          fall back on search&destroy.
        */
        if (m_dbDriver.supportsJenaReification() && s instanceof StatementImplRDB ) {
            IDBID sid = ((StatementImplRDB)s).getRDBStatementId();
            if (sid != null) {
                if (m_dbDriver.deleteStatement(sid, m_model)) {
                    return;
                }
                // ID delete fails fall through to normal delete
            }
        }
        // No apparatent statement index so use search&destroy method
        Resource subject = m_model.convert(s.getSubject());
        Property predicate = m_model.convert(s.getPredicate());
        RDFNode object = m_model.convert(s.getObject());
        m_dbDriver.deleteStatement(subject, predicate, object, m_model);
    }

    /** helper function the ensures the statement is made up of objects within this model */
    protected Statement convertStatement(Statement s) throws RDFException {
        Resource subject = s.getSubject();
        Resource newSubject = m_model.convert(subject);
        Property predicate = s.getPredicate();
        Property newPredicate = m_model.convert(predicate);
        RDFNode object = s.getObject();
        RDFNode newObject = m_model.convert(object);
        if (subject != newSubject || predicate != newPredicate || object != newObject) {
            return createStatement(newSubject, newPredicate, newObject, m_model);
        } else {
            return s;
        }
    }

    /** Return the total number of statements in the model by a brute force - list and count! */
    public synchronized int size() throws RDFException {
        Iterator it = list();
        int count = 0;
        while (it.hasNext()) {
            count++; it.next();
        }
        return count;
    }

    /** Return true if the model contains a statements corresponding this triple */
    public synchronized boolean contains(Statement s) throws RDFException {
        Resource subject = m_model.convert(s.getSubject());
        Property predicate = m_model.convert(s.getPredicate());
        RDFNode object = m_model.convert(s.getObject());
        return m_dbDriver.isStatementPresent(subject, predicate, object, m_model);
        //return notEmpty(m_dbDriver.listStatements(subject, predicate, object, m_model));
    }

    /** Return true if the model contains at least one statement with this subject/property pair */
    public synchronized boolean contains(Resource s, Property p) throws RDFException {
        return notEmpty(m_dbDriver.listStatements(m_model.convert(s), m_model.convert(p), null, m_model));
    }

    /** Helper function - true if the ResultSetIterator has at least one entry, cleans up the iterator */
    private boolean notEmpty(ResultSetIterator it) {
        boolean notEmpty = false;
        if (it.hasNext()) {
            notEmpty = true;
            it.close();
        }
        return notEmpty;
    }

    /** List all the statements in this model */
    public synchronized Iterator list() throws RDFException {
        return m_dbDriver.listStatements(null, null, null, m_model);
    }

    /** List all statements with this s/p/o combination. Any argument can be null which corresponds to a wildcard. */
    public synchronized Iterator list(Resource subject, Property predicate, RDFNode object) throws RDFException {
        return m_dbDriver.listStatements(m_model.convert(subject), m_model.convert(predicate), m_model.convert(object), m_model);
    }

    /**
     * List all the statements in the given moddel which make the given set of constraints
     * (a constraint set can be obtained using {@link #createConstraints}).
     */
    public ResultSetIterator listStatements(IConstraints constraints, ModelRDB m) throws RDFException {
        return constraints.runQuery(m_dbDriver, m);
    }

    /**
     * Create a constraint object appropriate for this type of database layout
     */
    public IConstraints createConstraints(ModelRDB model) {
        return m_dbDriver.createConstraints(model);
    }

    /** List all statements in the model with this subject */
    public synchronized Iterator listBySubject(Resource subject) throws RDFException {
        return m_dbDriver.listStatements(m_model.convert(subject), null, null, m_model);
    }

    /** List all statements in the model with this predicate */
    public synchronized Iterator listByPredicate(Property property) throws RDFException {
        return m_dbDriver.listStatements(null, m_model.convert(property), null, m_model);
    }

    /** List all statements in the model with this object */
    public synchronized Iterator listByObject(RDFNode object)  throws RDFException {
        return m_dbDriver.listStatements(null, null, m_model.convert(object), m_model);
    }

    /**
     * List all the subject resources in the model
     * @return ResIterator over Resource objects
     */
    public synchronized ResIterator listSubjects() throws RDFException {
        return new ResIteratorImpl(m_dbDriver.listSubjects(m_model), null);
    }

    public synchronized NsIterator listNameSpaces() throws RDFException {
        return new NsIteratorImpl(m_dbDriver.listNamespaces(m_model), null);
    }

//=======================================================================
// Helper functions specific to this store type

    /**
     * Returns true if the database layout supports implicit reification
     * of statements (i.e. statements can be treated as resources).
     */
    public boolean supportsJenaReification() {
        return m_dbDriver.supportsJenaReification();
    }

    /**
     * Return an iterator over all statements which are have been
     * added as reified statements.
     * Defect: The current implementation will not notice if all
     * references to a reified statement have been removed.
     */
    public ResultSetIterator listReifiedStatements(ModelRDB m) throws RDFException {
        return m_dbDriver.listReifiedStatements(m);
    }

}

/*
 *  (c) Copyright Hewlett-Packard Company 2000, 2001
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */