/* -*-Java-*-

*******************************************************************

*

* File:         TestGenericDriver.java

* RCS:          $Header: /cvsroot/jena/jena/src/com/hp/hpl/mesa/rdf/jena/rdb/TestGenericDriver.java,v 1.4 2002/03/01 14:25:13 der Exp $

* Author:       Dave Reynolds, HPLabs, Bristol

* Created:      30/9/01

* Modified:     $Date: 2002/03/01 14:25:13 $

* Status:       Experimental (Do Not Distribute)

*

* (c) Copyright 2001, Hewlett-Packard Company, all rights reserved.

* [See end of file]

*********************************************************************/



//=======================================================================

// Package

package com.hp.hpl.mesa.rdf.jena.rdb;



//=======================================================================

// Imports

import com.hp.hpl.mesa.rdf.jena.rdb.*;

import com.hp.hpl.mesa.rdf.jena.model.*;

import com.hp.hpl.mesa.rdf.jena.common.*;

import java.io.*;

import java.sql.SQLException;

import java.sql.Connection;

import java.sql.ResultSet;

import java.sql.ResultSetMetaData;

import java.util.*;

import junit.framework.*;



//=======================================================================

/**

 * Unit tests for the generic database driver.

 * The main tests are in TestStoreRDB - these were really bootstrap tests.

 * Need to set appropriate java properties to define the database connection:

 * <PRE>   jena.rdb.test.dbbase  - the database uri

 * jena.rdb.test.user  - the username to login under

 * jena.rdb.test.password  - the password for that user

 * </PRE>

 *

 * @author <a href="mailto:der@hplb.hpl.hp.com">Dave Reynolds</a>

 * @version $Revision: 1.4 $ on $Date: 2002/03/01 14:25:13 $

 */



public class TestGenericDriver extends TestCase {



    /** The database connection */

    protected DBConnection m_dbcon;



    /** The database type to be tested */

    protected String m_databaseType;



    /** The database layout to be tested */

    protected String m_layout;



    public TestGenericDriver(String name, String layout, String database) {

        super(name);

        m_databaseType = database;

        m_layout = layout;

    }



    public void setUp() throws IOException {

        String m_dbname = System.getProperty(TestRDB.DBFULL_PROP);

        String m_user =  System.getProperty(TestRDB.DBUSER_PROP, "DER");

        String m_password =  System.getProperty(TestRDB.DBPASSWORD_PROP, "key");

        m_dbcon = new DBConnection(m_dbname, m_user, m_password);

    }



    /** Support code - open a connection to a newly formatted database. */

    public IRDBDriver setupDB() throws RDFRDBException {

        IRDBDriver driver = m_dbcon.getDriver(m_layout, m_databaseType);

        if (driver.isDBFormatOK()) {

            driver.close();

            driver.cleanDB();

            driver.close();

        }

        driver.formatDB();

        return driver;

    }



    /** Support code - tear down the connection */

    public void tearDownDB(IRDBDriver driver) throws SQLException, RDFRDBException {

        driver.close();

        driver.cleanDB();

        driver.close();

        m_dbcon.close();

    }



    /** Test Database formating. */

    public void testDBFormat() throws RDFRDBException, IOException, SQLException {

        // Open connection

        IRDBDriver driver =  m_dbcon.getDriver(m_layout, m_databaseType);

        Connection con = driver.getConnection();

        assertNotNull("jdbc connection ok", con);

        // Try to clean up and pre-existing formating

        try {

            if (driver.isDBFormatOK()) {

                driver.cleanDB();

                driver.close();

            }

        } catch (Exception e) {

            // Don't care about errors here, this is just preparation for the real run

        }

        assertEquals("DB should be clean", false, driver.isDBFormatOK());

        driver.formatDB();

        assertEquals("DB should be formated", true, driver.isDBFormatOK());

        driver.close();         // Force release of resources to allow clean up

        driver.cleanDB();

        assertEquals("DB should be clean", false, driver.isDBFormatOK());

        driver.close();

    }



    /** Test loading of sql definition files. Only works on dev code, not released code so scratched from normal regression. */

    public void testSQLCache() throws IOException, SQLException {

        /*

        // Create the test file

        Writer outs = new FileWriter("classes/etc/test.sql");

        outs.write("# A comment line\n");

        outs.write("op1\nsql line 1;;\nsql line last;;\n\n");

        outs.write("op2 \nsql line 1\nsql line last\n\n");

        outs.write("op3\nsql ${id} line\n\n");

        outs.close();

        SQLCache cache = new SQLCache("etc/test.sql", null, null, "id-type");

        assertEquals(cache.getSQLStatement("op1"), "sql line 1;;\nsql line last;;\n");

        assertEquals(cache.getSQLStatement("op2"), "sql line 1\nsql line last\n");

        assertEquals(cache.getSQLStatement("op3"), "sql id-type line\n");

        assertNull(cache.getSQLStatement("op4"));

        Collection statements = cache.getSQLStatementGroup("op1");

        assertEquals("Statement group size", 2, statements.size());

        assertEquals("Statement group member", "sql line 1", statements.iterator().next());

        */

    }



    /** Test driver functions. Done as one big test instead of lots of small ones

     *  so that the setup/teardown overhead is only incurred once

     */

    public void testDriver() throws SQLException, RDFException {

        DriverGenericGeneric driver = (DriverGenericGeneric) setupDB();

        doTestNamespaceAdd(driver);

        doTestResourceAdd(driver);

        doTestLiteralAdd(driver);

        doTestStatementAdd(driver);

        if (driver.supportsMultipleModels()) {

            doTestMuliModels(driver);

        }

        tearDownDB(driver);

    }



    /** Test namespace manipulation */

    public void doTestNamespaceAdd(DriverGenericGeneric driver) throws SQLException, RDFException {

        String ns1 = "http://namespace/foo";

        String ns2 = "http://namespace/foo2";

        IDBID id1 = driver.addNamespace(ns1);

        //System.out.println("Namespace: " + ns1 + ", maps to: " + id1.getID());

        IDBID id1b = driver.getNamespaceID(ns1);

        assertEquals(id1, id1b);

        IDBID id2 = driver.addNamespace(ns2);

        IDBID id2b = driver.getNamespaceID(ns2);

        assertEquals(id1, id1b);

        assertTrue(!id2.equals(id1));

        IDBID id1c = driver.addNamespace(ns1);

        assertEquals(id1, id1c);

        String ns1b = driver.getNamespace(id1c);

        assertEquals(ns1, ns1b);

    }



    /** Test stored procedure for adding resources */

    public void doTestResourceAdd(DriverGenericGeneric driver) throws SQLException, RDFException {

        StoreRDB store = StoreRDB.createDB((IRDBDriver)driver);

        ModelRDB model = new ModelRDB(store);

        Resource r1 = model.createResource("http://namespace/foo#res1");

        Resource r2 = model.createResource("http://namespace/foo#res2");

        Resource r3 = model.createResource("http://namespace/foo2#res3");

        Resource r2b = model.createResource("http://namespace/foo#res2");

        IDBID id1 = driver.addResource(r1);

        IDBID id2 = driver.addResource(r2);

        IDBID id3 = driver.addResource(r3);

        IDBID id2b = driver.addResource(r2b);

        IDBID nsid = driver.getNamespaceID("http://namespace/foo#");

        assertEquals("Resource id allocation in clean database", id2, ((ResourceImplRDB)r2).getRDBId());

        assertEquals("Resource id allocation in clean database", id3, ((ResourceImplRDB)r3).getRDBId());

        assertEquals("NS ID alloc in clean db", nsid, ((ResourceImplRDB)r1).getRDBNSId());

        assertEquals("NS ID realloc in clean db", nsid, ((ResourceImplRDB)r2b).getRDBNSId());

        assertEquals("Detect prior id allocation", id2, ((ResourceImplRDB)r2b).getRDBId());

        ResourceImplRDB ra = new ResourceImplRDB(new AnonId("foobar1"), model);

        driver.addResource(ra);

        assertEquals("anon NS id", DriverGenericGeneric.ANON_NAMESPACE, driver.getNamespace(driver.m_anonNSId));

        assertEquals("anon resource", driver.m_anonNSId, ra.getRDBNSId());

        /** @todo check ra stored correctly */

        //listTable("RDF_RESOURCES", driver);

        //listTable("RDF_NAMESPACES", driver);

    }



    /** Test stored procedure for adding literals */

    public void doTestLiteralAdd(DriverGenericGeneric driver) throws SQLException, RDFException {

        IDBID id1 = driver.addLiteral(new LiteralImpl(1234));

        IDBID id2 = driver.addLiteral(new LiteralImpl("Foo bar"));

        IDBID id3 = driver.addLiteral(new LiteralImpl("<help/>", "English", true));

        IDBID id4 = driver.addLiteral(new LiteralImpl(1234));

        assertEquals(id1, id4);

        assertTrue(!id1.equals(id2));

        assertTrue(!id1.equals(id3));

        assertTrue(!id2.equals(id3));

        // Check table entries are as expected

        // TODO: Rewrite for order indendence

        /*

        java.sql.Statement sql = driver.getConnection().createStatement();

        ResultSet rs = sql.executeQuery("SELECT LITERAL FROM RDF_LITERALS");

        rs.next(); assertEquals(rs.getString(1), "1234");

        rs.next(); assertEquals(rs.getString(1), "Foo bar");

        rs.next(); assertEquals(rs.getString(1), "<help/>");

        rs.close();

        */

        // Check getLiteralID

        IDBID idt = ((DriverGenericGeneric)driver).getLiteralID(new LiteralImpl("Foo bar"));

        assertEquals(idt, id2);

        // Check big literals

        StoreRDB store = new StoreRDB(driver);

        store.begin();  // Wrap in a transaction so postgresql can handle blobs

        int maxLiteral = ((DriverGenericGeneric)driver).MAX_LITERAL;

        StringBuffer buf = new StringBuffer(maxLiteral);

        for (int i = 0; i < maxLiteral; i++) buf.append("abcdefghij".charAt(i % 10));

        String longString1 = buf.toString() + "a";

        String longString2 = buf.toString() + "b";

        // Big literal - max + 1 case

        IDBID ida = driver.addLiteral(new LiteralImpl(longString1));

        IDBID idb = driver.addLiteral(new LiteralImpl(longString2));

        assertTrue(!ida.equals(idb));

        Literal la = driver.getLiteral(ida);

        Literal lb = driver.getLiteral(idb);

        assertEquals(longString1, la.toString());

        assertEquals(longString2, lb.toString());

        // Big literal - max + 0 case

        ida = driver.addLiteral(new LiteralImpl(longString1.substring(1)));

        idb = driver.addLiteral(new LiteralImpl(longString2.substring(1)));

        assertTrue(!ida.equals(idb));

        la = driver.getLiteral(ida);

        lb = driver.getLiteral(idb);

        assertEquals(longString1.substring(1), la.toString());

        assertEquals(longString2.substring(1), lb.toString());

        // Big literal - max - 1 case

        ida = driver.addLiteral(new LiteralImpl(longString1.substring(2)));

        idb = driver.addLiteral(new LiteralImpl(longString2.substring(2)));

        assertTrue(!ida.equals(idb));

        la = driver.getLiteral(ida);

        lb = driver.getLiteral(idb);

        assertEquals(longString1.substring(2), la.toString());

        assertEquals(longString2.substring(2), lb.toString());

        store.commit();

    }



    /** Check statement creation - node valid for hash versions */

    public void doTestStatementAdd(DriverGenericGeneric driver) throws SQLException, RDFException {

        StoreRDB store = StoreRDB.createDB((IRDBDriver)driver);

        ModelRDB model = new ModelRDB(store);

        Resource r1 = model.createResource("http://namespace/foo#sub1");

        Property p1 = model.createProperty("http://namespace/foo#prop");

        com.hp.hpl.mesa.rdf.jena.model.Statement st = model.createStatement(r1, p1, 42);

        model.add(st);

        model.add(model.createStatement(r1,p1,r1));

        java.sql.Statement sql = driver.getConnection().createStatement();

        ResultSet rs = sql.executeQuery("SELECT SUBJECT, PREDICATE, OBJECT, OBJECT_ISLITERAL FROM RDF_STATEMENTS");

        assertTrue(rs.next());

        Object subj1 = rs.getObject(1);

        Object pred1 = rs.getObject(2);

        Object obj1 = rs.getObject(3);

        int il1 = rs.getInt(4);

        assertTrue(rs.next());

        Object subj2 = rs.getObject(1);

        Object pred2 = rs.getObject(2);

        Object obj2 = rs.getObject(3);

        int il2 = rs.getInt(4);

        assertEquals(subj1, subj2);

        assertEquals(subj2, obj2);

        assertEquals(pred1, pred2);

        assertTrue(!subj1.equals(pred1));

        assertEquals(il1, 1);

        assertEquals(il2, 0);

    }



    /** Test support for multiple models */

    public void doTestMuliModels(IRDBDriver driver) throws RDFException, SQLException {

        // Check basic low level model name machinery

        assertTrue(driver.supportsMultipleModels());

        IDBID id1 = driver.allocateModelID("model1");

        IDBID id2 = driver.allocateModelID("model2");

        IDBID id1b = driver.allocateModelID("model1");

        assertEquals(id1, id1b);

        assertTrue(!id1.equals(id2));

        assertIteratorValues(driver.listModels(), new Object[] {"default", "model1", "model2"});

        // Now create two models and check that the statements are independent

        ModelRDB m1 = ModelRDB.createModel(m_dbcon, "model1");

        ModelRDB m2 = ModelRDB.createModel(m_dbcon, "model2");

        Resource r1 = m1.createResource("http://namespace/foo#sub1");

        Resource r2 = m1.createResource("http://namespace/foo2#sub2");

        Property p = m1.createProperty("http://namespace/foo#prop");

        Property s = m1.createProperty("http://namespace/foop#props");

        Statement s1 = m1.createStatement(r1, p, 42); m1.add(s1);

        Statement s2 = m1.createStatement(r2, p, 13); m1.add(s2);

        //Statement s3 = m2.createStatement(r1, p, 42); m2.add(s3);

        Statement s3 = s1; m2.add(s3);

        Statement s4 = m2.createStatement(r1, s, 13); m2.add(s4);

        // Check list subjects

        assertIteratorValues(driver.listSubjects(m2), new Object[] {r1});

        assertIteratorValues(driver.listSubjects(m1), new Object[] {r1, r2});

        // Check list namespaces

        assertIteratorValues(driver.listNamespaces(m1), new Object[] {"http://namespace/foo#"});

        assertIteratorValues(driver.listNamespaces(m2), new Object[] {"http://namespace/foo#", "http://namespace/foop#"});

        // Check list statements - 8 variants

        assertIteratorValues(driver.listStatements(null, null, null, m1), new Object[] {s1, s2});

        assertIteratorValues(driver.listStatements(null, null, null, m2), new Object[] {s3, s4});

        assertIteratorValues(driver.listStatements(r1, null, null, m1), new Object[] {s1});

        assertIteratorValues(driver.listStatements(r1, null, null, m2), new Object[] {s3, s4});

        assertIteratorValues(driver.listStatements(null, p, null, m1), new Object[] {s1, s2});

        assertIteratorValues(driver.listStatements(null, p, null, m2), new Object[] {s3});

        assertIteratorValues(driver.listStatements(null, null, new LiteralImpl(13) , m1), new Object[] {s2});

        assertIteratorValues(driver.listStatements(null, null, new LiteralImpl(13) , m2), new Object[] {s4});

        assertIteratorValues(driver.listStatements(r1, p, null, m1), new Object[] {s1});

        assertIteratorValues(driver.listStatements(r1, p, null, m2), new Object[] {s3});

        assertIteratorValues(driver.listStatements(null, p, new LiteralImpl(13), m1), new Object[] {s2});

        assertIteratorValues(driver.listStatements(null, p, new LiteralImpl(13), m2), new Object[] {});

        assertIteratorValues(driver.listStatements(r1, null, new LiteralImpl(13), m1), new Object[] {});

        assertIteratorValues(driver.listStatements(r1, null, new LiteralImpl(13), m2), new Object[] {s4});

        assertIteratorValues(driver.listStatements(r1, s, new LiteralImpl(13), m1), new Object[] {});

        assertIteratorValues(driver.listStatements(r1, s, new LiteralImpl(13), m2), new Object[] {s4});

        // Check delete statements

        driver.deleteStatement(r1, p, new LiteralImpl(42), m1);

        assertIteratorValues(driver.listStatements(null, null, null, m1), new Object[] {s2});

        assertIteratorValues(driver.listStatements(null, null, null, m2), new Object[] {s3, s4});

        driver.deleteStatement(s2.getSubject(), s2. getPredicate(), s2.getObject(), m1);

        assertIteratorValues(driver.listStatements(null, null, null, m1), new Object[] {});

        assertIteratorValues(driver.listStatements(null, null, null, m2), new Object[] {s3, s4});



        m1.close();

        m2.close();

    }



    /** Helper function test an iterator against a list of objects - order dependent */

    public void assertIteratorValues(Iterator it, Object[] vals) {

        boolean[] found = new boolean[vals.length];

        for (int i = 0; i < vals.length; i++) found[i] = false;

        while (it.hasNext()) {

            Object n = it.next();

            boolean gotit = false;

            for (int i = 0; i < vals.length; i++) {

                if (n.equals(vals[i])) {

                    gotit = true;

                    found[i] = true;

                }

            }

            assertTrue(gotit);

        }

        for (int i = 0; i < vals.length; i++) {

            assertTrue(found[i]);

        }

    }



    /** Helper function during debug - list a table */

    public void listTable(String tname, IRDBDriver driver) throws SQLException, RDFRDBException {

        Connection con = driver.getConnection();

        System.out.println("Table - " + tname);

        java.sql.Statement st = con.createStatement();

        ResultSet rs = st.executeQuery("SELECT * FROM " + tname);

        ResultSetMetaData rsmd = rs.getMetaData();

        for (int i = 1; i <= rsmd.getColumnCount(); i++) {

            System.out.print("\t" + rsmd.getColumnLabel(i));

        }

        System.out.println("");

        while (rs.next()) {

            for (int i = 1; i <= rsmd.getColumnCount(); i++) {

                System.out.print("\t" + rs.getObject(i));

            }

            System.out.println("");

        }

    }



    /** Return full test suite suitable for the given database type */

    public static TestSuite suite(String layout, String database) {

        TestSuite suite = new TestSuite();

        suite.addTest(new TestGenericDriver("testDBFormat", layout, database));

        suite.addTest(new TestGenericDriver("testSQLCache", layout, database));

        suite.addTest(new TestGenericDriver("testDriver", layout, database));

        return suite;

    }

}



/*

 *  (c) Copyright Hewlett-Packard Company 2000, 2001

 *  All rights reserved.

 *

 * Redistribution and use in source and binary forms, with or without

 * modification, are permitted provided that the following conditions

 * are met:

 * 1. Redistributions of source code must retain the above copyright

 *    notice, this list of conditions and the following disclaimer.

 * 2. Redistributions in binary form must reproduce the above copyright

 *    notice, this list of conditions and the following disclaimer in the

 *    documentation and/or other materials provided with the distribution.

 * 3. The name of the author may not be used to endorse or promote products

 *    derived from this software without driverific prior written permission.



 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR

 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES

 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.

 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,

 * INCIDENTAL, driverIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT

 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,

 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY

 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT

 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF

 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 */