/* -*-Java-*-

*******************************************************************

*

* File:         TestStoreRDB.java

* RCS:          $Header: /cvsroot/jena/jena/src/com/hp/hpl/mesa/rdf/jena/rdb/TestStoreRDB.java,v 1.10 2002/10/04 14:04:14 der Exp $

* Author:       Dave Reynolds, HPLabs, Bristol

* Created:      7/9/01

* Modified:     $Date: 2002/10/04 14:04:14 $

* Status:       Experimental (Do Not Distribute)

*

* (c) Copyright 2001, Hewlett-Packard Company, all rights reserved.

* [See end of file]

*********************************************************************/



//=======================================================================

// Package

package com.hp.hpl.mesa.rdf.jena.rdb;



//=======================================================================

// Imports

import com.hp.hpl.mesa.rdf.jena.rdb.*;

import com.hp.hpl.mesa.rdf.jena.model.*;

import com.hp.hpl.mesa.rdf.jena.common.*;

import java.io.*;

import java.sql.SQLException;

import java.sql.PreparedStatement;

import java.sql.ResultSet;

import java.sql.ResultSetMetaData;

import java.util.*;

import junit.framework.*;



//=======================================================================

/**

 * Test harness for StoreRDB and supporting classes.

 *

 * @author <a href="mailto:der@hplb.hpl.hp.com">Dave Reynolds</a>

 * @version $Revision: 1.10 $ on $Date: 2002/10/04 14:04:14 $

 */



public class TestStoreRDB extends TestCase {



    /** The database connection */

    protected DBConnection m_dbcon;



    /** The database layout to test */

    protected String m_layout;



    /** The database type to test */

    protected String m_database;



    /** The driver implementation for this store */

    protected IRDBDriver m_driver;



    /** Store instance */

    protected StoreRDB store;



    /** Model instance which uses store */

    protected ModelRDB model;



    /** Test model elements */

    Property p1, p2, p3;

    Resource foo, bar, anon;



    /** Construct a test suitable for the given layout and database combo */

    public TestStoreRDB(String name, String layout, String database) {

        super(name);

        m_layout = layout;

        m_database = database;



        String dbname = System.getProperty(TestRDB.DBFULL_PROP);

        String user =  System.getProperty(TestRDB.DBUSER_PROP, "DER");

        String password =  System.getProperty(TestRDB.DBPASSWORD_PROP, "key");

        m_dbcon = new DBConnection(dbname, user, password);

    }



    /** Support code - open a connection to a newly formatted database. */

    public void setUp() throws RDFRDBException {

        m_driver = m_dbcon.getDriver(m_layout, m_database);

        if (m_driver.isDBFormatOK()) {

            m_driver.close();

            m_driver.cleanDB();

            m_driver.close();

        }

        store = StoreRDB.createDB(m_driver);

    }



    /** Support code - tear down the connection */

    public void tearDown() throws SQLException, RDFRDBException {

        m_driver.close();

        m_driver.cleanDB();

        m_driver.close();

    }



    /** Create a model to be tested */

    public void createModel(boolean hint) throws RDFException {

        model = new ModelRDB(store);

        p1 = model.createProperty("http://namespace/myprops#prop1");

        p2 = model.createProperty("http://namespace/myprops#prop2");

        p3 = model.createProperty("http://namespace/myprops#prop3");

        if (hint) {

            m_driver.hintPropertyTable(p1);

            m_driver.hintPropertyTable(p3);

        }

        anon = model.createResource().addProperty(p3, 13);

        bar = model.createResource("http://namespace/mine#bar")

                .addProperty(p2, "fred")

                .addProperty(p3, 50);

        foo = model.createResource("http://namespace/mine#foo")

                .addProperty(p1, 50)

                .addProperty(p1, anon)

                .addProperty(p1, bar);

        /*

        listTable("RDF_STATEMENTS");

        listTable("RDF_RESOURCES");

        listTable("RDF_NAMESPACES");

        listTable("RDF_LITERALS");

        */

    }



    /** remove all current statements */

    public void clearModel() throws RDFException {

        StmtIterator it = model.listStatements();

        while (it.hasNext()) {

            it.next();

            it.remove();

        }

        it.close();

    }



    /**

     * The main test function.

     * Do as one big function instead of lots of little ones to avoid creating and

     * tearing down the database too many times.

     */

    public void testStore() throws SQLException, RDFException {

        createModel(false);

        doTestListNamespaces();

        doTestListStatements();

        doTestConstraints();

        //doTestGetNamespaceId();

        doTestListSubjects();

        doTestMultipleAssert();

        doTestMultipleResources();

        doTestLangAttrib();

        doTestSQLCachePS();

        doTestDeleteStatement();

        if (model.supportsJenaReification())

            doTestReification();

        doTestBulkLoader();

        doTestModelCleaning();

        doTestLiteralCase();

    }



    /**

     * Try again with storage hints on p1

     */

    public void testStoreHint() throws SQLException, RDFException {

        createModel(true);

        doTestListNamespaces();

        doTestListStatements();

        doTestConstraints();

        doTestListSubjects();

        doTestMultipleAssert();

        doTestMultipleResources();

        doTestLangAttrib();

        doTestSQLCachePS();

        doTestDeleteStatement();

    }



    /** Check listSubjects */

    public void doTestListSubjects() throws SQLException, RDFException {

        Iterator it = m_driver.listSubjects(model);

        isResourceListOK(it, new Resource[] {bar, foo, anon});

    }



    /** Check listNamepsaces */

    public void doTestListNamespaces() throws RDFException, SQLException {

        Property baz = model.createProperty("http://todie#baz");

        Statement st = model.createStatement(foo, baz, foo);

        model.add(st);

        Iterator it = m_driver.listNamespaces(model);

        assertEquals("http://namespace/myprops#", it.next());

        assertEquals("http://todie#", it.next());

        assertTrue(!it.hasNext());

        model.remove(st);

        it = m_driver.listNamespaces(model);

        assertEquals("http://namespace/myprops#", it.next());

        assertTrue(!it.hasNext());

    }



    /** Check list statements - this version is dependent on result order, may need changing */

    public void doTestListStatements() throws RDFException, SQLException {

        /*

        listTable("RDF_STATEMENTS");

        listTable("RDF_RESOURCES");

        listTable("RDF_NAMESPACES");

        listTable("RDF_LITERALS");

        */

        // Case of list all statements

        Iterator it = m_driver.listStatements(null, null, null, model);

        isStatementListOK(it, new Statement[] {

            new StatementImpl(anon, p3, new LiteralImpl("13")),

            new StatementImpl(bar,  p2, new LiteralImpl("fred")),

            new StatementImpl(bar,  p3, new LiteralImpl("50")),

            new StatementImpl(foo,  p1, new LiteralImpl("50")),

            new StatementImpl(foo,  p1, anon),

            new StatementImpl(foo,  p1, bar)  });

        // Case of listS

        it = m_driver.listStatements(foo, null, null, model);

        isStatementListOK(it, new Statement[] {

            new StatementImpl(foo,  p1, new LiteralImpl("50")),

            new StatementImpl(foo,  p1, anon),

            new StatementImpl(foo,  p1, bar)  });

        // Case of listp

        it = m_driver.listStatements(null, p3, null, model);

        isStatementListOK(it, new Statement[] {

            new StatementImpl(anon, p3, new LiteralImpl("13")),

            new StatementImpl(bar,  p3, new LiteralImpl("50")) });

        // Case of listS

        it = m_driver.listStatements(bar, p3, null, model);

        isStatementListOK(it, new Statement[] {

             new StatementImpl(bar,  p3, new LiteralImpl("50")) });

        // Case of listO - resource

        it = m_driver.listStatements(null, null, bar, model);

        isStatementListOK(it, new Statement[] {

             new StatementImpl(foo,  p1, bar) });

        // Case of listO - literal

        it = m_driver.listStatements(null, null, model.createLiteral(50), model);

        isStatementListOK(it, new Statement[] {

            new StatementImpl(bar,  p3, new LiteralImpl("50")),

            new StatementImpl(foo,  p1, new LiteralImpl("50"))  });

        // Case of listSO - literal

        it = m_driver.listStatements(bar, null, model.createLiteral(50), model);

        isStatementListOK(it, new Statement[] {

            new StatementImpl(bar,  p3, new LiteralImpl("50")) });

        // Case of listPO - literal

        it = m_driver.listStatements(null, p3, model.createLiteral(50), model);

        isStatementListOK(it, new Statement[] {

            new StatementImpl(bar,  p3, new LiteralImpl("50")) });

        // Case of listSPO - literal

        it = m_driver.listStatements(bar, p3, model.createLiteral(50), model);

        isStatementListOK(it, new Statement[] {

            new StatementImpl(bar,  p3, new LiteralImpl("50")) });

        // Case of listSPO - resource

        it = m_driver.listStatements(foo, p1, bar, model);

        isStatementListOK(it, new Statement[] {

            new StatementImpl(foo,  p1, bar) });

        // Case of listSPO - null

        it = m_driver.listStatements(bar, p3, model.createLiteral(120), model);

        assertTrue(!it.hasNext());

    }



    /** Check list statements with constraints */

    public void doTestConstraints() throws RDFException, SQLException {

        // Case of list all statements

        IConstraints constraints = m_driver.createConstraints(model);

        Iterator it = m_driver.listStatements(constraints, model);

        isStatementListOK(it, new Statement[] {

            new StatementImpl(anon, p3, new LiteralImpl("13")),

            new StatementImpl(bar,  p2, new LiteralImpl("fred")),

            new StatementImpl(bar,  p3, new LiteralImpl("50")),

            new StatementImpl(foo,  p1, new LiteralImpl("50")),

            new StatementImpl(foo,  p1, anon),

            new StatementImpl(foo,  p1, bar)  });

        // Case of listS

        constraints.reset().addSubjectConstraint(foo);

        it = m_driver.listStatements(constraints, model);

        isStatementListOK(it, new Statement[] {

            new StatementImpl(foo,  p1, new LiteralImpl("50")),

            new StatementImpl(foo,  p1, anon),

            new StatementImpl(foo,  p1, bar)  });

        // Case of listp

        constraints.reset().addPropertyConstraint(p3);

        it = m_driver.listStatements(constraints, model);

        isStatementListOK(it, new Statement[] {

            new StatementImpl(anon, p3, new LiteralImpl("13")),

            new StatementImpl(bar,  p3, new LiteralImpl("50")) });

        // Case of listSP

        constraints.reset().addSubjectConstraint(bar).addPropertyConstraint(p3);

        it = m_driver.listStatements(constraints, model);

        isStatementListOK(it, new Statement[] {

             new StatementImpl(bar,  p3, new LiteralImpl("50")) });

        // Case of listO - resource

        constraints.reset().addObjectConstraint(bar);

        it = m_driver.listStatements(constraints, model);

        isStatementListOK(it, new Statement[] {

             new StatementImpl(foo,  p1, bar) });

        // Case of listO - literal

        constraints.reset().addObjectConstraint(model.createLiteral(50));

        it = m_driver.listStatements(constraints, model);

        isStatementListOK(it, new Statement[] {

            new StatementImpl(bar,  p3, new LiteralImpl("50")),

            new StatementImpl(foo,  p1, new LiteralImpl("50"))  });

        // Case of listSO - literal

        constraints.reset().addSubjectConstraint(bar).addObjectConstraint(model.createLiteral(50));

        it = m_driver.listStatements(constraints, model);

        isStatementListOK(it, new Statement[] {

            new StatementImpl(bar,  p3, new LiteralImpl("50")) });

        // Case of listPO - literal

        constraints.reset().addPropertyConstraint(p3).addObjectConstraint(model.createLiteral(50));

        it = m_driver.listStatements(constraints, model);

        isStatementListOK(it, new Statement[] {

            new StatementImpl(bar,  p3, new LiteralImpl("50")) });

        // Case of listSPO - literal

        constraints.reset()

                   .addSubjectConstraint(bar)

                   .addPropertyConstraint(p3)

                   .addObjectConstraint(model.createLiteral(50));

        it = m_driver.listStatements(constraints, model);

        it = m_driver.listStatements(constraints, model);

        isStatementListOK(it, new Statement[] {

            new StatementImpl(bar,  p3, new LiteralImpl("50")) });

        // Case of listSPO - resource

        constraints.reset()

                   .addSubjectConstraint(foo)

                   .addPropertyConstraint(p1)

                   .addObjectConstraint(bar);

        it = m_driver.listStatements(constraints, model);

        isStatementListOK(it, new Statement[] {

            new StatementImpl(foo,  p1, bar) });

        // Case of listSPO - null

        constraints.reset()

                   .addSubjectConstraint(bar)

                   .addPropertyConstraint(p3)

                   .addObjectConstraint(model.createLiteral(120));

        it = m_driver.listStatements(constraints, model);

        assertTrue(!it.hasNext());



        // String match functions

        Property p4 = model.createProperty("http://namespace/myprops#prop4");

        foo.addProperty(p4, "foo");

        foo.addProperty(p4, "foobar");

        foo.addProperty(p4, "bar");

        foo.addProperty(p4, "foo27bar");

        foo.addProperty(p4, "27");

        constraints.reset().addSubjectConstraint(foo).addPropertyConstraint(p4)

                           .addStringConstraint("LIKE", "foo%");

        it = m_driver.listStatements(constraints, model);

        isStatementListOK(it, new Statement[] {

            new StatementImpl(foo,  p4, new LiteralImpl("foo")),

            new StatementImpl(foo,  p4, new LiteralImpl("foobar")),

            new StatementImpl(foo,  p4, new LiteralImpl("foo27bar"))  });

        constraints.reset().addSubjectConstraint(foo).addPropertyConstraint(p4)

                           .addStringConstraint("NOT LIKE", "%foo%");

        it = m_driver.listStatements(constraints, model);

        isStatementListOK(it, new Statement[] {

            new StatementImpl(foo,  p4, new LiteralImpl("bar")),

            new StatementImpl(foo,  p4, new LiteralImpl("27"))  });

        constraints.reset().addSubjectConstraint(foo).addPropertyConstraint(p4)

                           .addStringConstraint("LIKE", "foo%")

                           .addStringConstraint("NOT LIKE", "%27%");

        it = m_driver.listStatements(constraints, model);

        isStatementListOK(it, new Statement[] {

            new StatementImpl(foo,  p4, new LiteralImpl("foo")),

            new StatementImpl(foo,  p4, new LiteralImpl("foobar"))  });

        // Int match functions

        foo.addProperty(p4, 3);

        foo.addProperty(p4, 4);

        foo.addProperty(p4, 42);

        foo.addProperty(p4, 43);

        //listTable("RDF_LITERALS");

        constraints.reset().addSubjectConstraint(foo).addPropertyConstraint(p4)

                           .addIntConstraint("<", 27);

        it = m_driver.listStatements(constraints, model);

        isStatementListOK(it, new Statement[] {

            new StatementImpl(foo,  p4, new LiteralImpl("3")),

            new StatementImpl(foo,  p4, new LiteralImpl("4"))  });

        constraints.reset().addSubjectConstraint(foo).addPropertyConstraint(p4)

                           .addIntConstraint(">=", 27);

        it = m_driver.listStatements(constraints, model);

        isStatementListOK(it, new Statement[] {

            new StatementImpl(foo,  p4, new LiteralImpl("27")),

            new StatementImpl(foo,  p4, new LiteralImpl("42")),

            new StatementImpl(foo,  p4, new LiteralImpl("43"))  });

        constraints.reset().addSubjectConstraint(foo).addPropertyConstraint(p4)

                           .addIntConstraint("<=", 42)

                           .addIntConstraint(">", 4);

        it = m_driver.listStatements(constraints, model);

        isStatementListOK(it, new Statement[] {

            new StatementImpl(foo,  p4, new LiteralImpl("27")),

            new StatementImpl(foo,  p4, new LiteralImpl("42"))  });

    }



    /** Check delete statement - this version is dependent on result order, may need changing */

    public void doTestDeleteStatement() throws RDFException {

        int base = store.size();

        // Add two statements and check they are there

        foo.addProperty(p2, bar);

        foo.addProperty(p2, "42");

        assertEquals(base+2, store.size());

        Iterator it = m_driver.listStatements(foo, p2, null, model);

        isStatementListOK(it, new Statement[] {

            new StatementImpl(foo,  p2, bar),

            new StatementImpl(foo,  p2, new LiteralImpl("42"))  });

        // Delete one and check it has gone

        m_driver.deleteStatement(foo, p2, bar, model);

        it = m_driver.listStatements(foo, p2, null, model);

        isStatementListOK(it, new Statement[] {

            new StatementImpl(foo,  p2, new LiteralImpl("42"))  });

        // Delete the last

        m_driver.deleteStatement(foo, p2, new LiteralImpl("42"),model);

        it = m_driver.listStatements(foo, p2, null, model);

        assertTrue(!it.hasNext());

        assertEquals(base, store.size());

    }



    /** Check SQLCache preparedStatement functions */

    public void doTestSQLCachePS() throws SQLException, RDFRDBException {

        SQLCache sql = m_driver.getSQLCache();

        IDBID id = ((DriverGenericGeneric)m_driver).addNamespace("foobar");

        // Use getNamespace op as a simple test case

        PreparedStatement ps1 = sql.getPreparedSQLStatement("getNamespace");

        PreparedStatement ps2 = sql.getPreparedSQLStatement("getNamespace");

        ps1.setObject(1, id.getID());

        ps2.setInt(1, 50);

        ResultSet rs = ps1.executeQuery();

        assertTrue(rs.next());

        assertEquals("foobar", rs.getString(1));

        rs.close();

        sql.returnPreparedSQLStatement(ps1, "getNamespace");

        sql.returnPreparedSQLStatement(ps2, "getNamespace");

        PreparedStatement ps =  sql.getPreparedSQLStatement("getNamespace");

        if (sql.getCachePreparedStatements())  assertEquals("PS realloc", ps1, ps);

    }



    /** Check getNamespaceId - this version can't check the cache miss case, that was done manually during development. */

    /*

    public void doTestGetNamespaceId() throws RDFException, SQLException {

        IDBID r1 = m_driver.getNamespaceID("http://namespace/myprops#");

        assertEquals(2, ((DBIDInt)r1).getIntID());

        IDBID r2 = m_driver.getNamespaceID("http://namespace/neverHeardOfIt#");

        assertNull(r2);

    }

    */



    /** Check case of asserting the same statement twice */

    public void doTestMultipleAssert() throws RDFException, SQLException {

        long size = model.size();

        Statement st = model.createStatement(foo, p3, model.createLiteral("check"));

        model.add(st);

        assertEquals("Assert once", size+1, model.size());

        model.add(st);

        assertEquals("Assert second time", size+1, model.size());

        model.remove(st);

        assertEquals("Removed assertions", size, model.size());

    }



    /** Check asserting resources with different namespaces but same local name */

    public void doTestMultipleResources() throws RDFException {

        long size = model.size();

        Property pt1 = model.createProperty("http://ns1/", "p");

        Property pt2 = model.createProperty("http://ns2/", "p");

        foo.addProperty(pt1, true);

        foo.addProperty(pt2, true);

        assertTrue(((PropertyImplRDB)pt1).getRDBId() != ((PropertyImplRDB)pt2).getRDBId());

    }



    /** Check xml:lang on literals - multibyte support depends on the database */

    public void doTestLangAttrib() throws RDFException {

        Resource fred = model.createResource("http://namespace#fred");

        Property value = model.createProperty("http://namespace#fred");

        fred.addProperty(value, "foo", "fr");

        //String mbtest = "foo\u03b0";

        //fred.addProperty(value,  mbtest, "fr");

        fred.addProperty(value, "foobar");

        Iterator it = m_driver.listStatements(fred, value, null, model);

        isStatementListOK(it, new Statement[] {

            new StatementImpl(fred, value, new LiteralImpl("foo", "fr")),

            //new StatementImpl(fred, value, new LiteralImpl(mbtest)),

            new StatementImpl(fred, value, new LiteralImpl("foobar"))

            });

        StmtIterator sit = fred.listProperties(value);

        while (sit.hasNext()) {

            Statement s = sit.next();

            sit.remove();

        }

        sit = fred.listProperties(value);

        assertTrue(!sit.hasNext());



        // Variant test for null properties and languages

        Resource fred2 = model.createResource("http://namespace#fred2");

        fred2.addProperty(value, "foo", "fr");

        fred2.addProperty(value, "foo", "");

        fred2.addProperty(value, "", "fr");

        fred2.addProperty(value, "");

        it = m_driver.listStatements(fred2, value, null, model);

        isStatementListOK(it, new Statement[] {

            new StatementImpl(fred2, value, new LiteralImpl("foo", "fr")),

            new StatementImpl(fred2, value, new LiteralImpl("foo", "")),

            new StatementImpl(fred2, value, new LiteralImpl("", "fr")),

            new StatementImpl(fred2, value, new LiteralImpl(""))

            });

    }



    /** check use of reification */

    public void doTestReification() throws RDFException {

        long size = model.size();

        Property rp = model.createProperty("http://ns3/", "reifiedAnnotation");

        Statement s1 = model.createStatement(foo, p1, "statement 1");

        assertEquals("no assert", size, model.size());



        // Statement as subject

        Statement rs1 = model.createStatement(s1, rp, "reified 1");

        assertTrue(!s1.isReified());

        model.add(rs1);

        assertTrue(s1.isReified());

        assertEquals("asserted reification", size+1, model.size());

        isStatementListOK(m_driver.listStatements(null, rp, (RDFNode)null, model), new Statement[] {rs1});        isStatementListOK(m_driver.listStatements(s1, null, (RDFNode)null, model), new Statement[] {rs1});



        // Statement as object

        Statement rs2 = model.createStatement(foo, rp, s1);

        model.add(rs2);

        assertEquals("asserted reification", size+2, model.size());

        isStatementListOK(m_driver.listStatements(null, rp, (RDFNode)null, model), new Statement[] {rs1, rs2});

        isStatementListOK(m_driver.listStatements(null, null, s1, model), new Statement[] {rs2});



        // Reification also asserted

        Statement s2 = model.createStatement(foo, p2, "statement 2");

        model.add(s2);

        assertEquals("assert s2", size+3, model.size());

        Statement rs3 = model.createStatement(s2, rp, "reified 2");

        model.add(rs3);

        assertEquals("asserted reification", size+4, model.size());

        isStatementListOK(m_driver.listStatements(null, rp, (RDFNode)null, model), new Statement[] {rs1, rs2, rs3});

        isStatementListOK(m_driver.listStatements(s2, null, (RDFNode)null, model), new Statement[] {rs3});

        isStatementListOK(m_driver.listReifiedStatements(model), new Statement[] {s1, s2});



        // Check id's tie up between statement and refering reification

        ResultSetIterator i = m_driver.listStatements(foo, p2, new LiteralImpl("statement 2"), model);

        assertTrue(i.hasNext());

        Statement rawS = (Statement)i.next();

        i.close();

        i = m_driver.listStatements(s2, null, (RDFNode)null, model);

        assertTrue(i.hasNext());

        Statement refS = (Statement)(((Statement)i.next()).getSubject());

        i.close();

        assertTrue(rawS.isAnon());

        assertTrue(refS.isAnon());

        assertEquals("statment and refernce have same anonid", rawS.getId(), refS.getId());



        // Assert reifification first

        Statement s4 = model.createStatement(foo, p2, "statement 4");

        Statement rs4 = model.createStatement(s4, rp, "reified 4");

        model.add(rs4);

        assertEquals("asserted reification only", size+5, model.size());

        isStatementListOK(m_driver.listStatements(null, rp, (RDFNode)null, model), new Statement[] {rs1, rs2, rs3, rs4});

        isStatementListOK(m_driver.listStatements(s4, null, (RDFNode)null, model), new Statement[] {rs4});

        assertTrue(!m_driver.listStatements(null, null, new LiteralImpl("statement 4"), model).hasNext());



        // Give a statement a real uri assert it then the reification list should contain it

        Statement s5 = model.createStatement(foo, p2, "statement 5");

        ((StatementImplRDB)s5).setURI(new ResourceImpl("http://namespace/mine#statement"));

        model.add(s5);

        isStatementListOK(m_driver.listReifiedStatements(model), new Statement[] {s1, s2, s4, s5});



        // Test removing all the statements

        for (StmtIterator iter = model.listStatements(); iter.hasNext(); ) {

            iter.next();

            iter.remove();

        }

        assertTrue("empty on db containing reified data", model.size() == 0);

    }



    /** Test model cleaning operations */

    public void doTestModelCleaning() throws RDFException {

        // Make sure model has at least 1 statement in it

        model.add(foo, p1, 44);

        assertTrue("non-empty db", model.size() != 0);



        // test clear operation

        model.clear();

        assertTrue("clear() failed", model.size() == 0);



        // Check model listing as well

        if (m_driver.supportsMultipleModels()) {

            ModelRDB model1 = ModelRDB.createModel(m_dbcon, "model 1");

            model1.add(foo, p1, 41);

            ModelRDB model2 = ModelRDB.createModel(m_dbcon, "model 2");

            model2.add(foo, p1, 42);

            assertTrue("model side effect detected", model.size() == 0);

            assertTrue("model creation problem", model1.size() == 1);

            assertTrue("model creation problem", model2.size() == 1);

            isStringListOK(ModelRDB.listModels(m_dbcon), new String[] {"default", "model 1", "model 2"});

            ModelRDB.deleteModel(m_dbcon, "model 1");

            assertTrue("model deletion problem", model1.size() == 0);

            assertTrue("model deletion problem", model2.size() == 1);

            isStringListOK(ModelRDB.listModels(m_dbcon), new String[] {"default", "model 2"});

        }

    }



    /** Test the bulk loader */

    public void doTestBulkLoader() throws RDFException {

        // Make sure model has at least 1 statement in it

        model.add(foo, p1, 142);

        assertTrue("non-empty db", model.size() != 0);



        Model lm = model.loadAll();

        assertTrue("bulk load error", model.size() == lm.size());



        // Repeat after a delete

        Statement s = model.createStatement(foo, p1, 143);

        model.add(s);

        model.remove(s);

        lm = model.loadAll();

        assertTrue("bulk load error", model.size() == lm.size());



        // Check contents

        ArrayList stl = new ArrayList();

        for (StmtIterator i = lm.listStatements(); i.hasNext(); ) {

            stl.add(i.next());

        }

        Statement[] sts = new Statement[stl.size()];

        sts = (Statement[])stl.toArray(sts);

        isStatementListOK(m_driver.listStatements(null, null, null, model), sts);

    }



    /**

     * Test that case insensitivity on literals is working.

     * Also checks a few odd literal handling problems that arose

     */

    protected void doTestLiteralCase() throws RDFException {

        Property p = model.createProperty("http://namespace/myprops#prop");

        Statement s1 = new StatementImpl(foo, p, new LiteralImpl("case"));

        Statement s2 = new StatementImpl(foo, p, new LiteralImpl("Case"));

        Statement s3 = new StatementImpl(foo, p, new LiteralImpl("cASE"));

        model.add(s1);

        model.add(s2);

        model.add(s3);

        isStatementListOK(m_driver.listStatements(foo, p, null, model), new Statement[] {s1, s2, s3});



        Statement s4 = new StatementImpl(foo, p, new LiteralImpl("foo "));

        model.add(s4);

        isStatementListOK(m_driver.listStatements(foo, p, null, model), new Statement[] {s1, s2, s3, s4});

    }



    /** Helper function compare a retrieved resource with a supplied one */

    protected void isResourceOK(Resource found, Resource expected) throws RDFException {

        if (expected.isAnon()) {

            assertTrue(found.isAnon());

            assertEquals(expected.getId(), found.getId());

        } else {

            assertEquals(expected.getURI(), found.getURI());

        }

    }



    /** Helper function - compare an iterator against a set of resources expected */

    protected void isResourceListOK(Iterator it, Resource[] expected) throws RDFException {

        boolean[] found = new boolean[expected.length];

        for (int i = 0; i < found.length; i++) found[i] = false;

        while (it.hasNext()) {

            Resource r = (Resource) it.next();

            boolean gotit = false;

            for (int i = 0; i < expected.length; i++) {

                if (!found[i]) {

                    if (r.isAnon()) {

                        if (expected[i].isAnon() && expected[i].getId().equals(r.getId())) {

                            gotit = true;

                        }

                    } else {

                        if (!expected[i].isAnon() && expected[i].getURI().equals(r.getURI())) {

                            gotit = true;

                        }

                    }

                    if (gotit) {

                        found[i] = true;

                        break;

                    }

                }

            }

            assertTrue(gotit);

        }

        for (int i = 0; i < expected.length; i++) {

            assertTrue(found[i]);

        }

    }



    /** Helper function - compare an iterator against a set of statements expected */

    protected void isStatementListOK(Iterator it, Statement[] expected) throws RDFException {

        boolean[] found = new boolean[expected.length];

        for (int i = 0; i < found.length; i++) found[i] = false;

        while (it.hasNext()) {

            Statement s = (Statement) it.next();

            boolean gotit = false;

            for (int i = 0; i < expected.length; i++) {

                if (!found[i] && s.equals(expected[i])) gotit = true;

                if (gotit) {

                    found[i] = true;

                    break;

                }

            }

            assertTrue(gotit);

        }

        for (int i = 0; i < expected.length; i++) {

            assertTrue(found[i]);

        }

    }



    /** Helper function - compare an iterator against a set of strings expected */

    protected void isStringListOK(Iterator it, String[] expected) throws RDFException {

        boolean[] found = new boolean[expected.length];

        for (int i = 0; i < found.length; i++) found[i] = false;

        while (it.hasNext()) {

            String s = (String) it.next();

            boolean gotit = false;

            for (int i = 0; i < expected.length; i++) {

                if (!found[i] && s.equals(expected[i])) gotit = true;

                if (gotit) {

                    found[i] = true;

                    break;

                }

            }

            assertTrue(gotit);

        }

        for (int i = 0; i < expected.length; i++) {

            assertTrue(found[i]);

        }

    }





    /** Helper function to check a retrieved statement, object is a Resource or a string */

    protected void isStatementOK(Statement st, Resource s, Property p, Object o) throws RDFException {

        isResourceOK(st.getSubject(), s);

        isResourceOK(st.getPredicate(), p);

        RDFNode obj = st.getObject();

        if (o instanceof String) {

            assertTrue(obj instanceof Literal);

            assertEquals(o, ((Literal)obj).toString());

        } else {

            assertTrue(obj instanceof Resource);

            isResourceOK((Resource)obj, (Resource)o);

        }

    }



    /** Helper function during debug - list a table */

    public void listTable(String tname) throws SQLException, RDFRDBException {

        System.out.println("Table - " + tname);

        java.sql.Statement st = m_driver.getConnection().createStatement();

        ResultSet rs = st.executeQuery("SELECT * FROM " + tname);

        ResultSetMetaData rsmd = rs.getMetaData();

        for (int i = 1; i <= rsmd.getColumnCount(); i++) {

            System.out.print("\t" + rsmd.getColumnLabel(i));

        }

        System.out.println("");

        while (rs.next()) {

            for (int i = 1; i <= rsmd.getColumnCount(); i++) {

                System.out.print("\t" + rs.getObject(i));

            }

            System.out.println("");

        }

    }



    public static  TestSuite suite(String layout, String database) {

        TestSuite suite = new TestSuite();

        suite.addTest(new TestStoreRDB("testStore", layout, database));

        suite.addTest(new TestStoreRDB("testStoreHint", layout, database));

        return suite;

    }



}



/*

 *  (c) Copyright Hewlett-Packard Company 2000, 2001

 *  All rights reserved.

 *

 * Redistribution and use in source and binary forms, with or without

 * modification, are permitted provided that the following conditions

 * are met:

 * 1. Redistributions of source code must retain the above copyright

 *    notice, this list of conditions and the following disclaimer.

 * 2. Redistributions in binary form must reproduce the above copyright

 *    notice, this list of conditions and the following disclaimer in the

 *    documentation and/or other materials provided with the distribution.

 * 3. The name of the author may not be used to endorse or promote products

 *    derived from this software without specific prior written permission.



 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR

 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES

 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.

 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,

 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT

 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,

 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY

 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT

 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF

 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 */