/*
 *  (c) Copyright Hewlett-Packard Company 2000 
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Created on 31 August 2000, 14:43
 */

package com.hp.hpl.mesa.rdf.jena.sample;

import com.hp.hpl.mesa.rdf.jena.common.*;
import com.hp.hpl.mesa.rdf.jena.mem.ModelMem;
import com.hp.hpl.mesa.rdf.jena.model.*;
import com.hp.hpl.mesa.rdf.jena.vocabulary.RDF;
import com.hp.hpl.mesa.rdf.jena.vocabulary.RSS;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStreamReader;
import java.io.PrintWriter;

import java.util.Iterator;
import java.util.Vector;

/** Render RSS 1.0 input as HTML
 * <P>This program reads a source of RSS 1.0 data and translates it to an HTML 
 *    representation.  The resulting HTML is crude - this
 *    program is intended as an illustration of how it can be done.</P>
 *
 *    Usage:<br>
 *       RenderToHTML [input URL] [output filename]<br>
 *    e.g.<br> 
 *       RenderToHTML http://purl.org/rss/1.0/convert?url=http://InternetAlchemy.org/channels/rss/cnet-news.xml c:\temp\rss.htm<br>
 *
 * @author  bwm
 * @version $Name:  $ $Revision: 1.2 $ $Date: 2001/07/04 17:54:06 $
 */
public class RenderToHTML extends Object {

   /** Render RSS 1.0 as HTML.
    * <P>This program reads the source of RSS 1.0 data whose URL is in 
    *    args[0] and writes an HTML representation to the file whose name is
    *    in args[1].  The translation is currently crude - this
    *    program is intended as an illustration of how it can be done.</P>
    *
    * 
    * @param args program arguments
    */
    public static void main (String args[]) {
        
        String rssIn = args[0];
        
        try {
            // open the output stream
            PrintWriter out = new PrintWriter(new FileOutputStream(args[1]));
            
            // read the rss stream
        
            Model model = new ModelMem();
            model.read(rssIn);
            
            // get an iterator over the channels in the file
            ResIterator channels = 
                          model.listSubjectsWithProperty(RDF.type, RSS.channel);
            
            // write the HTML file
        
            writeHTMLHeader(out);
            while (channels.hasNext()) {
                renderChannel(channels.next(), out);
            }
            writeHTMLTrailer(out);
            out.close();
            
        } catch (Exception e) {
            System.out.println(e);
        }
    }
    
    protected static void writeHTMLHeader(PrintWriter out) {
        out.println("<HTML>");
        out.println("<HEAD>");
        out.println("</HEAD>");
        out.println("<BODY>");
    }
    
    protected static void writeHTMLTrailer(PrintWriter out) {
        out.println("</BODY>");
    }
        
    protected static void renderChannel(Resource channel, PrintWriter out)
      throws RDFException {
        
        String url = null;
        
        // create the outer table
        out.print("<TABLE>\r\n"
                + "  <TR>\r\n");
            
        // begin first table entry
        out.println("    <TD>");
        
        // get the channel URL if it has one
        if (channel.hasProperty(RSS.link)) {
            url = channel.getProperty(RSS.link)
                         .getString();
        }
        
        // open the link tag if there is a url
        if (url != null) {
            out.print("<A HREF='" + url + "'>");
        }
            
        // output the title
        if (channel.hasProperty(RSS.title)) {
            out.print("      <H1>");            
            out.print(channel.getProperty(RSS.title)
                             .getString());
            out.print("</H1>");
        }
        
        // close the link tag if there is a link
        if (url != null) {
            out.print("</A>");
        }
        
        out.println("");        
            
        // output the description
        if (channel.hasProperty(RSS.description)) {
            out.print("      <P><H3>");
            out.print(channel.getProperty(RSS.description)
                             .getString());
            out.print("</H3></P>\r\n");
        }
            
        // end of first table entry
        out.print("    </TD><TD>\r\n");
            
        // output the image
        if (channel.hasProperty(RSS.image)) {
            renderImage(channel.getProperty(RSS.image)
                               .getResource(),
                        out);
        }
        
        // end of the second table entry
        out.print("    </TD><TD>\r\n");
        
        if (channel.hasProperty(RSS.textinput)) {
            renderTextInput(channel.getProperty(RSS.textinput)
                                   .getResource(),
                            out);
        }
            
        // end of the first row, start of the second
        out.print("    </TD>\r\n"
                + "  </TR>\r\n"
                + "  <TR>\r\n"
                + "    <TD>\r\n");
            
        // create the inner table
        out.print("      <TABLE>\r\n");

        // now render the items
        if (channel.hasProperty(RSS.items)) {
            Seq items = channel.getProperty(RSS.items)
                               .getSeq();
            for (int i=1; i<= items.size(); i++) {
                renderItem(items.getResource(i), out);
            }
        }
            
        // close the tables
        out.print("      </TABLE>\r\n"
                + "    </TD>\r\n"
                + "  </TR>\r\n"
                + "</TABLE>\r\n");
    }
    
    
    protected static void renderItem(Resource item, PrintWriter out) 
      throws RDFException {
        if (item == null) return;
        out.print("      <TR><TD>\r\n");
        out.print("          <B><A HREF=\"" + item.getProperty(RSS.link)
                                                  .getString() +"\">"
                                    + item.getProperty(RSS.title)
                                          .getString() + "</A></B>\r\n");
        out.print("        </TD><TD>");
        if (item.hasProperty(RSS.description)) {
            out.print("          " + item.getProperty(RSS.description)
                                         .getString() + "\r\n");
        }
        out.print("        </TD>\r\n"
                + "      </TR>");
    }
    
    protected static void renderImage(Resource image, PrintWriter out)
      throws RDFException {
        
        String url   = null;
        String title = null;
        
        // does the image have a link?
        if (image.hasProperty(RSS.link)) {
            url = image.getProperty(RSS.link)
                       .getString();
        }
        
        // does the image have a title?
        if (image.hasProperty(RSS.title)) {
            title = image.getProperty(RSS.title)
                         .getString();
        }
        
        // open the link tag if necessary
        if (url != null) {
            out.print("<A HREF='" + url + "'>");
        }
        
        // write the image tag        
       out.print("<IMG SRC='" + image.getURI() + "'");
       if (title != null) {
        out.print(" ALT='" + title + "'");
       }
       out.print(">");
    }
    
    protected static void renderTextInput(Resource ti, PrintWriter out)
      throws RDFException {
        
        String url = null;
        String title = null;
        String name = null;
        String description = null;
        
        if (ti.hasProperty(RSS.link)) {
            url = ti.getProperty(RSS.link)
                    .getString();
        } else return;
        if (ti.hasProperty(RSS.title)) {
            title = ti.getProperty(RSS.title)
                      .getString();
        } else return;
        if (ti.hasProperty(RSS.name)) {
            name = ti.getProperty(RSS.name)
                     .getString();
        } else return;
        if (ti.hasProperty(RSS.description)) {
            description = ti.getProperty(RSS.description)
                            .getString();
        } else return;
        
        out.println("<FORM METHOD='get' ACTION='" + url + "'>");
        out.println("<P>" + title + " <INPUT NAME='" + name + "' size='20'></p>");
        out.println("<p>" + description + "</P>");
        out.println("</FORM>");
    }
}