/*
 *  (c) Copyright Hewlett-Packard Company 2000 
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Sample2.java
 *
 * Example of using an enhanced resource class
 *
 * Created on 16 August 2000, 14:47
 */

package com.hp.hpl.mesa.rdf.jena.sample;

import com.hp.hpl.mesa.rdf.jena.common.*;
import com.hp.hpl.mesa.rdf.jena.model.*;
import com.hp.hpl.mesa.rdf.jena.mem.ModelMem;
import com.hp.hpl.mesa.rdf.jena.vocabulary.RDF;

/** Simple example of application defined enhanced resources..
 *
 *<p>This example uses an enhanced resource class Month.  An enhanced resource
 *   class encapsulates an ordinary resource and adds extra behaviour to it.
 *   A month resource is a sequence with a member for each day of the month.
 *   The enhanced functionality is to create a sequence with the correct
 *   number of days for a given month in a given year.</p>
 *<p>This class creates a month as anonymous resource, then uses a factory
 *   object to retrieve it from the model.</p>
 *
 * @author  bwm
 * @version Release='$Name:  $' Revision='$Revision: 1.2 $' Date='$Date: 2002/01/05 11:54:14 $'
 */
public class Sample2 extends Object {

    /**
    * @param args the command line arguments
    */
    public static void main (String args[]) {
        new Sample2();
    }
    
    // the main body of the application
    
    protected Sample2() {
        
        try {
        
            // create a new in memory model
            
            Model model = new ModelMem();
            
            // create a new month
        
            Month january = new Month(Month.JAN, 2000, model);
            
            // using the Date:numDays property of the month, process each
            // day resource in the month
        
            for (int i=1; i<=january.getProperty(Date.numDays).getInt(); i++) {
                
                // write out the day number of the day
                
                System.out.println(january.getResource(i)
                                          .getProperty(Date.num).getInt());
            }
            
            // create a new sequence object and attach the month to it
            
            Seq year = model.createSeq();
            year.add(january);
            
            // use a month factory to retrieve an instance of month from
            // the model by retrieving the first month in year.
            
            Month month = (Month) year.getResource(1, new MonthF());
            
            // and once again process the contents of month
            
            for (int i=1; i<=month.getProperty(Date.numDays).getInt(); i++) {
                
                // and process each day
                
                System.out.println(january.getResource(i)
                                          .getProperty(Date.num).getInt());
            }
            
            
        } catch (Exception e) {
            System.out.println(e);
        }
    }
}

// Month is an extended resource class.  A month is an RDF sequence with an
// anonymous resource for each day in the month.  The month class calculates
// the correct number of days.  A month resource also has properties Date.name
// for the name of the month, Date.numDays for the number of days in the
// month and Date.year for the year.

// Month is subclassed from EnhResource which takes care of some of the
// mechanics of making encapsulation work.  The mechanism of encapsulation
// was chosen so that enhanced resources such as Month could work with any
// model implementation, even if that implementation has its own implementations
// of Resource, Property etc.
    
class Month extends EnhSeq {
    
    // public constants for the months of the year
        
    static final int JAN = 1;
    static final int FEB = 2;
    static final int MAR = 3;
    static final int APR = 4;
    static final int MAY = 5;
    static final int JUN = 6;
    static final int JUL = 7;
    static final int AUG = 8;
    static final int SEP = 9;
    static final int OCT = 10;
    static final int NOV = 11;
    static final int DEC = 12;
    
    // the names of the months, indexed by constants above
    // note: the indexes don't start at 0
        
    static final String[] monthNames = 
        {"", "January", "February", "March", "April", "May", "June",
         "July", "August", "September", "October", "November", "December"};
        
    int month;   // the month number of this month
    int year;    // the year of this month
    
    // construct a new month resource in a model
                        
    protected Month (int month, int year, Model model) throws RDFException {
        
        // create the sequence which will represent this month
        // and call the superclass which handles some small mechanics
        
        super(model);
        
        // set up the instance variables
        
        this.month = month;
        this.year  = year;
        
        // add the appropriate properties to the resource
      
        resource.addProperty(RDF.type, Date.Month)
                .addProperty(Date.name, monthNames[month])
                .addProperty(Date.num, month)
                .addProperty(Date.year, year)
                .addProperty(Date.numDays, numDays(month));
        
        // create an anonymous resource for each day in the month,
        // attach its day number property to it and insert it into
        // the sequence.
        
        for (int i=1;i<=numDays(month); i++) {
            seq.add(model.createResource().addProperty(RDF.type, Date.Day)
                                          .addProperty(Date.num, i));
        }
    }
    
    // This constructor is called by the factory object when a Month is
    // retrieved from a model.  Creation of months is done with the
    // constructor above and not by callind the createResource method in
    // a model.
    
    public Month(Resource r) throws RDFException {
        
        // first call the super class constructor
        
        super(r);
        
        // Assume that the resource already exists in the model,
        // so its state can be retrieved by accessing its properties.
        // There ought to have some exception handling here
        // in case the properties don't exist.
        
        month = seq.getProperty(Date.num).getInt();
        year  = seq.getProperty(Date.year).getInt();
        seq = (Seq) resource;
    }
    
    // a function to compute the number of days in the month.
    
    static final int[] numDays = {0,31,0,31,30,31,30,31,31,30,31,30,31};
    
    protected int numDays(int month) {
        if (month != FEB) {
            return numDays[month];
        } else {
            if ( (year%4 != 0) || ((year%100==0) && (year%400!=0))) {
                return 28;
            } else {
                return 29;
            }
        }
    }
}

// A factory class for creating Month instances
    
class MonthF implements ResourceF{
    public Resource createResource(Resource r) throws RDFException {
        return new Month(r);
    }
}

// A simple vocabulary class defining the classes and properties used in
// this example.
    
class Date {
    static final String URI = "http://aldabaran.hp.hpl.com/month/";
    
           static final String   nMonth   = "Month";
    public static       Resource  Month   = null;
           static final String   nDay     = "Day";
    public static       Resource  Day     = null;
           static final String   nname    = "name";
    public static       Property  name    = null;
           static final String   nnum     = "num";
    public static       Property  num     = null;
           static final String   nnumDays = "numDays";
    public static       Property  numDays = null;
           static final String   nyear    = "year";
    public static       Property  year    = null;
    
    static {
        try {
            Month   = new ResourceImpl(URI+nMonth);
            Day     = new ResourceImpl(URI+nDay);
            name    = new PropertyImpl(URI, nname);
            num     = new PropertyImpl(URI, nnum);
            numDays = new PropertyImpl(URI, nnumDays);
            year    = new PropertyImpl(URI, nyear);
        } catch (Exception e) {
            System.out.println(e);
        }
    }
}